Require Import Metatheory.

Require Import Decidable.
Require Import List.

Require Import indexes.

Set Implicit Arguments.

Open Scope list_scope.
Open Scope type_scope.

Hint Constructors index.

(** Locations, thread IDs, and locks are all represented by the same structure:
   vars that support decidable equality. *)
Definition loc := var.
Definition tid := var.
Definition lock := var.

Section Actions.
  (** val is the set of values stored in the heap. *)
  Variable val : Set.

  (** Actions are visible both to the heap and to the program. *)
  Inductive action : Set :=
  | pure : action
  | alloc : loc -> val -> action
  | read : loc -> val -> action
  | write : loc -> val -> action
  | acquire : lock -> action
  | release : lock -> action
  | spawn : tid -> action.

  (** Every step taken by the program, the heap, or the system has a (possibly
     empty) effect. *)
  Definition effect := option (tid * action).

  (** A program trace is a list of effects. *)
  Definition trace := list effect.

  (** An "erased" trace is a list of effects with empty effects removed. *)
  Definition erased_trace := list (tid * action).

  (** Remove the empty effects from the trace. *)
  Fixpoint erase_trace (tr : trace) : erased_trace :=
    match tr with
      | nil => nil
      | None :: tr' => erase_trace tr'
      | Some (t, a) :: tr' => (t, a) :: (erase_trace tr')
    end.

  (** Convert an erased trace by to an "unerased" trace *)
  Fixpoint unerase_trace (tr : erased_trace) : trace :=
    match tr with
      | nil => nil
      | (t, a) :: tr => (Some (t,a)) :: (unerase_trace tr)
    end.

  Lemma erase_unerase : forall (tr : erased_trace),
    tr = erase_trace (unerase_trace tr).
  Proof.
    intros tr.
    induction tr as [ | [ t a ] tr ];
      simpl; f_equal; auto.
  Qed.

  Lemma app_erase : forall tr1 tr2,
    erase_trace (tr1 ++ tr2) = erase_trace tr1 ++ erase_trace tr2.
  Proof.
    intros tr1 tr2.
    induction tr1 as [ | [ [ t a ] | ] tr1 ]; simpl; f_equal; eauto.
  Qed.

  Lemma app_unerase : forall tr1 tr2,
    unerase_trace (tr1 ++ tr2) = unerase_trace tr1 ++ unerase_trace tr2.
  Proof.
    intros tr1 tr2.
    induction tr1 as [ | [ t a ] tr1 ]; simpl; f_equal; eauto.
  Qed.

  Lemma unerase_unique : forall tr1 tr2,
    unerase_trace tr1 = unerase_trace tr2 ->
    tr1 = tr2.
  Proof.
    intros tr1 tr2 eq.
    generalize dependent tr2.
    induction tr1 as [ | [ t1 a1 ] ]; intros tr2 eq;
        induction tr2 as [ | [ t2 a2 ] ];
          simpl in *; auto; inversion eq.
    subst. f_equal. auto.
  Qed.

  Lemma unerase_length : forall tr,
    length tr = length (unerase_trace tr).
  Proof.
    intros tr.
    induction tr as [ | [ ? ? ] ]; simpl; auto.
  Qed.

  (** A sequential trace is a trace with no empty effects. *)
  Inductive sequential_trace : trace -> Prop :=
  | sequential_trace_nil :
    sequential_trace nil
  | sequential_trace_cons : forall tr t a,
    sequential_trace tr ->
    sequential_trace (Some (t, a) :: tr).

  Lemma sequential_trace_index : forall tr j t a,
    sequential_trace tr ->
    index (erase_trace tr) j (t, a) ->
    index tr j (Some (t, a)).
  Proof.
    intros tr j t a seq idx.
    generalize dependent j.
    induction seq; intros j idx; inversion idx; subst; eauto.
  Qed.

  Lemma sequential_trace_erase_index : forall tr j t a,
    sequential_trace tr ->
    index tr j (Some (t, a)) ->
    index (erase_trace tr) j (t, a).
  Proof.
    intros tr j t a seq idx.
    generalize dependent j.
    induction seq; intros j idx; simpl; inversion idx; subst; eauto.
  Qed.

  Lemma index_unerase : forall tr j t a,
    index tr j (t, a) ->
    index (unerase_trace tr) j (Some (t, a)).
  Proof.
    intros tr j t a idx.
    induction idx as [ ? [ ? ? ] | tr ? [ ? ? ] [ ? ? ] ].
    apply index_0. simpl. apply index_S. auto.
  Qed.

  Lemma index_unerase_inversion : forall tr j t a,
    index (unerase_trace tr) j (Some (t, a)) ->
    index tr j (t, a).
  Proof.
    intros tr j t a idx.
    generalize dependent j.
    induction tr as [ | [ t' a' ] ]; intros j idx;
      simpl in idx; inversion idx; subst; auto.
  Qed.

  Lemma sequential_trace_index_inversion : forall tr j e,
    sequential_trace tr ->
    index tr j e ->
    exists t, exists a, e = (Some (t, a)).
  Proof.
    intros tr j e seq idx.
    generalize dependent e.
    generalize dependent j.
    induction tr as [ | e' tr ]; intros j e idx.
    inversion idx.
    inversion seq; subst.
    inversion idx; subst; eauto.
  Qed.

  Lemma sequential_trace_erase : forall tr,
    sequential_trace tr ->
    tr = unerase_trace (erase_trace tr).
  Proof.
    intros tr seq.
    induction seq.
    simpl. auto.
    simpl. rewrite <- IHseq. auto.
  Qed.

  Lemma sequential_trace_unerase : forall tr,
    sequential_trace (unerase_trace tr).
  Proof.
    intros tr.
    induction tr as [ | [ t a ] ]; simpl;
      auto using sequential_trace_nil, sequential_trace_cons.
  Qed.

  Lemma sequential_trace_append : forall tr1 tr2,
    sequential_trace tr1 ->
    sequential_trace tr2 ->
    sequential_trace (tr1 ++ tr2).
  Proof.
    intros tr1 tr2 seq1 seq2.
    induction seq1; simpl; eauto using sequential_trace_cons.
  Qed.

  Lemma sequential_trace_append_inversion : forall tr1 tr2,
    sequential_trace (tr1 ++ tr2) ->
    sequential_trace tr1 /\ sequential_trace tr2.
  Proof.
    intros tr1 tr2 seq.
    induction tr1; simpl in *.
    auto using sequential_trace_nil.
    inversion seq as [ | ? ? ? seq' ]; subst.
    pose proof IHtr1 seq' as [ ? ? ].
    auto using sequential_trace_cons.
  Qed.

  Lemma sequential_trace_length : forall tr,
    sequential_trace tr ->
    length tr = length (erase_trace tr).
  Proof.
    intros tr seq_tr.
    induction seq_tr; simpl; auto.
  Qed.

  Lemma index_erase : forall tr j t a,
    index tr j (Some (t, a)) ->
    exists j', index (erase_trace tr) j' (t, a).
  Proof.
    intros tr j t a idx.
    generalize dependent j.
    induction tr as [ | e ]; intros j idx; simpl in *.
    inversion idx.
    induction e as [ [ t' a' ] | ]; simpl in *.
    inversion idx as [ | ? ? ? ? idx' ]; subst.
    eauto.
    pose proof IHtr _ idx' as [ ? ? ]. eauto.
    inversion idx as [ | ? ? ? ? idx' ]; subst.
    pose proof IHtr _ idx' as [ ? ? ]. eauto.
  Qed.

  Lemma index_unerase_erase : forall tr j t a,
    index (unerase_trace (erase_trace tr)) j (Some (t, a)) ->
    exists j', index tr j' (Some (t, a)).
  Proof.
    intros tr j t a idx.
    generalize dependent j.
    induction tr as [ | e ]; intros j idx; simpl in *.
    inversion idx.
    induction e as [ [ t' a' ] | ]; simpl in *.
    inversion idx as [ | ? ? ? ? idx' ]; subst.
    eauto.
    pose proof IHtr _ idx' as [ ? ? ]. eauto.
    pose proof IHtr _ idx as [ ? ? ]. eauto.
  Qed.

  (** We define a conflict relation over actions. Actions conflict if they
     read/write/alloc the same location (unless both actions are reads) or if
     they acquire/release the same lock. *)
  Inductive actions_conflict : action -> action -> Prop :=
  | conflict_alloc_alloc : forall r v v',
    actions_conflict (alloc r v) (alloc r v')
  | conflict_alloc_read : forall r v v',
    actions_conflict (alloc r v) (read r v')
  | conflict_read_alloc : forall r v v',
    actions_conflict (read r v) (alloc r v')
  | conflict_alloc_write : forall r v v',
    actions_conflict (alloc r v) (write r v')
  | conflict_write_alloc : forall r v v',
    actions_conflict (write r v) (alloc r v')
  | conflict_write_write : forall r v v',
    actions_conflict (write r v) (write r v')
  | conflict_read_write : forall r v v',
    actions_conflict (read r v) (write r v')
  | conflict_write_read : forall r v v',
    actions_conflict (write r v) (read r v')
  | conflict_acquire_acquire : forall l,
    actions_conflict (acquire l) (acquire l)
  | conflict_acquire_release : forall l,
    actions_conflict (acquire l) (release l)
  | conflict_release_acquire : forall l,
    actions_conflict (release l) (acquire l)
  | conflict_release_release : forall l,
    actions_conflict (release l) (release l).

  Hint Constructors actions_conflict.

  (* The conflict relation is symmetric. *)
  Lemma actions_conflict_sym : forall a1 a2,
    actions_conflict a1 a2 ->
    actions_conflict a2 a1.
  Proof.
    intros a1 a2 conflict.
    induction conflict; auto.
  Qed.

  Lemma actions_conflict_dec : forall a1 a2,
    {actions_conflict a1 a2} + {~ actions_conflict a1 a2}.
  Proof.
    intros a1 a2.
    induction a1 as [ | r1 | r1 | r1 | l1 | l1 | t1 ];
      induction a2 as [ | r2 | r2 | r2 | l2 | l2 | t2 ];
        try (right; intro false; inversion false; fail);
          try (destruct (eq_var_dec r1 r2);
            [ subst; auto | right; intro false; inversion false; tauto ]; fail);
          try (destruct (eq_var_dec l1 l2);
            [ subst; auto | right; intro false; inversion false; tauto ]; fail).
  Qed.

  (* Two effects conflict if their actions conflict, if they are performed by
     the same thread, or if one is a spawn of the other's thread. *)
  Inductive effects_conflict : tid * action -> tid * action -> Prop :=
  | effects_conflict_actions : forall t1 a1 t2 a2,
    actions_conflict a1 a2 ->
    effects_conflict (t1, a1) (t2, a2)
  | effects_conflict_program_order : forall t a1 a2,
    effects_conflict (t, a1) (t, a2)
  | effects_conflict_spawn_left : forall t1 t2 a2,
    effects_conflict (t1, spawn t2) (t2, a2)
  | effects_conflict_spawn_right : forall t1 a1 t2,
    effects_conflict (t1, a1) (t2, spawn t1).

  Hint Constructors effects_conflict.

  (* The effect conflict relation is symmetric. *)
  Lemma effects_conflict_sym : forall e1 e2,
    effects_conflict e1 e2 ->
    effects_conflict e2 e1.
  Proof.
    intros e1 e2 conflict.
    inversion conflict; subst;
      auto using actions_conflict_sym.
  Qed.

  (* The effect conflict relation is decidable. *)
  Lemma effects_conflict_dec : forall e1 e2,
    {effects_conflict e1 e2} + {~ effects_conflict e1 e2}.
  Proof.
    intros e1 e2.
    destruct e1 as [ t1 a1 ]. destruct e2 as [ t2 a2 ].
    destruct (eq_var_dec t1 t2) as [ | t1_neq_t2 ].
    subst. auto.
    destruct (actions_conflict_dec a1 a2) as [ | not_conflict ]. auto.
    induction a1 as [ | | | | | | t3 ];
      induction a2 as [ | | | | | | t4 ];
        try (right; intro conflict; inversion conflict; subst; tauto);
          try (destruct (eq_var_dec t1 t4);
            [ subst; auto |
              right; intro conflict; inversion conflict; subst; tauto ]);
          try (destruct (eq_var_dec t2 t3);
            [ subst; auto |
              right; intro conflict; inversion conflict; subst; tauto ]).
    destruct (eq_var_dec t1 t4); destruct (eq_var_dec t2 t3); subst; auto.
    right. intro conflict. inversion conflict; subst; tauto.
  Qed.

  (** A projection is a list of actions performed by a particular thread. *)
  Definition projection := list action.

  (** trace_projection tr t is the subsequence of actions in tr performed by
     thread t. *)
  Fixpoint trace_projection (tr : erased_trace) (t : tid) {struct tr}
    : projection :=
    match tr with
      | nil => nil
      | (t', a) :: tr' =>
        if eq_var_dec t t'
          then a :: trace_projection tr' t
          else trace_projection tr' t
    end.

  (* The projection of two appended traces is the append of the projection of
     the two traces, and vice versa. *)
  Lemma app_projection : forall tr1 tr2 t,
    trace_projection (tr1 ++ tr2) t
    = trace_projection tr1 t ++ trace_projection tr2 t.
  Proof with auto.
    intros tr1 tr2 t.
    induction tr1 as [ | [ t' a ] tr1' ]; simpl...
    destruct (eq_var_dec t t')...
    rewrite IHtr1'...
  Qed.

  (** This is the standard definition of a data race. *)
  Inductive data_race : action -> action -> Prop :=
  | data_race_rw : forall r v v',
    data_race (read r v) (write r v')
  | data_race_wr : forall r v v',
    data_race (write r v) (read r v')
  | data_race_ww : forall r v v',
    data_race (write r v) (write r v').

  Hint Constructors data_race.

  (** Note that the conflict relation contains the data race relation. *)
  Lemma data_race_implies_conflict : forall a1 a2,
    data_race a1 a2 ->
    actions_conflict a1 a2.
  Proof.
    intros a1 a2 race12.
    inversion race12; auto.
  Qed.
End Actions.
