Require Import Arith.
Require Import Decidable.
Require Import FMaps.
Require Import Metatheory.
Require Import Morphisms.

Require Import actions.
Require Import buffers.
Require Import models.

Module VarPairMap.
  Include Type S with Module E := VarPair.
  Axiom extensionality : forall elt (M1 M2 : t elt),
    Equal M1 M2 ->
    M1 = M2.
End VarPairMap.

Set Implicit Arguments.

Open Scope nat_scope.
Open Scope signature_scope.

Module BufferFMap (L : LANGUAGE_MODEL) : BUFFER L.
  Import VarPairMap.
  Import L.

  (* A buffer is a partial map from TID/location pairs to queues of values. *)
  Definition buffer := t (list val).

  (* Enqueue a value in the buffer *)
  Definition enqueue (B : buffer) (t : tid) (r : loc) (v : val) : buffer :=
    add (t, r) (match find (t, r) B with
                  | Some q => q ++ v :: nil
                  | None => v :: nil
                end) B.

  (* Try to dequeue a value from the buffer *)
  Definition dequeue (B : buffer) (t : tid) (r : loc) : option (val * buffer) :=
    match find (t, r) B with
      | Some (v :: q) => Some (v, add (t, r) q B)
      | Some nil => None
      | None => None
    end.

  Fixpoint top_queue (A : Set) (l : list A) : option A :=
    match l with
      | nil => None
      | a :: nil => Some a
      | a :: l' => top_queue l'
    end.

  Definition top (B : buffer) (t : tid) (r : loc) : option val :=
    match find (t, r) B with
      | Some q => top_queue q
      | None => None
    end.

  (* The empty buffer. *)
  Definition empty := empty (list val).

  (* Holds if the queue for this pair is nil. *)
  Definition is_empty_queue (B : buffer) (t : tid) (r : loc) : Prop :=
    match find (t, r) B with
      | Some q => q = nil
      | None => True
    end.

  (* Holds if the entire buffer has no values enqueued. *)
  Definition is_empty (B : buffer) := forall t r, is_empty_queue B t r.

  (* Holds if there are no values enqueued for a thread. *)
  Definition is_empty_tid B t := forall r, is_empty_queue B t r.

  (* Holds if there are no values enqueued for a thread. *)
  Definition is_empty_loc B r := forall t, is_empty_queue B t r.

  (* A predicate holds for every value in a queue. *)
  Definition for_all_queue (B : buffer) (P : val -> Prop) (t : tid) (r : loc) :=
    match find (t, r) B with
      | None => True
      | Some q => forall v, List.In v q -> P v
    end.

  Definition for_all_loc B P r := forall t, for_all_queue B P t r.

  Definition for_all_tid B P t := forall r, for_all_queue B P t r.

  Definition for_all B P := forall t r, for_all_queue B P t r.

  Module MapFacts := FMapFacts.Facts VarPairMap.
  Module MapProperties := FMapFacts.Properties VarPairMap.

  (* The empty buffer lives up to its name. *)
  Lemma empty_is_empty : is_empty empty.
  Proof.
    intros t1 r1. unfold is_empty_queue.
    rewrite* MapFacts.empty_o.
  Qed.

  Lemma is_empty_queue_dequeue : forall B t r t' r' v B',
    is_empty_queue B t r ->
    dequeue B t' r' = Some (v, B') ->
    is_empty_queue B' t r.
  Proof.
    intros B1 t1 r1 t2 r2 v2 B2 Emp Deq.
    unfold dequeue in Deq.
    remember (find (t2, r2) B1) as maybe_q2.
    destruct maybe_q2 as [ [ | v2_ q2 ] | ]; inversion Deq.
    subst v2_ B2.
    unfold is_empty_queue in *.
    destruct ((t2, r2) ==== (t1, r1)) as [ [ ? ? ] | ].
    simpl in *. subst.
    rewrite <- Heqmaybe_q2 in Emp. inversion Emp.
    rewrite* MapFacts.add_neq_o.
  Qed.

  Lemma is_empty_queue_dequeue_diff : forall B t r v B' t' r',
    dequeue B t r = Some (v, B') ->
    ~ VarPair.eq (t, r) (t', r') ->
    (is_empty_queue B' t' r' <-> is_empty_queue B t' r').
  Proof.
    intros B1 t1 r1 v1 B2 t2 r2 Deq Neq.
    unfold dequeue in Deq.
    destruct (find (t1, r1) B1) as [ [ | ] | ]; inversion Deq.
    subst. unfold is_empty_queue.
    rewrite* MapFacts.add_neq_o.
  Qed.

  Lemma dequeue_not_is_empty_queue : forall B t r v B',
    dequeue B t r = Some (v, B') ->
    ~ is_empty_queue B t r.
  Proof.
    intros B1 t1 r1 v1 B2 Deq Emp.
    unfold dequeue, is_empty_queue in *.
    destruct (find (t1, r1) B1) as [ [ | ] | ]; inversion Deq.
    inversion Emp.
  Qed.

  Lemma is_empty_queue_enqueue_diff : forall B t r v t' r',
    (~ VarPair.eq (t, r) (t', r')) ->
    (is_empty_queue B t' r' <-> is_empty_queue (enqueue B t r v) t' r').
  Proof.
    intros B1 t1 r1 v1 t2 r2 Neq.
    unfold enqueue, is_empty_queue.
    rewrite* MapFacts.add_neq_o.
  Qed.

  Lemma dequeue_confluence : forall B t1 r1 v1 B1 t2 r2 v2 B2,
    dequeue B t1 r1 = Some (v1, B1) ->
    dequeue B t2 r2 = Some (v2, B2) ->
    (~ VarPair.eq (t1, r1) (t2, r2)) ->
    exists B',
      dequeue B1 t2 r2 = Some (v2, B') /\ dequeue B2 t1 r1 = Some (v1, B').
  Proof.
    intros B t1 r1 v1 B1 t2 r2 v2 B2 Deq1 Deq2 Neq.
    unfold dequeue in *.
    remember (t1, r1) as k1. remember (t2, r2) as k2.
    clear t1 r1 t2 r2 Heqk1 Heqk2.
    remember (find k1 B2) as maybe_q1'.
    remember (find k2 B1) as maybe_q2'.
    remember (find k1 B) as maybe_q1.
    destruct maybe_q1 as [ [ | v1_ q1 ] | ]; inversion Deq1; subst v1_ B1.
    remember (find k2 B) as maybe_q2.
    destruct maybe_q2 as [ [ | v2_ q2 ] | ]; inversion Deq2; subst v2_ B2.
    rewrite* MapFacts.add_neq_o in Heqmaybe_q1'.
    rewrite* MapFacts.add_neq_o in Heqmaybe_q2'.
    assert (maybe_q1' = Some (v1 :: q1)) as Heqq1. subst. auto.
    assert (maybe_q2' = Some (v2 :: q2)) as Heqq2. subst. auto.
    rewrite Heqq1. rewrite Heqq2.
    exists (add k2 q2 (add k1 q1 B)).
    split. auto. f_equal. f_equal.
    apply VarPairMap.extensionality.
    intros k3.
    destruct (k1 ==== k3).
    rewrite* MapFacts.add_eq_o.
    rewrite* MapFacts.add_neq_o.
    rewrite* MapFacts.add_eq_o.
    intro false. apply Neq.
    apply VarPair.eq_trans with (y := k3); auto.
    apply* VarPair.eq_sym. auto.
    destruct (k2 ==== k3).
    rewrite* MapFacts.add_neq_o.
    rewrite* MapFacts.add_eq_o.
    rewrite* MapFacts.add_eq_o.
    rewrite* MapFacts.add_neq_o.
    rewrite* MapFacts.add_neq_o.
    rewrite* MapFacts.add_neq_o.
    rewrite* MapFacts.add_neq_o.
    intro. apply Neq.
    apply VarPair.eq_sym. auto.
  Qed.

  Lemma dequeue_inversion : forall B t r v B',
    dequeue B t r = Some (v, B') ->
    exists q, find (t, r) B = Some (v :: q) /\
      B' = add (t, r) q B.
  Proof.
    intros B1 t1 r1 v1 B2 Deq.
    unfold dequeue in Deq.
    remember (find (t1, r1) B1) as q1.
    destruct q1 as [ [ | v1_ q1 ] | ]; inversion Deq.
    subst. eauto.
  Qed.

  Lemma dequeue_enqueue : forall B t r v B' t' r' v',
    dequeue B t r = Some (v, B') ->
    dequeue (enqueue B t' r' v') t r = Some (v, enqueue B' t' r' v').
  Proof.
    intros B1 t1 r1 v1 B2 t2 r2 v2 Deq.
    pose proof dequeue_inversion _ _ _ Deq as [ q1 [ q1_eq B2_eq ] ].
    subst B2.
    unfold dequeue, enqueue in *.
    remember (t1, r1) as k1.
    remember (t2, r2) as k2.
    clear t1 r1 t2 r2 Heqk1 Heqk2.
    destruct (k2 ==== k1) as [ Eq | Neq ].
    rewrite* MapFacts.add_eq_o.
      rewrite* MapFacts.add_eq_o.
      rewrite* (MapFacts.find_o B1 Eq).
      rewrite q1_eq. simpl.
      f_equal. f_equal.
      apply extensionality.
      intro k3. destruct (k1 ==== k3) as [ | Neq ].
      assert (VarPair.eq k2 k3).
        apply VarPair.eq_trans with (y := k1); auto.
        rewrite* MapFacts.add_eq_o.
        rewrite* MapFacts.add_eq_o.
      assert (~ VarPair.eq k2 k3).
        intro. apply Neq.
        apply VarPair.eq_trans with (y := k2).
        apply VarPair.eq_sym. auto. auto.
        rewrite* MapFacts.add_neq_o.
        rewrite* MapFacts.add_neq_o.
        rewrite* MapFacts.add_neq_o.
        rewrite* MapFacts.add_neq_o.
      apply VarPair.eq_sym. auto.
    rewrite* MapFacts.add_neq_o.
      rewrite* MapFacts.add_neq_o.
      rewrite* q1_eq.
      f_equal. f_equal.
      apply extensionality.
      intros k3. destruct (k1 ==== k3) as [ | neq13 ].
        rewrite* MapFacts.add_eq_o.
        rewrite* MapFacts.add_neq_o.
        rewrite* MapFacts.add_eq_o.
        intro eq23. apply Neq.
        apply VarPair.eq_trans with (y := k3). auto.
        apply VarPair.eq_sym. auto.
      rewrite* MapFacts.add_neq_o.
        destruct (k2 ==== k3) as [ | neq23 ].
        rewrite* MapFacts.add_eq_o.
          rewrite* MapFacts.add_eq_o.
        rewrite* MapFacts.add_neq_o.
          rewrite* MapFacts.add_neq_o.
          rewrite* MapFacts.add_neq_o.
      intro. apply Neq. apply VarPair.eq_sym. auto.
  Qed.

  (* eq_nat is an equivalence relation. *)
  Lemma Equivalence_eq_nat : Equivalence eq_nat.
  Proof.
    split. intro. apply eq_nat_refl.
    intros x y Eq. apply eq_eq_nat.
      symmetry. apply eq_nat_eq. auto.
    intros x y z Eq1 Eq2.
      apply eq_nat_eq in Eq1. apply eq_nat_eq in Eq2.
      apply eq_eq_nat. subst. auto.
  Qed.

  (* Used by the size funtion. *)
  Definition add_length (k : key) (q : list val) (n : nat) := length q + n.

  (* Number of values in the buffer. *)
  Definition size (B : buffer) :=
    fold add_length B 0.

  (* add_length is a morphism (whatever that means). *)
  Lemma Morphism_add_length :
    Morphism (VarPair.eq ==> Leibniz ==> eq_nat ==> eq_nat) add_length.
  Proof.
    assert (Reflexive (VarPair.eq==>(@eq (list val))==>eq_nat==>eq_nat)).
      intros g [ t1 r1 ] [ t2 r2 ] Eq.
      unfold VarPair.eq in Eq. simpl in Eq.
      destruct Eq. subst.
      intros q1 q2 Eq. subst.
      intros x y Eq. apply eq_nat_eq in Eq. subst.
      apply eq_eq_nat. auto.
    apply reflexive_morphism.
  Qed.

  (* add_length is commutative. *)
  Lemma transpose_add_length : MapProperties.transpose_neqkey eq_nat add_length.
  Proof.
    intros p1 p2 q1 q2 n Neq.
    unfold add_length.
    apply eq_eq_nat. ring.
  Qed.

  (* "Equal" buffers have the same size. *)
  Lemma size_Equal (B1 B2 : buffer) :
    Equal B1 B2 ->
    size B1 = size B2.
  Proof.
    intros B1 B2 Eq.
    apply eq_nat_eq.
    unfold size.
    exact (MapProperties.fold_Equal (elt := list val) Equivalence_eq_nat
      Morphism_add_length transpose_add_length 0 Eq).
  Qed.

  (* Adding a queue to a buffer increases the size by the length of the queue (if there was
     no queue for that pair previously).*)
  Lemma size_add : forall B t r q,
    ~ In (t, r) B ->
    size (add (t, r) q B) = length q + size B.
  Proof.
    intros B t1 r1 q not_in.
    apply eq_nat_eq.
    unfold size.
    pose proof (MapProperties.fold_add (elt := list val) Equivalence_eq_nat
      Morphism_add_length transpose_add_length q 0 not_in) as Eq.
    unfold add_length in *.
    auto.
  Qed.

  (* Removing a queue from a buffer decreases the size by the length of the queue. *)
  Lemma size_remove : forall B t r q,
    MapsTo (t, r) q B ->
    size B = length q + size (remove (t, r) B).
  Proof.
    intros B t1 r1 q1 Maps.
    assert (Equal B (add (t1, r1) q1 (remove (t1, r1) B))) as Eq.
    intros [ t2 r2 ].
    destruct ((t1, r1) ==== (t2, r2)) as [ [ ? ? ] | Neq ].
    simpl in *. subst.
      rewrite* MapFacts.add_eq_o. apply* find_1. auto.
    rewrite* MapFacts.add_neq_o.
      rewrite* MapFacts.remove_neq_o.
    apply size_Equal in Eq. rewrite Eq.
      rewrite size_add. auto.
      apply remove_1. apply VarPair.eq_refl.
  Qed.

  (* Removing a mapping from a buffer is that same as adding a different mapping for that
     same pair, then removing it again. *)
  Lemma MapsTo_add_remove : forall (B : buffer) t1 r1 q1 q2,
    MapsTo (t1, r1) q1 B ->
    Equal (remove (t1, r1) B) (remove (t1, r1) (add (t1, r1) q2 B)).
  Proof.
    intros B t1 r1 q1 q2 Maps [ t2 r2 ].
    destruct ((t1, r1) ==== (t2, r2)) as [ Eq | Neq ].
    rewrite* MapFacts.remove_eq_o.
    rewrite* MapFacts.remove_eq_o.
    rewrite* MapFacts.remove_neq_o.
    rewrite* MapFacts.remove_neq_o.
    rewrite* MapFacts.add_neq_o.
  Qed.

  (* Dequeueing a value decreases the size by 1. *)
  Lemma dequeue_size : forall B t r v B',
    dequeue B t r = Some (v, B') ->
    size B = size B' + 1.
  Proof.
    intros B t1 r1 v1 B' Eq.
    rewrite plus_comm.
    unfold dequeue in Eq.
    remember (find (t1, r1) B) as maybe_q1.
    symmetry in Heqmaybe_q1.
    destruct maybe_q1 as [ q1 | ].
    destruct q1 as [ | v q1 ];
      inversions Eq.

    (* buffer contains a value for this pair *)
    apply find_2 in Heqmaybe_q1.
    pose proof size_remove Heqmaybe_q1 as SizeEq1.
      rewrite SizeEq1. simpl. f_equal.
    pose proof (add_1 B q1 (VarPair.eq_refl (t1, r1))) as Maps2.
    pose proof size_remove Maps2 as SizeEq2.
    pose proof MapsTo_add_remove q1 Heqmaybe_q1 as Eq3.
    pose proof size_Equal Eq3 as SizeEq3.
      rewrite SizeEq3.
    symmetry. apply SizeEq2.

    (* buffer does not contain a queue for this pair *)
    inversion Eq.
  Qed.

  (* The next few lemmas are used in the empty_size proof. *)
  Lemma is_empty_Equal : forall B1 B2,
    is_empty B1 ->
    Equal B1 B2 ->
    is_empty B2.
  Proof.
    intros B1 B2 Emp Eq.
    unfold is_empty, is_empty_queue, Equal in *.
    intros t1 r1.
    specializes* Eq (>>> (t1, r1)).
    rewrite <- Eq. apply Emp.
  Qed.

  Lemma is_empty_add_nil : forall B t r q,
    is_empty (add (t, r) q B) ->
    q = nil.
  Proof.
    intros B1 t1 r1 q1 Emp.
    unfold is_empty, is_empty_queue in Emp.
    forwards Emp1 : Emp (>>> t1 r1).
    rewrite* MapFacts.add_eq_o in Emp1.
  Qed.

  Lemma is_empty_after_add : forall B t r,
    is_empty B ->
    is_empty (add (t, r) nil B).
  Proof.
    intros B1 t1 r1 Emp t2 r2.
    unfold is_empty, is_empty_queue in *.
    destruct ((t1, r1) ==== (t2, r2)) as [ [ ? ? ] | ].
    simpl in *. subst.
    rewrite* MapFacts.add_eq_o.
    rewrite* MapFacts.add_neq_o.
    apply Emp.
  Qed.

  Lemma is_empty_before_add : forall B t r q,
    is_empty (add (t, r) q B) ->
    ~ In (t, r) B ->
    is_empty B.
  Proof.
    intros B1 t1 r1 q1 Emp NotIn.
    unfold is_empty, is_empty_queue in *.
    intros t2 r2.
    remember (find (t2, r2) B1) as maybe_q2.
    symmetry in Heqmaybe_q2.
    destruct* maybe_q2 as [ q2 | ].
    destruct ((t1, r1) ==== (t2, r2)) as [ [ ? ? ] | ].
    simpl in *. subst t1 r1.
    rewrite <- MapFacts.find_mapsto_iff in Heqmaybe_q2.
    unfold In in NotIn. false NotIn. eauto.
    specializes Emp (>>> t2 r2).
    rewrite* MapFacts.add_neq_o in Emp.
    rewrite* Heqmaybe_q2 in Emp.
  Qed.

  (* The empty buffer has size 0. (Tricky because the buffer may have nil
     queues.) *)
  Lemma empty_size : forall B,
    is_empty B <-> size B = 0.
  Proof.
    apply MapProperties.map_induction_bis.

    (* holds over Equal *)
    intros B1 B2 Eq empty_size_B1.
    split.
    intros Emp2.
      pose proof is_empty_Equal Emp2 (MapFacts.Equal_sym Eq) as Emp1.
      rewrite empty_size_B1 in Emp1.
      rewrite <- Emp1. symmetry.
      apply* size_Equal. auto.
    intros Size2.
      apply* is_empty_Equal.
      rewrite empty_size_B1.
      rewrite <- Size2.
      apply* size_Equal. auto. eauto.

    (* holds for the empty map *)
    split; intro.
    unfold size. apply eq_nat_eq.
      apply MapProperties.fold_Empty.
      apply Equivalence_eq_nat.
      apply empty_1.
    intros t1 r1.
      unfold is_empty_queue.
      rewrite* MapFacts.empty_o.

    (* holds for add *)
    intros [ t1 r1 ] q1 B1 NotIn IH.
    rewrite* size_add.
    split.
    intros Emp.
      forwards* Eq : is_empty_add_nil.
      subst. simpl. rewrite <- IH.
      apply* is_empty_before_add.
    intros Size1.
      apply plus_is_O in Size1.
      destruct Size1 as [ length_q1 Size1 ].
      destruct q1.
      apply* is_empty_after_add.
      simpl in length_q1.
      inversion length_q1.
  Qed.

  Lemma dequeue_enqueue_empty : forall B t r v,
    is_empty B ->
    exists B', dequeue (enqueue B t r v) t r = Some (v, B')
      /\ is_empty B'.
  Proof.
    intros B1 t1 r1 v1 Emp.
    unfold is_empty, is_empty_queue, dequeue, enqueue in *.
    forwards Emp1 : Emp (>>> t1 r1).
    destruct (find (t1, r1) B1) as [ [ | ] | ].
    simpl. esplit.
      rewrite* MapFacts.add_eq_o.
      split*. intros t2 r2.
      destruct ((t1, r1) ==== (t2, r2)).
      rewrite* MapFacts.add_eq_o.
      rewrite* MapFacts.add_neq_o.
      rewrite* MapFacts.add_neq_o.
      apply Emp.
    inversion Emp1.
    simpl. esplit.
      rewrite* MapFacts.add_eq_o.
      split*. intros t2 r2.
      destruct ((t1, r1) ==== (t2, r2)).
      rewrite* MapFacts.add_eq_o.
      rewrite* MapFacts.add_neq_o.
      rewrite* MapFacts.add_neq_o.
      apply Emp.
  Qed.

  Lemma is_empty_queue_dec : forall B t r,
    decidable (is_empty_queue B t r).
  Proof.
    intros B1 t1 r1.
    unfold is_empty_queue, decidable.
    destruct* (find (t1, r1) B1) as [ q | ].
    destruct* q.
    right. intro false. inversion false.
  Qed.

  Lemma for_all_empty : forall (B : buffer) t r P,
    is_empty_queue B t r ->
    for_all_queue B P t r.
  Proof.
    unfold is_empty_queue, for_all_queue.
    intros B1 t1 r1 P Emp.
    destruct (find (t1, r1) B1).
    subst. intros v In_nil.
    pose proof List.in_nil In_nil as false.
    false. auto.
  Qed.

  Lemma for_all_generalize : forall B t r (P Q : L.val -> Prop),
    for_all_queue B P t r ->
    (forall v, P v -> Q v) ->
    for_all_queue B Q t r.
  Proof.
    intros B t1 r1 P Q All Gen.
    unfold for_all_queue in *.
    destruct* (find (t1, r1) B).
  Qed.

  Lemma for_all_enqueue_diff : forall B P t1 r1 t2 r2 v2,
    for_all_queue B P t1 r1 ->
    ~ VarPair.eq (t1, r1) (t2, r2) ->
    for_all_queue (enqueue B t2 r2 v2) P t1 r1.
  Proof.
    intros B P t1 r1 t2 r2 v2 All1 Neq12.
    unfold for_all_queue, enqueue in *.
    rewrite* MapFacts.add_neq_o. simpl.
    intros [ ? ? ]. subst. apply Neq12. apply VarPair.eq_refl.
  Qed.

  Lemma for_all_enqueue_same : forall B P t1 r1 v1,
    for_all_queue B P t1 r1 ->
    P v1 ->
    for_all_queue (enqueue B t1 r1 v1) P t1 r1.
  Proof.
    intros B P t1 r1 v1 All1 P_v1.
    unfold for_all_queue, enqueue in *.
    rewrite* MapFacts.add_eq_o.
    intros v2 In2.
    destruct (find (t1, r1) B) as [ q1 | ].
    apply List.in_app_or in In2.
    simpl in In2. destruct* In2 as [ | [ | ] ].
    subst. auto.
    simpl in In2. destruct* In2.
    subst. auto.
  Qed.

  Lemma for_all_dequeue_val : forall B1 P t1 r1 v1 B2,
    for_all_queue B1 P t1 r1 ->
    dequeue B1 t1 r1 = Some (v1, B2) ->
    P v1.
  Proof.
    intros B1 P t1 r1 v1 B2 All1 Deq.
    forwards* [ ? [ Find ? ] ] : dequeue_inversion.
    unfold for_all_queue in All1.
    rewrite Find in All1.
    apply All1. apply List.in_eq.
  Qed.

  Lemma for_all_dequeue : forall B1 P t1 r1 t2 r2 v2 B2,
    for_all_queue B1 P t1 r1 ->
    dequeue B1 t2 r2 = Some (v2, B2) ->
    for_all_queue B2 P t1 r1.
  Proof.
    intros B1 P t1 r1 t2 r2 v2 B2 All1 Deq.
    forwards* [ ? [ Find ? ] ] : dequeue_inversion. subst B2.
    unfold for_all_queue in *.
    destruct ((t1, r1) ==== (t2, r2)) as [ [ ? ? ] | Neq ].
    simpl in *. subst.
      rewrite* MapFacts.add_eq_o.
      rewrite Find in All1.
      intros v3 In3. apply All1.
      simpl. auto.
    rewrite* MapFacts.add_neq_o.
      intro. apply Neq. apply VarPair.eq_sym. auto.
  Qed.

  Lemma for_all_top : forall B P t r v,
    for_all_queue B P t r ->
    top B t r = Some v ->
    P v.
  Proof with auto.
    intros B P t' r v ForAll Top.
    unfold top, for_all_queue in *.
    destruct (find (t', r) B) as [ q | ].
    induction q as [ | v' q ];
      simpl in Top.

    inversion Top.

    destruct q.

    inversion Top.
    subst.
    apply ForAll.
    unfold List.In.
    auto.

    apply IHq.
    intros v'' InQueue.
    apply ForAll.
    right. auto.
    simpl in *.
    apply Top.

    inversion Top.
  Qed.    

  Lemma empty_top : forall B t r,
    is_empty_queue B t r ->
    top B t r = None.
  Proof.
    intros B t' r Emp.
    unfold is_empty_queue, top in *.
    destruct (find (t', r) B).
    subst. auto.
    auto.
  Qed.

  Lemma top_exists : forall (A : Set) (l : list A),
    match l with
      | nil => top_queue l = None
      | a :: l' => exists a', top_queue l = Some a'
    end.
  Proof.
    intros A l.
    induction l as [ | a l ].
    auto.
    destruct l.
    simpl. eauto.
    destruct IHl.
    simpl in *. eauto.
  Qed.

  Lemma top_dequeue : forall B t r v v' B',
    top B t r = Some v ->
    dequeue B t r = Some (v', B') ->
    (v = v' /\ is_empty_queue B' t r) \/ (top B' t r = Some v).
  Proof with eauto.
    intros B t' r v v' B' top_eq deq_eq.
    unfold top, dequeue, is_empty_queue in *.
    destruct (find (t', r) B) as [ q | ].
    destruct q as [ | v'' q ];
      inversion deq_eq.
    subst.
    destruct q as [ | v''' q ].
    simpl in top_eq.
    inversion top_eq; subst.
    left. split...
    rewrite MapFacts.add_eq_o...
    rewrite MapFacts.add_eq_o...
    inversion top_eq.
  Qed.

  Lemma top_dequeue_diff : forall B t r opt t' r' v' B',
    top B t r = opt ->
    dequeue B t' r' = Some (v', B') ->
    ~ VarPair.eq (t, r) (t', r') ->
    top B' t r = opt.
  Proof with eauto.
    intros B t' r opt t'' r' v' B' top_eq deq_eq Neq.
    unfold top, dequeue in *.
    destruct (find (t'', r') B) as [ [ | v'' q ] | ];
      inversion deq_eq.
    subst.
    rewrite MapFacts.add_neq_o...
    simpl. intros [ ? ? ]. subst.
    apply Neq. apply VarPair.eq_refl.
  Qed.

  Lemma top_not_empty : forall B t r v,
    top B t r = Some v ->
    ~ is_empty_queue B t r.
  Proof.
    intros B t' r v top_eq empty_q.
    unfold top, is_empty_queue in *.
    destruct (find (t', r) B).
    subst. simpl in top_eq. inversion top_eq.
    inversion top_eq.
  Qed.
End BufferFMap.
