Require Import init_wb.

Set Implicit Arguments.

Open Scope nat_scope.

Import WB.

(* Buffer B is empty for location r, except that thread t may have some values
   buffered. *)
Definition empty_except (B : Buffer.buffer) (t : tid) (r : loc) : Prop :=
  forall t', t <> t' -> Buffer.is_empty_queue B t' r.

(* A heap is coherent if at most one thread has values enqueued for each
   location. *)
Definition coherent_heap (H : heap) :=
  forall t r, ~ Buffer.is_empty_queue (snd (fst H)) t r ->
    empty_except (snd (fst H)) t r.

Hint Unfold coherent_heap empty_except.

(* The empty heap is coherent. *)
Lemma coherent_empty :
  coherent_heap WB.empty_heap.
Proof.
  unfold empty_heap, coherent_heap, empty_except.
  simpl. intros.
  pose proof Buffer.empty_is_empty as is_empty.
  unfold Buffer.is_empty in is_empty.
  auto.
Qed.

(* The lock map has no effect on coherency. *)
Lemma coherent_independent_of_lock_map : forall S B Lambda Lambda',
  coherent_heap (S, B, Lambda) ->
  coherent_heap (S, B, Lambda').
Proof.
  intros. unfold coherent_heap in *. auto.
Qed.

(* Taking an empty step preserves coherency. *)
Lemma empty_step_preserves_coherent : forall H_wb H_wb',
  coherent_heap H_wb ->
  WB.heap_step H_wb None H_wb' ->
  coherent_heap H_wb'.
Proof.
  intros H_wb H_wb' coherent steps.
  destruct H_wb as [ [ S B ] Lambda ].
  destruct H_wb' as [ [ S' B' ] Lambda' ].
  inversion steps as [ | | | ? ? ? t r v ? B'_eq | | | | ]; subst.
  unfold coherent_heap in *.
  intros t' r' not_empty t'' t'_neq_t''. simpl in *.

  destruct (eq_var_dec r r'); subst.

  (* Case: r = r' *)
  destruct (eq_var_dec t t'); subst.

  (* Subcase: r = r' and t = t' *)
  eapply Buffer.is_empty_queue_dequeue; eauto.
  eapply coherent; eauto.
  eapply Buffer.dequeue_not_is_empty_queue; eauto.

  (* Subcase: r = r' and t <> t' *)
  contradiction not_empty.
  eapply Buffer.is_empty_queue_dequeue; eauto.
  eapply coherent; eauto.
  eapply Buffer.dequeue_not_is_empty_queue; eauto.

  (* Case: r <> r' *)
  eapply Buffer.is_empty_queue_dequeue; eauto.
  eapply coherent; eauto. simpl in *.
  rewrite <- Buffer.is_empty_queue_dequeue_diff; eauto.
  intros [ ? ? ]. simpl in *. subst. tauto.
Qed.

(* Taking many empty steps preserves coherency. *)
Lemma empty_step_many_preserves_coherent : forall H1 H2 tr,
  coherent_heap H1 ->
  heap_step_many H1 H2 tr ->
  erase_trace tr = nil ->
  coherent_heap H2.
Proof.
  intros H1 H2 tr coherent_H1 steps_many tr_nil.
  induction steps_many as [ | H1 H2 H3 tr e steps_many ? steps ].
  auto.
  destruct e as [ [ ] | ]; simpl in tr_nil.
  inversion tr_nil.
  eapply empty_step_preserves_coherent.
  apply IHsteps_many. auto. auto. auto.
Qed.

(* An alloc step preserves coherency. *)
Lemma alloc_preserves_coherent : forall H1 H2 t r v,
  coherent_heap H1 ->
  heap_step H1 (Some (t, alloc r v)) H2 ->
  coherent_heap H2.
Proof.
  intros H1 H2 t r v coherent steps.
  inversion steps; subst.
  unfold coherent_heap in *.
  auto.
Qed.

(* A write step preserved coherency (assuming the thread doing the write is the
   only thread with any values buffered). *)
Lemma write_preserves_coherent : forall H1 H2 t r v,
  coherent_heap H1 ->
  empty_except (snd (fst H1)) t r ->
  heap_step H1 (Some (t, write r v)) H2 ->
  coherent_heap H2.
Proof.
  intros H1 H2 t r v coherent except steps.
  inversion steps; subst. simpl in *.
  unfold empty_except in except.
  intros t' r' not_empty_t' t'' t'_neq_t''.
  simpl in *.
  destruct (eq_var_dec r r').
  subst r'.
  destruct (eq_var_dec t t'').
  subst t''.
  contradiction not_empty_t'.
  rewrite <- Buffer.is_empty_queue_enqueue_diff; eauto.
  intros [ ? ? ]. simpl in *. subst. tauto.
  rewrite <- Buffer.is_empty_queue_enqueue_diff; eauto.
  intros [ ? ? ]. simpl in *. subst. tauto.
  rewrite <- Buffer.is_empty_queue_enqueue_diff; eauto.
  apply coherent with (t := t'); auto.
  intro false. apply not_empty_t'.
  rewrite <- Buffer.is_empty_queue_enqueue_diff; auto.
  intros [ ? ? ]. simpl in *. subst. tauto.
  intros [ ? ? ]. simpl in *. subst. tauto.
Qed.
