Require Import Metatheory.

Require Import Arith.
Require Import List.

Require Import actions.
Require Import events.
Require Import indexes.

Set Implicit Arguments.

Section Traces.
  Variable val : Set.

  Inductive equivalence : erased_trace val -> erased_trace val -> Prop :=
  | equivalence_flip : forall e1 e2,
    ~ effects_conflict e1 e2 ->
    equivalence (e1 :: e2 :: nil) (e2 :: e1 :: nil)
  | equivalence_append : forall tr1 tr2 tr1' tr2',
    equivalence tr1 tr1' ->
    equivalence tr2 tr2' ->
    equivalence (tr1 ++ tr2) (tr1' ++ tr2')
  | equivalence_refl : forall tr,
    equivalence tr tr
  | equivalence_trans : forall tr1 tr2 tr3,
    equivalence tr1 tr2 ->
    equivalence tr2 tr3 ->
    equivalence tr1 tr3.

  Hint Constructors equivalence.

  (* The equivalence relation is symmetric. *)
  Lemma equivalence_sym : forall tr1 tr2,
    equivalence tr1 tr2 ->
    equivalence tr2 tr1.
  Proof.
    intros tr1 tr2 equiv.
    induction equiv;
      eauto 6 using effects_conflict_sym.
  Qed.

  (* Equivalent traces are of the same length. *)
  Lemma equivalence_length : forall tr1 tr2,
    equivalence tr1 tr2 ->
    length tr1 = length tr2.
  Proof with auto.
    intros tr1 tr2 equiv.
    induction equiv...
    repeat rewrite app_length...
    rewrite IHequiv1. rewrite IHequiv2...
  Qed.

  (* Equivalent traces have identical projections. *)
  Lemma equivalence_projection :
    forall tr1 tr2 t,
      equivalence tr1 tr2 ->
      trace_projection tr1 t = trace_projection tr2 t.
  Proof with auto.
    intros tr1 tr2 t equiv.
    induction equiv as [ e1 e2 not_conflict concurrent | | | ]...

    (* Two concurrent, non-conflicting actions. *)
    simpl.
    destruct e1 as [ t1 a1 ]. destruct e2 as [ t2 a2 ].
    destruct (eq_var_dec t t1); destruct (eq_var_dec t t2)...
    subst t1 t2.
    contradiction not_conflict.
    apply effects_conflict_program_order.

    (* Append case. *)
    repeat rewrite app_projection.
    rewrite IHequiv1. rewrite IHequiv2...

    (* Transitive case *)
    rewrite IHequiv1. rewrite IHequiv2...
  Qed.

  (* If a event index (t, i) is valid for tr, then it is also valid for any
     tr' that is equivalent to tr. *)
  Lemma equivalence_valid : forall tr t i tr',
    event_valid tr t i ->
    equivalence tr tr' ->
    event_valid tr' t i.
  Proof with auto.
    intros tr t i tr' valid equiv.
    rewrite (valid_iff_lt_length_projection tr' t i).
    rewrite <- (equivalence_projection t equiv).
    rewrite <- (valid_iff_lt_length_projection tr t i)...
  Qed.

  (* If action a appears at event index (t, i) in tr, and tr' is equivalent
     to tr, then action a also appears at event index (t, i) in tr'. *)
  Lemma equivalence_event : forall tr1 tr2 t i j1 j2 a,
    event tr1 t i j1 ->
    event tr2 t i j2 ->
    index tr1 j1 (t, a) ->
    equivalence tr1 tr2 ->
    index tr2 j2 (t, a).
  Proof.
    intros tr1 tr2 t i j1 j2 a evt1 evt2 idx1 equiv.
    pose proof event_inversion evt1 as [ a1 [ idx1' idx_proj1 ] ].
    pose proof event_inversion evt2 as [ a2 [ idx2 idx_proj2 ] ].
    rewrite <- (equivalence_projection t equiv) in idx_proj2.
    pose proof index_unique idx1 idx1' as eq1. inversion eq1; subst.
    pose proof index_unique idx_proj1 idx_proj2 as eq2. inversion eq2; subst.
    auto.
  Qed.

  Lemma equivalence_append_flip : forall tr1 tr2 e1 e2,
    ~ effects_conflict e1 e2 ->
    equivalence (tr2 ++ (e2 :: nil) ++ (e1 :: nil) ++ tr1)
    (tr2 ++ (e1 :: nil) ++ (e2 :: nil) ++ tr1).
  Proof.
    intros tr1 tr2 e1 e2 not_conflict.
    apply equivalence_sym.
    rewrite <- (app_ass _ _ tr1).
    rewrite <- (app_ass _ (e1 :: nil) tr1).
    apply equivalence_append; auto.
    apply equivalence_append; simpl; auto.
  Qed.

  (* If tr1 and tr1' are equivalent, and event index t i occrs in tr1, and event
     index t i occurs at j' in tr1' ++ tr2', then event index t i occurs at j' in
     tr1'. *)
  Lemma equivalence_append_1 : forall tr1 tr1' tr2' t i j j',
    event tr1 t i j ->
    equivalence tr1 tr1' ->
    event (tr1' ++ tr2') t i j' ->
    j' < length tr1' /\ event tr1' t i j'.
  Proof with eauto.
    intros tr1 tr1' tr2' t i j j' evt equiv evt'.

    pose proof event_inversion evt as [ a [ index index_proj_tr1 ] ].
    rewrite equivalence_projection with (tr2 := tr1') in index_proj_tr1...
    pose proof index_lt_length index_proj_tr1 as i_lt.
    rewrite <- valid_iff_lt_length_projection in i_lt.
    destruct i_lt as [ j'' evt''_tr1' ].
    pose proof event_append_1 tr2' evt''_tr1' as evt''.
    pose proof event_unique_index evt' evt''.
    subst. split...
    eapply event_lt_length...
  Qed.

  (* If tr1 and tr1' are equivalent, and event index t i occrs in tr2, and event
     index t i occurs at j' in tr1' ++ tr2', then event index t i falls in tr2' *)
  Lemma equivalence_append_2 : forall tr1 tr2 tr1' tr2' t i j j',
    event (tr1 ++ tr2) t i (j + length tr1) ->
    equivalence tr1 tr1' ->
    event (tr1' ++ tr2') t i j' ->
    exists i'', exists j'',
      j' = j'' + length tr1' /\
      i = i'' + length (trace_projection tr1' t) /\
      event tr2' t i'' j''.
  Proof with eauto.
    intros tr1 tr2 tr1' tr2' t i j j' evt equiv1 evt'.
    pose proof event_append_inversion _ _ evt'
      as [ [ j'_lt evt'_tr1' ]
        | [ i'' [ j'' [ j''_eq [ i''_eq evt'' ] ] ] ] ]...

    (* Vacuous, but finding a contradicition is a little tricky. *)
    assert (event tr1 t i (j + length tr1)) as evt_tr1.
    eapply proj2. eapply equivalence_append_1; eauto.
    apply equivalence_sym. auto.
    pose proof (event_inversion evt_tr1) as [ a [ index_tr1 ? ] ].
    assert (j + length tr1 < length tr1). eapply index_lt_length. eauto.
    assert (j + length tr1 >= length tr1) as false. auto with arith.
    unfold ge in false. apply le_not_lt in false. tauto.
  Qed.
End Traces.