(** We define the happens-before relation over program traces, and prove that
   DRF programs produce traces consistent with happens-before when running on
   a sequentially-consistent system. *)

Require Import init_wb.
Require Import modulo_values.
Require Import preservation.
Require Import sc_facts.
Require Import sc_type_preservation.

Import L.
Import SC.
Import L_SC.
Import DRF.
Import SCFacts.
Import Typ.
Import Pres.
Import L_Axioms.
Import SC_Type_Preservation.

Module SysPres := System_Model_Type_Preservation Pres SC_Type_Preservation L_SC.
Import SysPres.

Set Implicit Arguments.

Open Scope nat_scope.

Hint Constructors system_step system_step_many.

(** The happens-before relation is an ordering on actions in traces. It is the
   union of program order (actions performed by the same thread) and the
   synchronization order (between relases and acquires of the same lock). *)
Inductive happens_before (tr : erased_trace L.val) : Relation nat :=
| happens_before_program_order : forall t a1 j1 a2 j2,
  index tr j1 (t, a1) ->
  index tr j2 (t, a2) ->
  j2 < j1 ->
  happens_before tr j1 j2
| happens_before_sync_order : forall t1 j1 t2 j2 l,
  index tr j1 (t1, release _ l) ->
  index tr j2 (t2, acquire _ l) ->
  j2 < j1 ->
  happens_before tr j1 j2
| happens_before_spawn_order : forall t1 j1 t2 j2 a2,
  index tr j1 (t1, spawn _ t2) ->
  index tr j2 (t2, a2) ->
  j2 < j1 ->
  happens_before tr j1 j2.

(** Happens-before-plus is the transitive closure of the happens-before
   relation. (We'll refer to happens-before-plus as happens-before.) Note that
   happens-before-plus is irreflexive and asymmetric. *)
Definition happens_before_plus tr := Rplus _ (happens_before tr).

(** A program is well-synchronized if all data races are synchronized by
   happens-before. *)
Definition well_synchronized (tr : erased_trace L.val) :=
  forall t1 a1 j1 t2 a2 j2,
    index tr j1 (t1, a1) ->
    index tr j2 (t2, a2) ->
    data_race a1 a2 ->
    j2 < j1 ->
    happens_before_plus tr j1 j2.

(** The happens-before relation is oblivious to the values recorded by alloc,
   read and write actions. *)
Lemma equal_modulo_values_happens_before : forall tr1 tr2 j1 j2,
  happens_before tr1 j1 j2 ->
  equal_modulo_values tr1 tr2 ->
  happens_before tr2 j1 j2.
Proof.
  intros tr1 tr2 j1 j2 hb1 mod.
  inversion hb1 as [ ? ? ? ? ? idx1 idx2 | ? ? ? ? ? idx1 idx2
    | ? ? ? ? ? idx1 idx2 ]; subst;
    pose proof equal_modulo_values_index idx1 mod as [ ? [ ? actions_mod1 ] ];
      pose proof equal_modulo_values_index idx2 mod as [ ? [ ? actions_mod2 ] ].
  eapply happens_before_program_order; eauto.
  inversion actions_mod1. inversion actions_mod2. subst.
  eapply happens_before_sync_order; eauto.
  inversion actions_mod1. subst.
  eapply happens_before_spawn_order; eauto.
Qed.

Lemma equal_modulo_values_happens_before_plus : forall tr1 tr2 j1 j2,
  happens_before_plus tr1 j1 j2 ->
  equal_modulo_values tr1 tr2 ->
  happens_before_plus tr2 j1 j2.
Proof.
  intros tr1 tr2 j1 j2 plus mod.
  induction plus.
  apply Rplus_0. eapply equal_modulo_values_happens_before; eauto.
  eapply Rplus_n. eapply equal_modulo_values_happens_before; eauto. auto.
Qed.

Lemma equal_modulo_values_well_synchronized : forall tr1 tr2,
  well_synchronized tr1 ->
  equal_modulo_values tr1 tr2 ->
  well_synchronized tr2.
Proof.
  intros tr1 tr2 sync mod.
  unfold well_synchronized in *.
  intros t1 a1 j1 t2 a2 j2 idx1 idx2 race12 j2_lt.
  apply equal_modulo_values_happens_before_plus with (tr1 := tr1); auto.
  pose proof equal_modulo_values_sym mod as mod'.
  pose proof equal_modulo_values_index idx1 mod' as [ a1' [ idx1' mod1' ] ].
  pose proof equal_modulo_values_index idx2 mod' as [ a2' [ idx2' mod2' ] ].
  eapply sync; eauto.
  eapply actions_equal_modulo_values_race; eauto.
Qed.

Lemma happens_before_plus_implies_lt : forall tr j1 j2,
  happens_before_plus tr j1 j2 ->
  j2 < j1.
Proof.
  intros tr j1 j2 hb.
  induction hb as [ ? ? hb | ? ? ? hb ].
  inversion hb; auto.
  inversion hb; omega.
Qed.

(** The inherently-precedes relation contains the happens-before relation. *)
Lemma happens_before_implies_inherently_precedes :
  forall tr t1 i1 j1 t2 i2 j2,
    happens_before tr j1 j2 ->
    event tr t1 i1 j1 ->
    event tr t2 i2 j2 ->
    inherently_precedes tr t1 i1 t2 i2.
Proof.
  intros tr t1 i1 j1 t2 i2 j2 hb evt1 evt2.
  pose proof event_index evt1 as [ a1 idx1 ].
  pose proof event_index evt2 as [ a2 idx2 ].
  eapply conflict_inherently_precedes; eauto.
  inversion hb; auto.
  inversion hb as [ ? ? ? ? ? idx1' idx2' | ? ? ? ? ? idx1' idx2'
    | ? ? ? ? ? idx1' idx2' ];
  pose proof index_unique idx1 idx1' as eq1; inversion eq1;
    pose proof index_unique idx2 idx2' as eq2; inversion eq2;
      subst.
  apply effects_conflict_program_order.
  apply effects_conflict_actions. auto.
  apply effects_conflict_spawn_left.
Qed.

(** Happens-before translates to equivalent traces. *)
Lemma equivalence_happens_before : forall tr tr' t1 i1 j1 j1' t2 i2 j2 j2',
  happens_before tr j1 j2 ->
  event tr t1 i1 j1 ->
  event tr t2 i2 j2 ->
  equivalence tr tr' ->
  event tr' t1 i1 j1' ->
  event tr' t2 i2 j2' ->
  happens_before tr' j1' j2'.
Proof.
  intros tr tr' t1 i1 j1 j1' t2 i2 j2 j2' hb evt1 evt2 equiv evt1' evt2'.
  pose proof event_index evt1 as [ a1 idx1 ].
  pose proof event_index evt2 as [ a2 idx2 ].
  pose proof event_index evt1' as [ a1' idx1' ].
  pose proof event_index evt2' as [ a2' idx2' ].
  pose proof happens_before_implies_inherently_precedes hb evt1 evt2 as
    [ ? [ ? precedes ] ].
  assert (j2' < j1'). apply precedes with (tr' := tr'); auto.
  inversion hb as [ ? ? ? ? ? idx1_ idx2_ | ? ? ? ? ? idx1_ idx2_
    | ? ? ? ? ? idx1_ idx2_ ]; subst;
    pose proof index_unique idx1 idx1_ as eq1; inversion eq1; subst;
      pose proof index_unique idx2 idx2_ as eq2; inversion eq2; subst.
  eapply happens_before_program_order; eauto.
  pose proof equivalence_event evt1 evt1' idx1 equiv as idx1'_.
  pose proof index_unique idx1' idx1'_ as eq1'; inversion eq1'; subst.
  pose proof equivalence_event evt2 evt2' idx2 equiv as idx2'_.
  pose proof index_unique idx2' idx2'_ as eq2'; inversion eq2'; subst.
  eapply happens_before_sync_order; eauto.
  pose proof equivalence_event evt1 evt1' idx1 equiv as idx1'_.
  pose proof index_unique idx1' idx1'_ as eq1'; inversion eq1'; subst.
  pose proof equivalence_event evt2 evt2' idx2 equiv as idx2'_.
  pose proof index_unique idx2' idx2'_ as eq2'; inversion eq2'; subst.
  eapply happens_before_spawn_order; eauto.
Qed.

Lemma equivalence_happens_before_plus : forall tr tr' t1 i1 j1 j1' t2 i2 j2 j2',
  happens_before_plus tr j1 j2 ->
  event tr t1 i1 j1 ->
  event tr t2 i2 j2 ->
  equivalence tr tr' ->
  event tr' t1 i1 j1' ->
  event tr' t2 i2 j2' ->
  happens_before_plus tr' j1' j2'.
Proof.
  intros tr tr' t1 i1 j1 j1' t2 i2 j2 j2' plus evt1 evt2 equiv evt1' evt2'.
  generalize dependent j1'.
  generalize dependent i1.
  generalize dependent t1.
  induction plus as [ j1 j2 hb | j1 j3 j2 hb ]; intros t1 i1 evt1 j1' evt1'.

  apply Rplus_0. eapply equivalence_happens_before; eauto.

  assert (exists t3, exists a3, index tr j3 (t3, a3)) as [ t3 [ a3 idx3 ] ].
  inversion hb; eauto.
  pose proof index_event idx3 as [ i3 evt3 ].
  assert (event_valid tr' t3 i3) as [ j3' evt3' ].
  eapply equivalence_valid; eauto.
  apply Rplus_n with (y := j3').
  eapply equivalence_happens_before; eauto.
  eapply IHplus; eauto.
Qed.

(** The base case for the next lemma. If two adjacent actions conflict (and are
   performed by concurrent threads), then they are ordered by happens-before. *)
Lemma drf_conflict_adjacent_implies_happens_before :
  forall Sigma1 H1 P1 H2 P2 tr e1 e2 j,
    well_typed Sigma1 H1 ->
    typing Sigma1 P1 ->
    race_free_state H1 P1 ->
    system_step_many H1 P1 H2 P2 (unerase_trace tr) ->
    index tr (j + 1) e1 ->
    index tr j e2 ->
    effects_conflict e1 e2 ->
    happens_before tr (j + 1) j.
Proof.
  intros Sigma1 H1 P1 H2 P2 tr e1 e2 j typing_H1 typing_P1 drf steps_many idx1
    idx2 conflict.
  assert (j < j + 1) as j_lt. omega.
  destruct e1 as [ t1 a1 ]. destruct e2 as [ t2 a2 ].
  destruct (eq_var_dec t1 t2) as [ | t1_neq_t2 ].
  subst. eapply happens_before_program_order; eauto.

  (* Invert the step_many relation *)
  pose proof index_unerase idx1 as idx1'.
  pose proof index_unerase idx2 as idx2'.
  pose proof SCFacts.system_step_many_index_inversion_adjacent steps_many idx1'
    idx2' as [ H3 [ P3 [ H4 [ P4 [ H5 [ P6 [ tr1 [ tr2 [ steps_many1 [ steps1 [
      steps2 [ steps_many2 tr_eq ] ] ] ] ] ] ] ] ] ] ] ].

  (* The state just before e1 is race-free. *)
  assert (no_imminent_races H3 P3) as no_races. eauto.
  unfold no_imminent_races in no_races.
  pose proof no_races _ _ _ _ _ _ _ _ _ _ steps1 steps2 as not_race.

  (* The heap and program just before e1 typecheck. *)
  pose proof preservation_many typing_H1 typing_P1 steps_many1
    as [ Sigma3 [ ext [ typing_H3 typing_P3 ] ] ].

  (* Destruct the heaps into their component pieces. *)
  destruct H1 as [ S1 Lambda1 ].
  destruct H2 as [ S2 Lambda2 ].
  destruct H3 as [ S3 Lambda3 ].

  (* What type of conflict is it? *)
  inversion conflict as [ ? ? ? ? conflict_actions | | | ]; subst.
  inversion conflict_actions as [
    r v1 v2
    | r v1 v2
    | r v1 v2
    | r v1 v2
    | r v1 v2
    | r v1 v2
    | r v1 v2
    | r v1 v2
    | l
    | l
    | l
    | l
  ]; subst a1 a2;
  inversion steps1 as [ ? ? ? ? ? ? program_steps1 heap_steps1 | | ]; subst;
    inversion steps2 as [ ? ? ? ? ? ? program_steps2 heap_steps2 | | ]; subst.

  (* alloc, alloc (violates heap semantics) *)
  inversion heap_steps1; subst;
    inversion heap_steps2 as [ ? ? ? ? ? fresh_S3 | | | | | ]; subst.
  pose proof SC.Store.not_fresh_after_update fresh_S3 as false.
  inversion false.

  (* alloc, read (violates program semantics) *)
  forwards [ ? [ program_steps1' ? ] ] : program_asynchrony program_steps1.
  apply program_steps2.
  intros prgm_conflict.
  inversion prgm_conflict.
  tauto.
  forwards : well_formed_read (>>> typing_P3 program_steps1').
  forwards typing_inc : typing_step (>>> typing_H3 heap_steps1).
  inversion typing_inc; subst.
  tauto.

  (* read, alloc (violates heap semantics) *)
  inversion heap_steps1 as [ | ? ? ? ? ? get_S1 | | | | ]; subst;
    inversion heap_steps2 as [ ? ? ? ? ? fresh_S2 | | | | | ]; subst.
  pose proof SC.Store.not_fresh_after_get get_S1 fresh_S2 as false.
  inversion false.

  (* alloc, write (violates program semantics) *)
  forwards [ ? [ program_steps1' ? ] ] : program_asynchrony program_steps1.
  apply program_steps2.
  intros prgm_conflict.
  inversion prgm_conflict.
  tauto.
  forwards : well_formed_write (>>> typing_P3 program_steps1').
  forwards typing_inc : typing_step (>>> typing_H3 heap_steps1).
  inversion typing_inc; subst.
  tauto.

  (* write, alloc (violates heap semantics) *)
  inversion heap_steps1; subst;
    inversion heap_steps2 as [ ? ? ? ? ? fresh_S2 | | | | | ]; subst.
  pose proof SC.Store.not_fresh_after_update fresh_S2 as false.
  inversion false.

  (* write, write (violates data-race freedom) *)
  destruct not_race as [ false | ].
  contradiction false. apply data_race_ww. tauto.

  (* read, write (violates data-race freedom) *)
  destruct not_race as [ false | ].
  contradiction false. apply data_race_rw. tauto.

  (* write, read (violates data-race freedom) *)
  destruct not_race as [ false | ].
  contradiction false. apply data_race_wr. tauto.

  (* acquire, acquire (violates lock semantics) *)
  inversion heap_steps1; subst;
    inversion heap_steps2 as [ | | | ? ? ? ? avail | | ]; subst.
  pose proof LockMap.not_available_after_acquire avail as false.
  inversion false.

  (* acquire, release (violates lock semantics) *)
  inversion heap_steps1; subst;
    inversion heap_steps2 as [ | | | | ? ? ? ? held | ]; subst.
  pose proof LockMap.not_held_by_after_acquire t1_neq_t2 held as false.
  inversion false.

  (* release, acquire (one legal scenario) *)
  eapply happens_before_sync_order; eauto.

  (* release, release (violates lock semantics) *)
  inversion heap_steps1; subst;
    inversion heap_steps2 as [ | | | | ? ? ? ? held | ]; subst.
  pose proof LockMap.not_held_by_after_release held as false.
  inversion false.

  (* program order (vacuous) *)
  tauto.

  (* thread spawn (the other legal scenario) *)
  eapply happens_before_spawn_order; eauto.

  (* backwards thread spawn (violates program semantics) *)
  assert (t1 <> t1).
  eapply spawn_fresh with (P1 := P1) (P2 := P4) (P3 := P6).
  eapply SCFacts.system_step_many_implies_program_step_many.
  eapply system_step_many_trans. apply steps_many1. apply steps1.
  inversion steps2. eauto.
  simpl. eauto. tauto.
Qed.

(** For DRF programs, the (transitive closure of the) happens-before relation
   is the same as the inherently-precedes relation. (We proved that inherently-
   precedes contains happens-before earlier; the converse holds only for DRF
   programs.) We assume in this lemma that the two actions are performed by
   different threads, since any two actions performed by the same thread are
   trivially ordered by happens-before. *)

Lemma drf_inherently_precedes_implies_happens_before_plus :
  forall Sigma1 H1 P1 H2 P2 tr t1 i1 j1 a1 t2 i2 j2 a2,
    well_typed Sigma1 H1 ->
    typing Sigma1 P1 ->
    race_free_state H1 P1 ->
    system_step_many H1 P1 H2 P2 (unerase_trace tr) ->
    event tr t1 i1 j1 ->
    event tr t2 i2 j2 ->
    index tr j1 (t1, a1) ->
    index tr j2 (t2, a2) ->
    t1 <> t2 ->
    inherently_precedes tr t1 i1 t2 i2 ->
    happens_before_plus tr j1 j2.
Proof.
  intros Sigma1 H1 P1 H2 P2 tr t1 i1 j1 a1 t2 i2 j2 a2 typing_H1 typing_H2 drf
    steps_many evt1 evt2 idx1 idx2 concurrent precedes12.

  (* Note that j2 < j1 by inherently-precedes. *)
  assert (j2 < j1) as j2_lt.
  destruct precedes12 as [ ? [ ? precedes12 ] ].
  eapply precedes12; eauto.

  (* Proof by induction on n, the number of steps between j1 and j2 *)
  assert (exists n, j1 = j2 + n + 1) as [ n j1_le ].
  exists (j1 - j2 - 1). omega.
  apply weaken_le in j1_le.
  generalize dependent j2.
  generalize dependent a1.
  generalize dependent i1.
  generalize dependent j1.
  generalize dependent t1.
  generalize dependent tr.
  induction n as [ | n ]; intros tr steps_many t1 concurrent j1 i1 evt1
    precedes12 a1 idx1 j2 evt2 idx2 j2_lt j1_le;
    pose proof index_unerase idx1 as idx1';
      pose proof index_unerase idx2 as idx2'.

  Focus.
  (* n = 0 - j1 and j2 are adjacent. Use our previous lemma. *)
  assert (j1 = j2 + 1). omega. subst j1. clear j2_lt j1_le.
  pose proof inherently_precedes_adjacent_conflict evt1 evt2 idx1 idx2
    precedes12 as conflict.
  apply Rplus_0.
  eapply drf_conflict_adjacent_implies_happens_before; eauto.
  Unfocus.

  (* Inductive case *)
  inversion j1_le as [ j1_eq | ].

  Focus.
  clear j1_le. rewrite <- j1_eq.

  (* Infer the structure of the trace. *)
  pose proof SCFacts.system_step_many_index_inversion_2 steps_many idx1'
    idx2' j2_lt as [ H3 [ P3 [ H4 [ P4 [ H5 [ P5 [ H6 [ P6 [ tr1_ [ tr2_ [ tr3_
      [ steps_many1 [ steps1 [ steps_many2 [ steps2 [ steps_many3 [ tr__eq
        [ j2_eq j1_eq' ] ] ] ] ] ] ] ] ] ] ] ] ] ] ] ] ] ].
  remember (erase_trace tr1_) as tr1.
  pose proof step_many_sequential_trace steps_many1 as seq_tr1.
  pose proof sequential_trace_erase seq_tr1 as tr1_eq.
  rewrite <- Heqtr1 in tr1_eq. subst tr1_. clear Heqtr1 seq_tr1.
  remember (erase_trace tr2_) as tr2.
  pose proof step_many_sequential_trace steps_many2 as seq_tr2.
  pose proof sequential_trace_erase seq_tr2 as tr2_eq.
  rewrite <- Heqtr2 in tr2_eq. subst tr2_. clear Heqtr2 seq_tr2.
  remember (erase_trace tr3_) as tr3.
  pose proof step_many_sequential_trace steps_many3 as seq_tr3.
  pose proof sequential_trace_erase seq_tr3 as tr3_eq.
  rewrite <- Heqtr3 in tr3_eq. subst tr3_. clear Heqtr3 seq_tr3.
  repeat rewrite <- unerase_length in *.
  assert (tr = tr3 ++ ((t2, a2) :: nil) ++ tr2 ++ ((t1, a1) :: nil) ++ tr1)
    as tr_eq.
  apply unerase_unique. repeat rewrite app_unerase. simpl in *. auto.
  clear tr__eq.

  (* What are i1 and i2? *)
  pose proof event_length (tr3 ++ ((t2, a2) :: nil) ++ tr2) a1 tr1 t1 as evt1'.
  pose proof event_length tr3 a2 (tr2 ++ ((t1, a1) :: nil) ++ tr1) t2 as evt2'.
  repeat rewrite app_ass in evt1'. rewrite <- tr_eq in evt1', evt2'.
  repeat rewrite app_projection in evt1'. repeat rewrite app_length in evt1'.
  assert (length (trace_projection ((t2, a2) :: nil) t1) = 0) as length_proj_t2.
  simpl. destruct (eq_var_dec t1 t2). subst. tauto. auto.
  rewrite length_proj_t2 in evt1'. rewrite plus_0_l in evt1'.
  simpl in evt1'. rewrite <- plus_1 in evt1'. rewrite plus_assoc in evt1'.
  rewrite j1_eq' in evt1'.
  pose proof index_unique_event evt1 evt1' as [ t1_eq i1_eq ].
  rewrite j2_eq in evt2'.
  pose proof index_unique_event evt2 evt2' as [ t2_eq i2_eq ].
  clear t1_eq t2_eq evt1' evt2'.

  (* There is at least one action between j1 and j2. Let's call the action at
     (j1 - 1 = j2 + n + 1) (t3, a3). *)
  assert (length tr2 = n + 1) as length_tr2. omega.
  pose proof length_append_exists_1 _ _ length_tr2
    as [ tr2' [ [ t3 a3 ] [ length_tr2' tr2_eq ] ] ].
  subst tr2. repeat rewrite app_ass in tr_eq.
  rewrite app_projection in i1_eq. rewrite app_length in i1_eq.
  rewrite plus_assoc in i1_eq.
  rewrite app_unerase in steps_many2. simpl in steps_many2.
  remember (j2 + n + 1) as j3.
  assert (j1 = j3 + 1) as j1_eq''. omega.

  (* a3 is at index j2 + n + 1 *)
  pose proof event_length (tr3 ++ ((t2, a2) :: nil) ++ tr2') a3
    (((t1, a1) :: nil) ++ tr1) t3 as evt3.
  repeat rewrite app_ass in evt3. repeat rewrite app_projection in evt3.
  repeat rewrite app_length in evt3. rewrite <- tr_eq in evt3.
  replace (length tr3 + (length ((t2, a2) :: nil) + length tr2'))
    with j3 in evt3; [ | subst j3 j2 n; simpl; ring ].
  assert (index tr j3 (t3, a3)) as idx3.
  subst tr j3 j2 n.
  rewrite <- plus_assoc. rewrite plus_comm. apply index_append_2.
  rewrite plus_1. rewrite <- cons_nil_app. apply index_S.
  rewrite <- (plus_0_l (length tr2')). apply index_append_2. apply index_0.
  remember (length (trace_projection tr3 t3) +
    (length (trace_projection ((t2, a2) :: nil) t3) +
      length (trace_projection tr2' t3))) as i3.

  (* Does (t3, i3) inherently-precede (t2, i2)? *)
  pose proof inherently_precedes_dec tr t3 i3 t2 i2
    as [ precedes32 | not_precedes32 ].

  (* j3 happens-before j2 (either by the inductive hypothesis, or by program
     order. *)
  assert (happens_before_plus tr j3 j2) as hb32.
  destruct (eq_var_dec t2 t3) as [ | concurrent23 ].
  subst t3. apply Rplus_0.
  eapply happens_before_program_order; eauto. omega.
  apply IHn with (t1 := t3) (i1 := i3) (a1 := a3); auto. omega. omega.

  (* Does t1 = t3? *)
  destruct (eq_var_dec t1 t3) as [ | concurrent13 ]. subst t3.

  (* t1 = t3. j1 happens-before j3 by program order, and j1 happens-before j2 by
     transitivity of happens_before. *)
  eapply Rplus_n; eauto. eapply happens_before_program_order; eauto. omega.

  (* t1 <> t3. Does a1 conflict with a3? *)
  assert (length (trace_projection ((t3, a3) :: nil) t1) = 0) as proj_t3_t1.
  simpl. destruct (eq_var_dec t1 t3). subst. tauto. auto.
  rewrite proj_t3_t1 in i1_eq. rewrite plus_0_r in i1_eq.
  destruct (effects_conflict_dec (t1, a1) (t3, a3))
    as [ conflict13 | not_conflict13 ].

  (* t1 <> t3 and a1 conflicts with a3. By the previous lemma, j1 happens-before
     j3, so j1 happens-before j2 by transitivity. *)
  apply Rplus_n with (y := j3); auto. rewrite j1_eq''.
  eapply drf_conflict_adjacent_implies_happens_before; eauto.
  rewrite <- j1_eq''. eauto.

  Focus.
  (* t1 <> t3 and e1 does not conflict with e3.  We can construct an equivalent
     trace in which j1 and j3 are flipped, then apply the IH. *)
  (* tr' is tr with a1 and a3 transposed. *)
  remember (tr3 ++ ((t2, a2) :: nil) ++ tr2' ++ ((t1, a1) :: nil) ++
    ((t3, a3) :: nil) ++ tr1) as tr'.

  (* tr and tr' (erased) are equivalent. *)
  pose proof equivalence_append_flip tr1 (tr3 ++ ((t2, a2) :: nil) ++ tr2')
    not_conflict13 as equiv.
  repeat rewrite app_ass in equiv.
  rewrite <- tr_eq in equiv. rewrite <- Heqtr' in equiv.

  (* tr' is an equivalent trace for the system *)
  assert (system_step_many H1 P1 H2 P2 (unerase_trace tr')) as steps_many'.
  eapply computational_equivalence. eauto.
  apply sequential_trace_unerase.
  repeat rewrite <- erase_unerase. auto.

  (* a1 is at index j3 in the new trace. *)
  assert (index tr' j3 (t1, a1)) as idx1_tr'.
  rewrite Heqtr'. subst j3 j2 n.
  rewrite <- plus_assoc. rewrite plus_comm. apply index_append_2.
  rewrite plus_1. apply index_S.
  rewrite <- (plus_0_l (length tr2')). apply index_append_2.
  apply index_0.

  (* a2 is at index j2 in the new trace. *)
  assert (index tr' j2 (t2, a2)) as idx2_tr'.
  rewrite Heqtr'. subst j2.
  rewrite <- (plus_0_l (length tr3)). apply index_append_2. apply index_0.

  (* a3 is at index j1 in the new trace. *)
  assert (index tr' (j3 + 1) (t3, a3)) as idx3_tr'.
  rewrite Heqtr'. subst j3 j2 n.
  repeat rewrite <- plus_assoc. rewrite plus_comm. apply index_append_2.
  rewrite plus_assoc. rewrite plus_1. apply index_S.
  rewrite plus_comm. apply index_append_2.
  apply index_S. apply index_0.

  (* Event index for a1. *)
  pose proof event_length (tr3 ++ ((t2, a2) :: nil) ++ tr2') a1
    (((t3, a3) :: nil) ++ tr1) t1 as evt1_tr'.
  repeat rewrite app_ass in evt1_tr'. rewrite <- Heqtr' in evt1_tr'.
  repeat rewrite app_projection in evt1_tr'.
  repeat rewrite app_length in evt1_tr'.
  assert (length (trace_projection ((t2, a2) :: nil) t1) = 0) as proj_t2_t1.
  simpl. destruct (eq_var_dec t1 t2). subst. tauto. auto.
  rewrite proj_t2_t1 in evt1_tr'. rewrite plus_0_l in evt1_tr'.
  rewrite <- i1_eq in evt1_tr'.
  simpl in evt1_tr'. rewrite <- plus_1 in evt1_tr'.
  rewrite plus_assoc in evt1_tr'.
  replace (length tr3 + length tr2' + 1) with j3 in evt1_tr'; [ | omega ].

  (* Event index for a2. *)
  pose proof event_length tr3 a2 (tr2' ++ ((t1, a1) :: nil) ++
    ((t3, a3) :: nil) ++ tr1) t2 as evt2_tr'.
  repeat rewrite app_ass in evt2_tr'. rewrite <- Heqtr' in evt2_tr'.
  rewrite <- i2_eq in evt2_tr'. rewrite j2_eq in evt2_tr'.

  (* The indexes are within range. *)
  assert (j2 < j3) as j2_lt_j3. omega.
  assert (j3 <= j2 + n + 1) as j3_le. omega.

  (* And inherently-precedes holds for tr' as well. *)
  pose proof equivalence_inherently_precedes precedes12 equiv as precedes12'.

  (* Finally, apply the induction hypothesis. *)
  pose proof IHn _ steps_many' _ concurrent _ _ evt1_tr' precedes12' _ idx1_tr'
    _ evt2_tr' idx2_tr' j2_lt_j3 j3_le as hb12'.
  eapply equivalence_happens_before_plus; eauto using equivalence_sym.
  Unfocus.

  Focus.
  (* (t3, i3) does not inherently precede (t2, i2). Apply the transposition
     lemma to (t3, a2) and (t2, a2). Then apply the induction hypothesis to
     (t1, i1) and (t2, i2). *)

  (* j2 < j3 *)
  assert (j2 < j3) as j2_lt_j3. omega.

  (* tr3 subtrace *)
  assert (subtrace tr 0 j2 tr3) as subtrace_tr3.
  eapply subtrace_def with (tr1 := nil).
  simpl. eauto. auto. auto.

  (* (t1, a1) ++ tr1 subtrace *)
  assert (subtrace tr (j3 + 1) (length tr1 + 1) (((t1, a1) :: nil) ++ tr1))
    as subtrace_a1_tr1.
  apply subtrace_def with (tr3 := nil)
    (tr1 := tr3 ++ ((t2, a2) :: nil) ++ tr2' ++ ((t3, a3) :: nil)).
  repeat rewrite app_ass. rewrite <- app_nil_end. auto.
  repeat rewrite app_length. simpl. omega.
  repeat rewrite app_length. simpl. ring.

  (* length of tr *)
  assert (length tr = (length tr1 + 1) + j3 + 1) as length_tr.
  subst tr j3 j2 n. repeat rewrite app_length. simpl. ring.

  (* Apply the transposition lemma. *)
  pose proof transposition evt3 evt2 j2_lt_j3 not_precedes32 length_tr
    subtrace_tr3 subtrace_a1_tr1 as [ tr' [ j4 [ equiv [ evt3_tr' [ evt2_tr'
      [ subtrace'_tr3 [ subtrace'_a1_tr1 [ j4_ge j3_le ] ] ] ] ] ] ] ].

  (* tr' is a valid computation *)
  assert (system_step_many H1 P1 H2 P2 (unerase_trace tr')) as steps_many'.
  apply computational_equivalence with (tr := unerase_trace tr).
  auto. apply sequential_trace_unerase.
  repeat rewrite <- erase_unerase. auto.

  (* Indexes for a3 and a2. *)
  assert (index tr' (j4 + 1) (t2, a2)) as idx2_tr'.
  eauto using equivalence_event.
  assert (index tr' j4 (t3, a3)) as idx3_tr'.
  eauto using equivalence_event.

  (* Note lengths of tr and tr' are equal. *)
  pose proof equivalence_length equiv as lengths_equal.

  (* Infer structure of tr'. *)
  inversion subtrace'_a1_tr1
    as [ tr3' tr1' tr0 tr'_eq length_tr3' length_a1_tr1 ]. subst tr1'.
  assert (j1 = length tr3') as j1_eq'''. omega.
  assert (tr0 = nil). rewrite length_nil.
  assert (length tr' = length (tr3' ++ (((t1, a1) :: nil) ++ tr1) ++ tr0))
    as length_tr'. rewrite tr'_eq. auto.
  repeat rewrite app_length in length_tr'. simpl in length_tr'.
  repeat rewrite app_length in length_a1_tr1. simpl in length_a1_tr1.
  subst j3 j2 n. omega. subst tr0. rewrite <- app_nil_end in tr'_eq.

  (* a1 stayed at the same index. *)
  assert (index tr' j1 (t1, a1)) as idx1_tr'.
  subst tr'. rewrite <- (plus_0_l j1). rewrite j1_eq'''.
  apply index_append_2. simpl. apply index_0.

  (* Event index for a1. *)
  assert (event tr' t1 i1 j1) as evt1_tr'.
  assert (trace_projection tr t1 = trace_projection tr' t1) as proj_eq.
  apply equivalence_projection. auto.
  rewrite tr_eq in proj_eq. rewrite tr'_eq in proj_eq.
  repeat rewrite <- app_ass in proj_eq.
  repeat rewrite (app_projection _ tr1) in proj_eq.
  apply append_end_unique in proj_eq.
  destruct proj_eq as [ proj_eq ? ].
  repeat rewrite <- app_ass in proj_eq.
  repeat rewrite (app_projection _ ((t1, a1) :: nil)) in proj_eq.
  apply append_end_unique in proj_eq.
  destruct proj_eq as [ proj_eq ? ].
  repeat rewrite app_projection in proj_eq.
  pose proof event_length tr3' a1 tr1 t1 as evt3_tr''.
  rewrite <- proj_eq in evt3_tr''.
  repeat rewrite app_length in evt3_tr''.
  rewrite length_proj_t2 in evt3_tr''. rewrite plus_0_r in evt3_tr''.
  rewrite j1_eq'''. subst tr' i1. auto.
  auto. auto.

  (* Indexes are within range. *)
  assert (j4 + 1 < j1) as j4_plus_lt_j1. omega.
  assert (j1 <= (j4 + 1) + n + 1) as j1_le. omega.

  (* Inherently-precedes continues to hold. *)
  assert (inherently_precedes tr' t1 i1 t2 i2) as precedes12'.
  eapply equivalence_inherently_precedes; eauto.

  (* Apply the induction hypothesis! *)
  pose proof IHn tr' steps_many' t1 concurrent j1 i1 evt1_tr' precedes12' a1
    idx1_tr' (j4 + 1) evt2_tr' idx2_tr' j4_plus_lt_j1 j1_le as hb12'.
  eapply equivalence_happens_before_plus; eauto using equivalence_sym.
  Unfocus.

  (* The last case is just a simple application of the inductive hypothesis. *)
  eapply IHn; eauto; omega.
Qed.

(** DRF programs produce well-synchronized traces. *)

Theorem drf_implies_well_synchronized : forall P0 H P tr,
  typing empty_heap_typing P0 ->
  race_free_program P0 ->
  system_step_many empty_heap P0 H P (unerase_trace tr) ->
  well_synchronized tr.
Proof.
  intros P0 H P tr typing_P0 drf steps_many t1 a1 j1 t2 a2 j2 idx1 idx2 race12
    j2_lt.
  pose proof index_event idx1 as [ i1 evt1 ].
  pose proof index_event idx2 as [ i2 evt2 ].

  (* If t1 = t2, we can simply use program order. *)
  destruct (eq_var_dec t1 t2). subst.
  apply Rplus_0. eauto using happens_before_program_order.

  (* If t1 <> t2, use the previous lemma. *)
  eapply drf_inherently_precedes_implies_happens_before_plus; eauto.
  apply well_typed_empty.
  eapply conflict_inherently_precedes; eauto.
  apply effects_conflict_actions.
  apply data_race_implies_conflict. auto.
Qed.
