(** This file supports indexes for list elements: the head of the list has index
   0, the next index 1, and so on. The final element in the list has index
   (length L - 1), where length L is the length of a list named L. *)

Require Import Arith.
Require Import Decidable.
Require Import List.
Require Import Omega.

Require Import nat_facts.

Set Implicit Arguments.

Section List.
  Variable A : Set.

  Inductive index : list A -> nat -> A -> Prop :=
  | index_0 : forall l a,
    index (a :: l) 0 a
  | index_S : forall l j a1 a2,
    index l j a1 ->
    index (a2 :: l) (S j) a1.

  Hint Constructors index.

  (* Indexes are always less than the length of the list. *)
  Lemma index_lt_length : forall l j a,
    index l j a ->
    j < length l.
  Proof.
    intros l j a idx.
    induction idx; simpl; auto with arith.
  Qed.

  Lemma lt_length_index : forall l j,
    j < length l ->
    exists a, index l j a.
  Proof with eauto.
    intros l.
    induction l as [ | a l ]; intros j j_lt.
    inversion j_lt.
    destruct j...
    simpl in j_lt. apply lt_S_n in j_lt.
    pose proof IHl _ j_lt as [ ? ? ]...
  Qed.

  Lemma index_valid_dec : forall l j,
    decidable (exists a, index l j a).
  Proof.
    intros l j.
    destruct (le_lt_dec (length l) j) as [ j_ge | ].
    right. intros [ a idx ].
    pose proof index_lt_length idx.
    pose proof le_not_lt _ _ j_ge. tauto.
    left. apply lt_length_index. auto.
  Qed.

  (* Each index corresponds to exactly one element. *)
  Lemma index_unique : forall l j a a',
    index l j a ->
    index l j a' ->
    a = a'.
  Proof.
    intros l j a a' idx idx'.
    generalize dependent j.
    induction l; intros j idx idx';
      inversion idx; subst; inversion idx'; subst; eauto.
  Qed.

  (* If a is at index j in l1, a is at index j in any list that starts with l1.
     *)
  Lemma index_append_1 : forall l1 l2 j a,
    index l1 j a ->
    index (l1 ++ l2) j a.
  Proof.
    intros l1 l2 j a idx.
    induction idx; simpl; auto.
  Qed.

  (* If a is at index j in l2, then prepending a prefix l1 to l2 makes the new
     index of a (j + length l1). *)
  Lemma index_append_2 : forall l1 l2 j a,
    index l2 j a ->
    index (l1 ++ l2) (j + length l1) a.
  Proof.
    intros l1 l2 j a idx.
    induction l1; simpl.
    rewrite plus_0_r. auto.
    rewrite <- plus_Snm_nSm. simpl. auto.
  Qed.

  (* If a is at index j in the list l1 ++ l2, and j falls in l1, then a is also
     at index j in list l1. *)
  Lemma index_append_inversion_1 : forall l1 l2 j a,
    index (l1 ++ l2) j a ->
    j < length l1 ->
    index l1 j a.
  Proof.
    intros l1 l2 j a idx j_lt.
    generalize dependent j.
    induction l1; intros j idx j_lt.
    inversion j_lt.
    inversion idx; subst; auto with arith.
  Qed.

  (* If a is at index j in the list l1 ++ l2, and j falls in l2, then a is at
     index j - length l1 in l2. *)
  Lemma index_append_inversion_2 : forall l1 l2 j a,
    index (l1 ++ l2) (j + length l1) a ->
    index l2 j a.
  Proof.
    intros l1 l2 j a idx.

    generalize dependent j.
    induction l1; intros j idx; simpl in *.

    rewrite plus_0_r in idx. auto.

    apply IHl1.
    inversion idx as [ ? ? [] false | ? j' ? ? ? [] j'_eq ]; subst.
    rewrite plus_comm in false. inversion false.
    rewrite S_eq_plus_S in j'_eq. subst. auto.
  Qed.

  Hint Resolve index_append_inversion_1 index_append_inversion_2.

  (* Inferring the structure of a list, given the index of a single element. *)
  Lemma index_inversion : forall l j a,
    index l j a ->
    exists l1, exists l2,
      l = l1 ++ (a :: nil) ++ l2 /\ length l1 = j.
  Proof.
    intros l j a idx.
    induction idx as [ ? a | l j a1 a2 ].
    exists (nil:list A). exists l. simpl. auto.
    destruct IHidx as [ l1 [ l2 [ ? [ ] ] ] ].
    exists (a2 :: l1). exists l2. subst l. simpl. auto.
  Qed.

  (* Inferring the structure of a list, given the indexes of two elements. *)
  Lemma index_inversion_2 : forall l j1 a1 j2 a2,
    index l j1 a1 ->
    index l j2 a2 ->
    j2 < j1 ->
    exists l1, exists l2, exists l3,
      l = l1 ++ (a2 :: nil) ++ l2 ++ (a1 :: nil) ++ l3 /\ length l1 = j2
      /\ length l1 + length l2 + 1 = j1.
  Proof.
    intros l j1 a1 j2 a2 idx1 idx2 j2_lt.
    pose proof index_inversion idx1 as [ l12 [ l3 [ ? ? ] ] ]. subst.
    pose proof index_append_inversion_1 _ _ idx2 j2_lt as idx2_l12.
    pose proof index_inversion idx2_l12 as [ l1 [ l2 [ ? [ ] ] ] ].
    subst.
    exists l1. exists l2. exists l3.
    split. rewrite app_ass. auto.
    split. auto.
    repeat rewrite app_length. simpl. ring.
  Qed.
End List.