(** We define the inherently-precedes relation over events in a trace. Two
   events are ordered by inherently-precedes for a trace tr if their order is
   preserved by all traces equivalent to tr. In other words, we can never
   transpose the two events by flipping concurrent, non-conflicting actions. (I
   believe that inherently-precedes is essentially the transitive closure of the
   union of the conflict and program order relations, although I have not proved
   that.) *)

Require Import Metatheory.

Require Import Arith.
Require Import Decidable.
Require Import List.

Require Import actions.
Require Import equivalence.
Require Import events.
Require Import indexes.
Require Import list_facts.
Require Import nat_facts.
Require Import subtraces.

Set Implicit Arguments.

Open Scope nat_scope.

Hint Constructors index equivalence event.
Hint Unfold event_valid.

Section Traces.
  Variable val : Set.

  (** An event (t1, i1) inherently precedes an event (t2, i2) in tr if, in all
     traces tr' equivalent to tr, (t1, i1) occurs before (t2, i2). *)
  Definition inherently_precedes (tr : erased_trace val) t1 i1 t2 i2 : Prop :=
    event_valid tr t1 i1 /\
    event_valid tr t2 i2 /\
    forall tr' j1 j2,
      equivalence tr tr' ->
      event tr' t1 i1 j1 ->
      event tr' t2 i2 j2 ->
      j2 < j1.

  Lemma equivalence_inherently_precedes : forall tr tr' t1 i1 t2 i2,
    inherently_precedes tr t1 i1 t2 i2 ->
    equivalence tr tr' ->
    inherently_precedes tr' t1 i1 t2 i2.
  Proof with eauto.
    unfold inherently_precedes.
    intros tr tr' t1 i1 t2 i2 [ valid1 [ valid2 precedes ] ] equiv.
    split. eauto using equivalence_valid.
    split. eauto using equivalence_valid.
    intros tr'' j1 j2 equiv' evt1 evt2...
  Qed.

  (* The inherently-precedes relation is transitive. *)
  Lemma inherently_precedes_trans : forall tr t1 i1 t2 i2 t3 i3,
    inherently_precedes tr t1 i1 t2 i2 ->
    inherently_precedes tr t2 i2 t3 i3 ->
    inherently_precedes tr t1 i1 t3 i3.
  Proof with eauto.
    intros tr t1 i1 t2 i2 t3 i3 [ valid1 [ valid2 precedes1 ] ]
      [ ? [ valid3 precedes2 ] ].
    split... split...
    intros tr' j1 j3 equiv index1 index3.
    pose proof equivalence_valid valid2 equiv as [ j2' index2' ].
    eauto using lt_trans.
  Qed.

  (** Two conflicting or non-concurrent actions in a trace are ordered by the
     inherently-precedes relation. *)
  Lemma conflict_inherently_precedes :
    forall tr t1 i1 j1 a1 t2 i2 j2 a2,
      event tr t1 i1 j1 ->
      event tr t2 i2 j2 ->
      j2 < j1 ->
      index tr j1 (t1, a1) ->
      index tr j2 (t2, a2) ->
      effects_conflict (t1, a1) (t2, a2) ->
      inherently_precedes tr t1 i1 t2 i2.
  Proof.
    intros tr t1 i1 j1 a1 t2 i2 j2 a2 evt1 evt2 j2_le index1 index2 conflict.

    (* It's trivial to prove that the event indexes are valid, but it will be
       useful to keep that information handy. *)
    assert (event_valid tr t1 i1) as valid1. eauto.
    assert (event_valid tr t2 i2) as valid2. eauto.
    split. auto. split. auto.

    (* Let tr' be such that tr is equivalent to tr'. Proof by induction on the
       rule used to derive this equivalence. *)
    intros tr' j1' j2' equiv evt1' evt2'.

    (* Now we must prove that j2' < j1' *)
    generalize dependent j2'.
    generalize dependent j1'.
    generalize dependent j2.
    generalize dependent i2.
    generalize dependent j1.
    generalize dependent i1.
    induction equiv as [
      e3 e4 not_conflict (* flip *)
      | tr1 tr2 tr1' tr2' equiv1 ? equiv2 (* append *)
      | (* reflexivity *)
      | tr tr'' tr' equiv ? equiv' (* transitivity *)
    ]; intros i1 valid1 j1 evt1 idx1 i2 valid2 j2 evt2 j2_lt
    idx2 j1' evt1' j2' evt2'.

    (* Case: both traces have exactly two actions, in either order. *)
    (* j1 must be 1. *)
    inversion idx1 as [ | ? P_j1 ? ? j1_is_1 ]; subst.
    inversion j2_lt.
    inversion j1_is_1 as [ | ? ? ? ? false ]; [ subst | inversion false ].
    (* And j2 must be 0. *)
    inversion j2_lt as [ | ? false ]; [ subst | inversion false ].
    inversion idx2; subst.
    (* Contradiction: the two events do not conflict, which violates our
       assumption. *)
    contradiction not_conflict. apply effects_conflict_sym. auto.

    (* Case: tr = tr1 ++ tr2 and tr' = tr1' ++ tr2' and tr1 is equivalent to
       tr1' and tr2 is equivalent to tr2' *)
    pose proof equivalence_length equiv1 as lengths1.
    pose proof equivalence_length equiv2 as lengths2.
    pose proof equivalence_projection t1 equiv1 as projections_equal1_t1.
    pose proof equivalence_projection t2 equiv1 as projections_equal1_t2.

    (* Basically we need to know whether the conflicting actions are both in
       tr1, both in tr2, or one in tr1 and one in tr2. *)
    destruct (event_append_inversion tr1 tr2 evt1)
      as [ [ j1_lt evt1_tr1 ]
        | [ i1'' [ j1'' [ j1_eq [ i1_eq evt1_tr2 ] ] ] ] ];
      [ pose proof equivalence_append_1 _ evt1_tr1 equiv1
        evt1' as [ j1'_lt evt1'_tr1' ]
        | subst j1; pose proof equivalence_append_2 _ _ _ evt1
          equiv1 evt1'
          as [ i1''' [ j1''' [ j1'_eq [ i1_eq' evt1'_tr2' ] ] ] ] ];
      (destruct (event_append_inversion tr1 tr2 evt2)
        as [ [ ? evt2_tr1 ]
          | [ i2'' [ j2'' [ j2_eq [ i2_eq evt2_tr2 ] ] ] ] ];
        [ pose proof equivalence_append_1 _ evt2_tr1 equiv1
          evt2' as [ j2'_lt evt2'_tr1' ]
          | subst j2; pose proof equivalence_append_2 _ _ _ evt2
            equiv1 evt2'
            as [ i2''' [ j2''' [ j2'_eq [ i2_eq' evt2'_tr2' ] ] ] ] ]).

    (* Both in tr1 *)
    eauto 8 using index_append_inversion_1.

    (* j1 in tr1, j2 in tr2 (contradiction) *)
    subst. apply lt_asym in j2_lt. contradiction j2_lt. omega.

    (* j1 in tr2, j2 in tr1 *)
    pose proof event_lt_length evt2'_tr1' as j1'_lt.
    subst. omega.

    (* Both in tr2 *)
    assert (j2''' < j1''').
    eapply IHequiv2; eauto using index_append_inversion_2.
    eapply plus_lt_reg_l with (p := length tr1).
    rewrite plus_comm. replace (length tr1 + j1'') with (j1'' + length tr1).
    auto. apply plus_comm.
    rewrite projections_equal1_t1 in i1_eq.
    subst i1. apply plus_reg_r in i1_eq'. subst i1''. auto.
    rewrite projections_equal1_t2 in i2_eq.
    subst i2. apply plus_reg_r in i2_eq'. subst i2''. auto.
    subst. apply plus_lt_compat_r. auto.

    (* Case: tr = tr' *)
    pose proof event_unique_index evt1 evt1'; subst.
    pose proof event_unique_index evt2 evt2'; subst.
    auto.

    (* Case: tr is equivalent to tr'', tr'' is equivalent to tr' *)
    pose proof (equivalence_valid valid1 equiv) as [ j1'' evt1'' ].
    pose proof (equivalence_valid valid2 equiv) as [ j2'' evt2'' ].
    assert (j2'' < j1''). eauto.
    eauto 8 using equivalence_event.
  Qed.

  (** The contrapositive of the above. *)
  Lemma not_inherently_precedes_not_conflict :
    forall tr t1 i1 j1 a1 t2 i2 j2 a2,
      event tr t1 i1 j1 ->
      event tr t2 i2 j2 ->
      j2 < j1 ->
      index tr j1 (t1, a1) ->
      index tr j2 (t2, a2) ->
      ~ inherently_precedes tr t1 i1 t2 i2 ->
      ~ effects_conflict (t1, a1) (t2, a2).
  Proof.
    intros tr t1 i1 j1 a1 t2 i2 j2 a2 evt1 evt2 j2_lt idx1 idx2 not_precedes
      conflict.
    eapply not_precedes.
    eapply conflict_inherently_precedes; eauto.
  Qed.

  (** If two adjacent actions in a trace are ordered by inherently-precedes,
     then they must be either conflicting or non-concurrent. Basically the
     argument is that if the actions are concurrent and non-conflicting, we can
     construct an equivalent trace in which they're flipped, violating
     inherently-precedes.  *)
  Lemma inherently_precedes_adjacent_conflict :
    forall tr j t1 i1 a1 t2 i2 a2,
      event tr t1 i1 (j + 1) ->
      event tr t2 i2 j ->
      index tr (j + 1) (t1, a1) ->
      index tr j (t2, a2) ->
      inherently_precedes tr t1 i1 t2 i2 ->
      effects_conflict (t1, a1) (t2, a2).
  Proof.
    intros tr j t1 i1 a1 t2 i2 a2 evt1 evt2 idx1 idx2 precedes.

    destruct (effects_conflict_dec (t1, a1) (t2, a2))
      as [ | not_conflict ]; auto.

    assert (j < j + 1) as j_lt. omega.
    pose proof index_inversion_2 idx1 idx2 j_lt
      as [ tr1 [ tr2 [ tr3 [ tr_eq [ ? ? ] ] ] ] ].
    assert (tr2 = nil). rewrite length_nil. omega. subst tr2.

    rewrite <- app_nil_begin in tr_eq.
    remember (tr1 ++ ((t1, a1) :: nil) ++ ((t2, a2) :: nil) ++ tr3) as tr'.
    assert (equivalence tr tr') as equiv.
    subst. apply equivalence_append_flip; auto.

    assert (i1 = length (trace_projection tr1 t1)).
    subst tr j. rewrite plus_comm in evt1.
    pose proof event_append_inversion_2 _ _ _ evt1 as [ ? [ ? evt1' ] ].
    inversion evt1' as [ | | ? ? ? ? ? ? evt1'' ]; subst.
    contradiction not_conflict. apply effects_conflict_program_order.
    inversion evt1''; subst. auto.

    assert (i2 = length (trace_projection tr1 t2)).
    subst tr j. rewrite <- plus_0_l in evt2.
    pose proof event_append_inversion_2 _ _ _ evt2 as [ ? [ ? evt2' ] ].
    inversion evt2'; subst. auto.

    assert (event tr' t1 i1 j).
    subst. rewrite <- plus_0_l.
    rewrite <- (plus_0_l (length (trace_projection tr1 t1))).
    apply event_append_2. apply event_0.

    assert (event tr' t2 i2 (j + 1)).
    subst. rewrite plus_comm.
    rewrite <- (plus_0_l (length (trace_projection tr1 t2))).
    apply event_append_2. apply event_neq_S. apply event_0. auto.
    intro. subst.
    contradiction not_conflict. apply effects_conflict_program_order.

    assert (j + 1 < j) as false.
    destruct precedes as [ ? [ ? precedes ] ].
    eauto.

    contradiction (lt_irrefl j). omega.
  Qed.

  (** Inherently-precedes is decidable for adjacent actions in a trace. *)
  Lemma inherently_precedes_adjacent_dec : forall tr t1 i1 t2 i2 j2,
    event tr t1 i1 (j2 + 1) ->
    event tr t2 i2 j2 ->
    decidable (inherently_precedes tr t1 i1 t2 i2).
  Proof.
    intros tr t1 i1 t2 i2 j2 evt1 evt2.
    pose proof event_index evt1 as [ a1 idx1 ].
    pose proof event_index evt2 as [ a2 idx2 ].
    destruct (effects_conflict_dec (t1, a1) (t2, a2)) as [ | not_conflict ].
    left. eapply conflict_inherently_precedes; eauto. omega.
    right. intro precedes. apply not_conflict.
    eapply inherently_precedes_adjacent_conflict; eauto.
  Qed.

  (** The base case for the transposition lemma. If two adjacent actions are
     not ordered by inherently-precedes, then we can construct an equivalent
     trace in which the two actions are flipped. *)
  Lemma transposition_adjacent : forall tr t1 i1 t2 i2 j tr1 tr3 len_tr3,
    event tr t1 i1 (j + 1) ->
    event tr t2 i2 j ->
    ~ inherently_precedes tr t1 i1 t2 i2 ->
    length tr = len_tr3 + j + 2 ->
    subtrace tr 0 j tr1 ->
    subtrace tr (j + 2) len_tr3 tr3 ->
    exists tr',
      equivalence tr tr' /\
      event tr' t1 i1 j /\
      event tr' t2 i2 (j + 1) /\
      subtrace tr' 0 j tr1 /\
      subtrace tr' (j + 2) len_tr3 tr3.
  Proof.
    intros tr t1 i1 t2 i2 j tr1 tr3 len_tr3 evt1 evt2 not_precedes
      length_tr subtrace_tr1 subtrace_tr3.

    (* Which action does each index point to? (a1 and a2) *)
    pose proof event_inversion evt1 as [ a1 [ idx1 idx1_proj ] ].
    pose proof event_inversion evt2 as [ a2 [ idx2 idx2_proj ] ].

    (* Note that j < j + 1 *)
    assert (j < j + 1) as j_lt. omega.

    (* By a previous lemma, we can show that a1 and a2 are non-conflicting, and
       they do not occur in the same thread. *)
    pose proof not_inherently_precedes_not_conflict evt1
      evt2 j_lt idx1 idx2 not_precedes as not_conflict.

    (* We can infer the structure of the trace (namely, tr1 ++ a2 ++ a1 ++ tr3).
       *)
    pose proof index_inversion_2 idx1 idx2 j_lt
      as [ tr1' [ tr2 [ tr3' [ tr_eq [ j2_eq j1_eq ] ] ] ] ].

    (* Establish that the tr1' from the trace inversion is indeed tr1. *)
    assert (tr1 = tr1').
    eapply subtrace_unique. eauto. eapply subtrace_def with (tr1 := nil).
    simpl. eauto. auto. auto. subst tr1'.

    (* And that tr2 is nil. *)
    assert (tr2 = nil). rewrite length_nil. omega. subst tr2.

    (* len_tr3 is the length of tr3. *)
    assert (len_tr3 = length tr3') as len_tr3_eq'.
    rewrite tr_eq in length_tr.
    repeat rewrite app_length in length_tr.
    simpl in length_tr. subst j. omega.

    (* The tr3' from the trace inversion is indeed tr3. *)
    assert (tr3 = tr3').
    eapply subtrace_unique. eauto. eapply subtrace_def with (tr3 := nil).
    rewrite <- app_ass. rewrite <- app_nil_end.
    repeat rewrite <- app_ass in tr_eq. apply tr_eq.
    repeat rewrite app_length. simpl. subst j. ring.
    auto. subst tr3'.

    (* We also have formulas for i1 and i2 in terms of the lengths of the
       projections of tr1 and tr2 onto t. *)
    pose proof event_length tr1 a2 (((t1, a1) :: nil) ++ tr3) t2 as evt2'.
    assert (i2 = length (trace_projection tr1 t2)).
    subst.
    pose proof index_unique_event evt2 evt2' as [ ]. auto.
    pose proof event_length (tr1 ++ ((t2, a2) :: nil)) a1 tr3 t1 as evt1'.
    replace (length (tr1 ++ ((t2, a2) :: nil))) with (length tr1 + 1) in evt1';
      [ | repeat rewrite app_length; simpl; ring ].
    repeat rewrite app_ass in evt1'.
    assert (i1 = length (trace_projection (tr1 ++ ((t2, a2) :: nil)) t1)).
    subst.
    pose proof index_unique_event evt1 evt1' as [ ]. auto.

    (* This is mostly a nightmarish manipulation of appends and lengths and such.
       Basically I'm just showing you can flip the two actions. *)
    exists (tr1 ++ ((t1, a1) :: nil) ++ ((t2, a2) :: nil) ++ tr3).

    (* The traces are equivalent. *)
    split. subst. apply equivalence_append; auto.
    repeat rewrite <- app_ass.
    apply equivalence_append; auto.
    simpl. auto using effects_conflict_sym.

    (* a1 appears at position j in the new trace. *)
    split. subst.
    rewrite app_projection.
    replace (trace_projection ((t2, a2) :: nil) t1)
      with (nil : projection val).
    rewrite <- app_nil_end.
    apply event_length.
    simpl.
    destruct (eq_var_dec t1 t2).
    subst. contradiction not_conflict. apply effects_conflict_program_order.
    auto.

    (* a2 appears at position j2 + 1 in the new trace. *)
    split. subst.
    replace (length tr1 + 1) with (length (tr1 ++ ((t1, a1) :: nil))).
    replace (length (trace_projection tr1 t2))
      with (length (trace_projection (tr1 ++ ((t1, a1) :: nil)) t2)).
    rewrite <- app_ass.
    apply event_length.
    rewrite app_projection. simpl.
    destruct (eq_var_dec t2 t1) as [ false | ].
    subst. contradiction not_conflict. apply effects_conflict_program_order.
    rewrite <- app_nil_end. auto.
    rewrite app_length. simpl. auto.

    (* tr1 is preserved at the beginning of tr'. *)
    split. eapply subtrace_def with (tr1 := nil); simpl; eauto.

    (* tr3 is preserved at the end of tr'. *)
    eapply subtrace_def with (tr3 := nil)
      (tr1 := tr1 ++ ((t1, a1) :: nil) ++ ((t2, a2) :: nil)).
    repeat rewrite app_ass. rewrite <- app_nil_end. auto.
    subst. repeat rewrite app_length. simpl. auto. auto.
  Qed.

  (** This is lemma 3.10 from the paper (transposition), plus the decidability
     of inherently-precedes. If two actions are not ordered by the inherently-
     precedes relation, then we can construct a trace in which (1) they appear
     in reverse order and (2) the new trace is identical to the old trace
     except in between the two actions. That is, the subtrace from the beginning
     of the trace to just before the first action does not change, and neither
     does the subtrace from right after the second action to the end. *)
  Lemma transposition_helper : forall tr t1 i1 j1 t2 i2 j2 tr1 tr3 len_tr3,
    event tr t1 i1 j1 ->
    event tr t2 i2 j2 ->
    j2 < j1 ->
    length tr = len_tr3 + j1 + 1 ->
    subtrace tr 0 j2 tr1 ->
    subtrace tr (j1 + 1) len_tr3 tr3 ->
    inherently_precedes tr t1 i1 t2 i2 \/
    ~ inherently_precedes tr t1 i1 t2 i2 /\
    exists tr', exists j',
      equivalence tr tr' /\
      event tr' t1 i1 j' /\
      event tr' t2 i2 (j' + 1) /\
      subtrace tr' 0 j2 tr1 /\
      subtrace tr' (j1 + 1) len_tr3 tr3 /\
      j' >= j2 /\
      S j' <= j1.
  Proof.
    intros tr t1 i1 j1 t2 i2 j2 tr1 tr3 len_tr3 evt1 evt2 j2_lt len_tr3_eq
      subtrace_tr1 subtrace_tr3.

    (** Let n = j1 - j2 - 1 (the number of actions between j1 and j2. We use
       strong induction on n. *)
    assert (exists n, j1 = j2 + n + 1) as [ n j1_le ].
    exists (j1 - j2 - 1). omega.
    apply weaken_le in j1_le.

    (* Generalize everything and do the induction. *)
    generalize dependent len_tr3.
    generalize dependent tr3.
    generalize dependent tr1.
    generalize dependent j2.
    generalize dependent i2.
    generalize dependent t2.
    generalize dependent j1.
    generalize dependent i1.
    generalize dependent t1.
    generalize dependent tr.
    induction n as [ | n ];
      intros tr t1 i1 j1 evt1 t2 i2 j2 evt2 j2_lt j1_le tr1 subtrace_tr1 tr3
        len_tr3 length_tr subtrace_tr3.

    Focus.
    (** Base case : n = 0, so j1 = j2 + 1. Use previous lemmas to decide if j1
       inherently-precedes, and if it does not, to construct a trace in which j1
       and j2 are flipped. *)
    assert (j1 = j2 + 1). omega. subst j1.
    replace (len_tr3 + (j2 + 1) + 1) with (len_tr3 + j2 + 2) in length_tr.
    replace (j2 + 1 + 1) with (j2 + 2) in subtrace_tr3.
    destruct (inherently_precedes_adjacent_dec evt1 evt2) as [ | not_precedes ].
    auto. right. split. auto.
    pose proof transposition_adjacent evt1 evt2 not_precedes
      length_tr subtrace_tr1 subtrace_tr3
      as [ tr' [ equiv [ evt1' [ evt2' [ subtrace'_tr1
        subtrace'_tr3 ] ] ] ] ].
    exists tr'. exists j2.
    repeat (split; auto).
    replace (j2 + 1 + 1) with (j2 + 2). auto.
    ring. ring. ring.
    Unfocus.

    (** Inductive case. Assume j1 = j2 + n + 2. *)
    inversion j1_le as [ j1_eq | ]. clear j1_le.
    replace (j2 + S n + 1) with (j2 + n + 2) in j1_eq; [ | ring ].

    (** Let j3 be the index right next to j2. *)
    remember (j2 + 1) as j3.
    assert (j3 < length tr) as j3_lt.
    subst. rewrite length_tr. omega.
    pose proof lt_length_event _ j3_lt as [ t3 [ i3 evt3 ] ].

    (* First, we need a name for the action pointed to (t2, i2). Call it a2. *)
    pose proof event_inversion evt2 as [ a2 [ idx2 idx2_proj ] ].

    (* Note that the first j3 (= j2 + 1) elements of tr are tr1 ++ (a2 :: nil)
       *)
    pose proof subtrace_plus subtrace_tr1 idx2 as subtrace_tr1_a2.

    (* And that j2 = length tr1. *)
    pose proof subtrace_inversion_length subtrace_tr1 as length_tr1.

    (** Does (t3, i3) inherently-precede (t2, i2)? *)
    assert (decidable (inherently_precedes tr t3 i3 t2 i2))
      as [ precedes32 | not_precedes32 ].
    subst j3. eapply inherently_precedes_adjacent_dec; eauto.

    (** (t3, i3) inherently-precedes (t2, i2). Apply the IH to (t1, i1) and
       (t3, i3) to determine if (t1, i1) inherently-precedes (t3, i3).  If not,
       the IH provides a new trace in which (t1, i1) and (t3, i3) are
       transposed. *)
    assert (inherently_precedes tr t1 i1 t3 i3 \/
      (~ inherently_precedes tr t1 i1 t3 i3 /\
        exists tr', exists j1',
          equivalence tr tr' /\
          event tr' t1 i1 j1' /\
          event tr' t3 i3 (j1' + 1) /\
          subtrace tr' 0 j3 (tr1 ++ ((t2, a2) :: nil)) /\
          subtrace tr' (j1 + 1) len_tr3 tr3 /\
          j1' >= j3 /\
          S j1' <= j1))
    as [ precedes13 | [ not_precedes13 [ tr' [ j1' [ equiv [ evt1' [ evt3'
      [ subtrace'_tr1_a2 [ subtrace'_tr3 [ j1'_ge S_j1'_le ] ] ] ] ] ] ] ] ] ].
    apply IHn; subst j3; auto; omega.

    (** (t1, i1) inherently-precedes (t3, i3). That implies that (t1, i1)
       inherently-precedes (t2, i2) by transitivity, so we can conclude. *)
    left. eauto using inherently_precedes_trans.

    Focus.
    (** (t1, i1) does not inherently-precede (t3, i3).  We take the new trace
       in which (t1, i1) and (t3, i3) are transposed, and apply the IH again
       on (t1, i1) and (t2, i2). *)

    (* We need a name for the subtrace of tr' starting at j1' and ending just
       before tr3 - don't really care what's in it, just need to know it's
       there. *)
    assert (exists len_tr4, length tr = (j1' + 1) + (len_tr4 + len_tr3))
      as [ len_tr4 length_tr' ].
    rewrite length_tr. exists (j1 - j1'). omega.
    assert (exists tr4, subtrace tr' (j1' + 1) (len_tr4 + len_tr3) (tr4 ++ tr3))
      as [ tr4 subtrace'_tr4_tr3 ].
    apply subtrace_extend_begin.
    rewrite length_tr in length_tr'.
    assert (j1' + 1 + len_tr4 = j1 + 1) as j1'_eq. omega.
    rewrite j1'_eq. auto.
    assert (length tr4 = len_tr4).
    pose proof subtrace_inversion_length subtrace'_tr4_tr3 as length_tr4_tr3.
    rewrite app_length in length_tr4_tr3.
    pose proof subtrace_inversion_length subtrace_tr3 as length_tr3. omega.

    (* a2 is still in the same place as before. *)
    assert (index tr' j2 (t2, a2)) as idx2'.
    rewrite <- (plus_0_l j2).
    eapply subtrace_plus_inversion.
    subst j3. apply subtrace'_tr1_a2.

    (* And (t2, i2) has the same trace event index in the new trace. *)
    assert (event tr' t2 i2 j2) as evt2'.
    pose proof subtrace_inversion_begin subtrace_tr1_a2 as [ ].
    pose proof subtrace_inversion_begin subtrace'_tr1_a2 as [ ].
    subst tr tr'. simpl. apply event_append_1.
    eapply event_append_inversion_1.
    eauto. subst j2. rewrite app_length. simpl. omega.

    (** Apply the IH to (t1, i1) and (t2, i2) in the new trace.  If (t1, i1)
       does not inherently-precede (t2, i2), this will produce a new trace in
       which (t1, i1) and (t2, i2) are transposed. *)
    assert (inherently_precedes tr' t1 i1 t2 i2 \/
      ~ inherently_precedes tr' t1 i1 t2 i2 /\
      exists tr'', exists j1'',
        equivalence tr' tr'' /\
        event tr'' t1 i1 j1'' /\
        event tr'' t2 i2 (j1'' + 1) /\
        subtrace tr'' 0 j2 tr1 /\
        subtrace tr'' (j1' + 1) (len_tr4 + len_tr3) (tr4 ++ tr3) /\
        j1'' >= j2 /\
        S j1'' <= j1')
    as [ precedes12 | [ not_precedes12 [ tr'' [ j1'' [ equiv' [ evt1'' [ evt2''
      [ subtrace''_tr1 [ subtrace''_tr4_tr3 [ j1''_ge S_j1''_le ] ] ] ] ] ] ]
    ] ] ].
    apply IHn; auto. omega. omega.
    eapply subtrace_minus. subst. eauto.
    rewrite <- equivalence_length with (tr1 := tr). omega. auto.

    (** (t1, i1) inherently-precedes (t2, i2). Conclude. *)
    left. eapply equivalence_inherently_precedes. eauto.
    eapply equivalence_sym. auto.

    (** (t1, i1) does not inherently-precede (t2, i2). Conclude using the trace
       given by the inductive hypothesis. *)
    right. split. intro precedes12. apply not_precedes12.
    eapply equivalence_inherently_precedes; eauto.

    (* Note tr'' still has tr3 at the end. *)
    assert (subtrace tr'' (j1 + 1) len_tr3 tr3) as subtrace''_tr3.
    assert (j1 + 1 = j1' + 1 + len_tr4) as j1_plus_1_eq. omega.
    rewrite j1_plus_1_eq.
    eapply subtrace_extend_begin_inversion. eauto. auto.

    (* The new trace tr'' we got from the second inductive step is the one we
       need to finish off this case. *)
    exists tr''. exists j1''.
    split. eauto using equivalence_trans.
    split. auto.
    split. auto.
    split. auto.
    split. replace (j2 + S n + 1 + 1) with (j1 + 1). auto. subst. ring.
    split. auto.
    subst. omega.
    Unfocus.

    Focus.
    (** (t3, i3) does not inherently-precede (t2, i2).  Apply a previous lemma
       to produce a new trace in which (t3, i3) and (t2, i2) are flipped, then
       apply the IH to decide whether (t1, i1) inherently-precedes (t2, i2).
       If (t1, i1) does not inherently-precede (t2, i2), the IH produces a
       trace in which (t1, i1) and (t2, i2) are transposed. *)

    (* We need a name for the subtrace of tr starting at j2 + 2 and ending just
       before tr3 - don't really care what's in it, just need to know it's
       there. *)
    assert (exists len_tr4, length tr = (j2 + 2) + (len_tr4 + len_tr3))
      as [ len_tr4 length_tr4 ].
    rewrite length_tr. exists (j1 - j2 - 1). omega.
    assert (exists tr4, subtrace tr (j2 + 2) (len_tr4 + len_tr3) (tr4 ++ tr3))
      as [ tr4 subtrace_tr4_tr3 ].
    apply subtrace_extend_begin.
    rewrite length_tr in length_tr4.
    assert (j2 + 2 + len_tr4 = j1 + 1) as j1_plus_1_eq. omega.
    rewrite j1_plus_1_eq. auto.
    assert (length tr4 = len_tr4).
    pose proof subtrace_inversion_length subtrace_tr4_tr3 as length_tr4_tr3.
    rewrite app_length in length_tr4_tr3.
    pose proof subtrace_inversion_length subtrace_tr3 as length_tr3. omega.

    (** Apply the adjacent-transposition lemma to (t3, i3) and (t2, i2) *)
    assert (exists tr',
      equivalence tr tr' /\
      event tr' t3 i3 j2 /\
      event tr' t2 i2 (j2 + 1) /\
      subtrace tr' 0 j2 tr1 /\
      subtrace tr' (j2 + 2) (len_tr4 + len_tr3) (tr4 ++ tr3))
    as [ tr' [ equiv [ evt3' [ evt2' [ subtrace'_tr1
      subtrace'_tr4_tr3 ] ] ] ] ].
    subst j3. apply transposition_adjacent; auto; omega.

    (* We need a names for the actions pointed to by (t1, a1), (t2, a2) and
       (t3, i3). Call them a1, a2 and a3. *)
    pose proof event_inversion evt1 as [ a1 [ idx1 idx1_proj ] ].
    pose proof event_inversion evt3 as [ a3 [ idx3 idx3_proj ] ].

    (* a2 and a2 are at positions j2 and j2 + 1, respectively, in tr'. *)
    assert (index tr' (j2 + 1) (t2, a2)) as idx2'.
    eauto using equivalence_event.
    assert (index tr' j2 (t3, a3)) as idx3'.
    eauto using equivalence_event.

    (* Infer the structure of tr. *)
    subst j3. rewrite (plus_comm (j2 + 2)) in length_tr4.
    rewrite plus_assoc in length_tr4.
    pose proof subtrace_inversion_2 subtrace_tr1 idx2 idx3 length_tr4
      subtrace_tr4_tr3 as tr_eq.

    (* Infer the structure of tr'. *)
    rewrite (equivalence_length equiv) in length_tr4.
    pose proof subtrace_inversion_2 subtrace'_tr1 idx3' idx2' length_tr4
      subtrace'_tr4_tr3 as tr'_eq.

    (* j2 = length tr1 *)
    pose proof subtrace_inversion_length subtrace_tr1 as j2_eq.

    (* (t1, i1) is at the same position (j1) in tr'. (This one's a bit ugly.) *)
    assert (event tr' t1 i1 j1) as evt1'.
    rewrite <- app_ass in tr_eq. rewrite <- app_ass in tr_eq.
    rewrite (app_ass _ ((t2, a2) :: nil)) in tr_eq. subst tr.
    assert (j1 = n + (length (tr1 ++ ((t2, a2) :: nil) ++ (t3, a3) :: nil)))
      as j1_eq'.
    subst j2. repeat rewrite app_length. simpl. omega.
    rewrite j1_eq' in evt1.
    pose proof event_append_inversion_2 _ _ _ evt1
      as [ m [ i1_eq evt1_tr4_tr3 ] ].
    rewrite <- app_ass in tr'_eq. rewrite <- app_ass in tr'_eq.
    rewrite (app_ass _ ((t3, a3) :: nil)) in tr'_eq. subst tr'.
    assert (j1 = n + (length (tr1 ++ ((t3, a3) :: nil) ++ (t2, a2) :: nil)))
      as j1_eq''.
    subst j2. repeat rewrite app_length. simpl. omega.
    rewrite j1_eq''. subst i1.
    replace (length (trace_projection (tr1 ++ ((t2, a2) :: nil) ++
      (t3, a3) :: nil) t1)) with (length (trace_projection (tr1 ++
        ((t3, a3) :: nil) ++ (t2, a2) :: nil) t1)).
    apply event_append_2. auto.
    repeat rewrite app_projection. repeat rewrite app_length. ring.

    (* The beginning of tr' is tr1 ++ (t3, a3) :: nil. *)
    assert (subtrace tr' 0 (j2 + 1) (tr1 ++ ((t3, a3) :: nil)))
      as subtrace'_tr1_a3.
    apply subtrace_def with (tr1 := nil) (tr3 := ((t2, a2) :: nil) ++ tr4 ++ tr3).
    simpl. rewrite app_ass. rewrite <- cons_nil_app. simpl in tr'_eq. auto.
    simpl. auto. rewrite app_length. simpl. subst j2. auto.

    (* The end of tr' is tr3 *)
    assert (subtrace tr' (j1 + 1) len_tr3 tr3) as subtrace'_tr3.
    apply subtrace_def with (tr1 := tr1 ++ ((t3, a3) :: nil) ++ ((t2, a2) :: nil)
      ++ tr4) (tr3 := nil).
    simpl. repeat rewrite app_ass. simpl. rewrite <- app_nil_end.
    simpl in tr'_eq. auto.
    subst j1 j2. repeat rewrite app_length. simpl.
    rewrite <- (equivalence_length equiv) in length_tr4. omega.
    eapply subtrace_inversion_length. eauto.

    (** Apply the inductive hypothesis to (t1, i1) and (t2, i2). *)
    assert (inherently_precedes tr' t1 i1 t2 i2 \/
      ~ inherently_precedes tr' t1 i1 t2 i2 /\
      exists tr'', exists j1'',
        equivalence tr' tr'' /\
        event tr'' t1 i1 j1'' /\
        event tr'' t2 i2 (j1'' + 1) /\
        subtrace tr'' 0 (j2 + 1) (tr1 ++ ((t3, a3) :: nil)) /\
        subtrace tr'' (j1 + 1) len_tr3 tr3 /\
        j1'' >= j2 + 1 /\
        S j1'' <= j1)
    as [ precedes12 | [ not_precedes12 [ tr'' [ j1'' [ equiv' [ evt1'' [ evt2''
      [ subtrace''_tr1_a3 [ subtrace''_tr3 [ j1''_ge S_j1''_le ] ] ] ] ] ] ] ]
    ] ].
    apply IHn; auto. omega. omega.
    rewrite <- equivalence_length with (tr1 := tr); auto.

    (** (t1, i1) inherently-precedes (t2, i2). Conclude. *)
    left. eapply equivalence_inherently_precedes; eauto.
    apply equivalence_sym; auto.

    (** (t1, i1) does not inherently-precede (t2, i2). The trace given by the IH
       is our witness. *)
    right. split. intro precedes12. apply not_precedes12.
    eapply equivalence_inherently_precedes; eauto.

    (* tr'' is the trace we need. *)
    exists tr''. exists j1''.
    split. eauto using equivalence_trans.
    split. auto.
    split. auto.
    split. eapply subtrace_minus. eauto.
    split. subst j1.
    replace (j2 + S n + 1 + 1) with (j2 + n + 2 + 1). auto. ring.
    split. omega.
    subst j1. omega.

    Unfocus.

    (** This is a trivial case resulting from the way I set up strong induction.
       There's probably a better way to do strong induction. *)
    apply IHn; auto; omega.
  Qed.

  (** If two events are not ordered by inherently-precedes, then there exists
     an equivalent trace in which those two events are transposed. Moreover,
     the order of events before the first event and after the second event is
     unchanged. *)
  Corollary transposition : forall tr t1 i1 j1 t2 i2 j2 tr1 tr3 len_tr3,
    event tr t1 i1 j1 ->
    event tr t2 i2 j2 ->
    j2 < j1 ->
    ~ inherently_precedes tr t1 i1 t2 i2 ->
    length tr = len_tr3 + j1 + 1 ->
    subtrace tr 0 j2 tr1 ->
    subtrace tr (j1 + 1) len_tr3 tr3 ->
    exists tr', exists j',
      equivalence tr tr' /\
      event tr' t1 i1 j' /\
      event tr' t2 i2 (j' + 1) /\
      subtrace tr' 0 j2 tr1 /\
      subtrace tr' (j1 + 1) len_tr3 tr3 /\
      j' >= j2 /\
      S j' <= j1.
  Proof.
    intros tr t1 i1 j1 t2 i2 j2 tr1 tr3 len_tr3 evt1 evt2 j2_lt
      not_precedes len_tr3_eq subtrace_tr1 subtrace_tr3.

    pose proof transposition_helper evt1 evt2 j2_lt len_tr3_eq subtrace_tr1
      subtrace_tr3 as [ precedes12 | [ not_precedes12 ? ] ].
    tauto. auto.
  Qed.

  (** Given two events in a trace, we can decide whether the two events are
     ordered by inherently-precedes. (This is a technical point that keeps us
     in the constructive logic world.) *)
  Corollary inherently_precedes_dec : forall tr t1 i1 t2 i2,
    decidable (inherently_precedes tr t1 i1 t2 i2).
  Proof.
    intros tr t1 i1 t2 i2.
    pose proof event_valid_dec tr t1 i1 as [ valid1 | not_valid1 ].
    pose proof event_valid_dec tr t2 i2 as [ valid2 | not_valid2 ].
    destruct valid1 as [ j1 evt1 ].
    destruct valid2 as [ j2 evt2 ].
    destruct (le_lt_dec j1 j2) as [ j1_le | j2_lt ].

    (* j1 <= j2 - does not precede *)
    right. intros [ ? [ ? precedes ] ].
    pose proof precedes _ _ _ (equivalence_refl tr) evt1 evt2.
    assert (j2 < j2) as false. omega.
    contradiction (lt_irrefl _ false).

    (* j2 < j1 - apply helper lemma *)
    pose proof transposition_helper.
    pose proof event_lt_length evt2 as j2_lt_length.
    assert (exists m, length tr = j2 + m) as [ m m_eq ].
    exists (length tr - j2). omega.
    pose proof length_append_exists _ _ _ m_eq
      as [ tr1 [ ? [ ? [ ? ? ] ] ] ].
    assert (subtrace tr 0 j2 tr1) as subtrace_tr1.
    eapply subtrace_def with (tr1 := nil); eauto.
    pose proof event_lt_length evt1 as j1_lt_length.
    assert (exists n, length tr = j1 + 1 + n) as [ n n_eq ].
    exists (length tr - j1 - 1). omega.
    pose proof length_append_exists _ _ _ n_eq
      as [ ? [ tr3 [ ? [ ? ? ] ] ] ].
    assert (subtrace tr (j1 + 1) n tr3) as subtrace_tr3.
    eapply subtrace_def with (tr3 := nil); eauto.
    rewrite <- app_ass. rewrite <- app_nil_end. eauto.
    rewrite plus_comm in n_eq. rewrite plus_assoc in n_eq.
    pose proof transposition_helper evt1 evt2 j2_lt n_eq subtrace_tr1
      subtrace_tr3 as [ precedes | [ not_precedes ? ] ].
    left. auto. right. auto.

    (* invalid event. *)
    right. intros [ ? [ ? ? ] ]. tauto.
    right. intros [ ? [ ? ? ] ]. tauto.
  Qed.
End Traces.
