Require Import Metatheory.

Require Import actions.
Require Import drf_language_axioms.
Require Import indexes.
Require Import lambda_calculus.
Require Import lambda_calculus_locations_ok.
Require Import model_facts.
Require Import var_map.

Set Implicit Arguments.

(** This module guarantees that the lamba calculus satisfies the necessary
   properties for the DRF proof to go through. *)
Module Lambda_Calculus_DRF_Axioms <: DRF_LANGUAGE_AXIOMS.
  Module L := Lambda_Calculus.
  Module Typ := Locations_Ok.
  Module Pres := Locations_Ok_Preservation.
  Import Lambda_Calculus.
  Import Locations_Ok.
  Import Locations_Ok_Preservation.

  Module Facts := FMapFacts.Facts Map.
  Module VarMapFacts := VarMapFacts Map.
  Module LFacts := LanguageModelFacts L.

  (** If a thread allocates a fresh location, only that thread is able to read
     it immediately. *)
  Lemma well_formed_alloc_read : forall Sigma P id1 r1 v1 P1 id2 v2 P2,
    typing Sigma P ->
    ~ Map.In r1 Sigma ->
    program_step P (Some (id1, alloc r1 v1)) P1 ->
    program_step P1 (Some (id2, read r1 v2)) P2 ->
    id1 = id2.
  Proof.
    introv Typ Fr Step1 Step2.
    destruct* (id1 == id2).
    false.
    inversion Step1 as [ ? ? t1 ? t1' Maps1 Red1 | ]; subst.
    inversion Step2 as [ ? ? t2 ? t2' Maps2 Red2 | ]; subst.
    apply Map.add_3 in Maps2; auto.
    unfold typing in Typ.
    forwards Typ2 : Typ (>>> id2); eauto.
    clear Maps2 Step2.
    inductions Typ2 gen t2' end;
      introv Red2; inversions* Red2.
    inversion Typ2. unfold Map.In in Fr. eauto.
  Qed.

  (** If a thread allocates a fresh location, only that thread is able to write
     to it immediately. *)
  Lemma well_formed_alloc_write : forall Sigma P id1 r1 v1 P1 id2 v2 P2,
    typing Sigma P ->
    ~ Map.In r1 Sigma ->
    program_step P (Some (id1, alloc r1 v1)) P1 ->
    program_step P1 (Some (id2, write r1 v2)) P2 ->
    id1 = id2.
  Proof.
    introv Typ Fr Step1 Step2.
    destruct* (id1 == id2).
    false.
    inversion Step1 as [ ? ? t1 ? t1' Maps1 Red1 | ]; subst.
    inversion Step2 as [ ? ? t2 ? t2' Maps2 Red2 | ]; subst.
    apply Map.add_3 in Maps2; auto.
    unfold typing in Typ.
    forwards Typ2 : Typ (>>> id2); eauto.
    clear Maps2 Step2.
    inductions Typ2 gen t2' end;
      introv Red2; inversions* Red2.
    inversion Typ2_1. unfold Map.In in Fr. eauto.
  Qed.

  Inductive program_conflict :
    tid * action L.val -> tid * action L.val -> Prop :=
  | program_conflict_tid : forall t a1 a2,
    program_conflict (t, a1) (t, a2)
  | program_conflict_spawn_left : forall t1 t2 a2,
    program_conflict (t1, spawn _ t2) (t2, a2)
  | program_conflict_spawn_right : forall t1 a1 t2,
    program_conflict (t1, a1) (t2, spawn _ t1).

  (** Non-conflicting steps taken by different threads may be transposed. *)
  Lemma program_asynchrony : forall P P1 P2 e1 e2,
    program_step P (Some e1) P1 ->
    program_step P1 (Some e2) P2 ->
    ~ program_conflict e1 e2 ->
    exists P1',
      program_step P (Some e2) P1'
      /\ program_step P1' (Some e1) P2.
  Proof.
    intros P P1 P2 e1 e2 Step1 Step2 NotConflict.
    destruct e1 as [ id1 a1 ].
    destruct e2 as [ id2 a2 ].
    assert (id1 <> id2) as IDNeq12.
      intro. subst. apply NotConflict.
      apply program_conflict_tid.
    inversion Step1 as [ ? ? t1 ? t1' Maps1 Red1
      | ? ? t1 t1' id3 t3 Maps1 NotIn1 Red1 ];
      inversion Step2 as [ ? ? t2 ? t2' Maps2 Red2
        | ? ? t2 t2' id4 t4 Maps2 NotIn2 Red2 ];
        subst.

    (* Neither thread does a spawn. *)
    pose proof VarMapFacts.eq_add_add P t1' t2' IDNeq12 as Eq.
    apply Map.extensionality in Eq.
    esplit. split.
    apply* step_thread.
      rewrite* <- Facts.add_neq_mapsto_iff.
    rewrite Eq.
      apply* step_thread.
      rewrite* Facts.add_neq_mapsto_iff.

    (* Thread 2 does a spawn. *)
    assert (id1 <> id4) as IDNeq14.
      intro. subst. apply NotConflict.
      apply program_conflict_spawn_right.
    pose proof VarMapFacts.eq_add_add P t1' t2' IDNeq12 as Eq1.
    apply Map.extensionality in Eq1.
    pose proof VarMapFacts.eq_add_add (Map.add id2 t2' P) t1' t4 IDNeq14 as Eq2.
    apply Map.extensionality in Eq2.
    rewrite <- Eq1 in Eq2.
    esplit. split.
    apply* step_spawn.
      rewrite* <- Facts.add_neq_mapsto_iff.
      rewrite* <- Facts.add_neq_in_iff.
    rewrite Eq2.
      apply* step_thread.
        rewrite* Facts.add_neq_mapsto_iff.
        rewrite* Facts.add_neq_mapsto_iff.

    (* Thread 1 does a spawn. *)
    assert (id2 <> id3) as IDNeq23.
      intro. subst. apply NotConflict.
      apply program_conflict_spawn_left.
    pose proof VarMapFacts.eq_add_add (Map.add id1 t1' P) t2' t3 IDNeq23 as Eq1.
    apply Map.extensionality in Eq1.
    pose proof VarMapFacts.eq_add_add P t1' t2' IDNeq12 as Eq2.
    apply Map.extensionality in Eq2.
    rewrite Eq2 in Eq1.
    rewrite <- Eq1.
    esplit. split.
    apply* step_thread.
      rewrite* Facts.add_neq_mapsto_iff in Maps2.
      rewrite* Facts.add_neq_mapsto_iff in Maps2.
    apply* step_spawn.
      rewrite* Facts.add_neq_mapsto_iff.
      rewrite* Facts.add_neq_in_iff.

    (* Both threads do spawns. *)
    assert (id1 <> id4) as IDNeq14.
      intro. subst. apply NotConflict.
      apply program_conflict_spawn_right.
    assert (id2 <> id3) as IDNeq23.
      intro. subst. apply NotConflict.
      apply program_conflict_spawn_left.
    assert (id3 <> id4) as IDNeq34.
      intro. subst. apply NotIn2.
      unfold Map.In. exists t3.
      apply Map.find_2. rewrite* Facts.add_eq_o.
    pose proof VarMapFacts.eq_add_add (Map.add id1 t1' P) t2' t3 IDNeq23 as Eq1.
    apply Map.extensionality in Eq1.
    rewrite <- Eq1.
    pose proof VarMapFacts.eq_add_add P t1' t2' IDNeq12 as Eq2.
    apply Map.extensionality in Eq2.
    rewrite Eq2.
    pose proof VarMapFacts.eq_add_add (Map.add id1 t1' (Map.add id2 t2' P)) t3
      t4 IDNeq34 as Eq3.
    apply Map.extensionality in Eq3.
    rewrite Eq3.
    pose proof VarMapFacts.eq_add_add (Map.add id2 t2' P) t1' t4 IDNeq14 as Eq4.
    apply Map.extensionality in Eq4.
    rewrite Eq4.
    esplit. split.
    apply* step_spawn.
      rewrite* Facts.add_neq_mapsto_iff in Maps2.
      rewrite* Facts.add_neq_mapsto_iff in Maps2.
      intro. apply NotIn2.
      rewrite* Facts.add_neq_in_iff.
      rewrite* Facts.add_neq_in_iff.
    apply* step_spawn.
      rewrite* Facts.add_neq_mapsto_iff.
      rewrite* Facts.add_neq_mapsto_iff.
      rewrite* Facts.add_neq_in_iff.
      rewrite* Facts.add_neq_in_iff.
  Qed.

  Hint Constructors red.

  Ltac value_indep_ind H :=
    specializes* H (>>> __ __);
    destruct* H as [ ? ? ].

  (** The program does not care what value is returned by a read. *)
  Lemma value_independent_reads : forall P P' id r v v',
    program_step P (Some (id, read r v)) P' ->
    exists P'', program_step P (Some (id, read r v')) P''.
  Proof.
    introv Step.
    inversion Step as [ ? ? t ? t' Maps Red | ]; subst.
    assert (exists t'', t -- read r v' --> t'' ;; None) as [ ? ? ].
    clear Maps Step.
    remember (read r v) as a. remember (@None trm) as t_opt.
    induction Red; inversion Heqa; subst;
      try value_indep_ind IHRed.
    exists* v'.
    esplit. apply* step_thread.
  Qed.

  (** The set of thread IDs grows from step to step. *)
  Lemma program_step_many_preserves_tids : forall P1 P2 tr id1,
    program_step_many P1 P2 tr ->
    Map.In id1 P1 ->
    Map.In id1 P2.
  Proof.
    intros P1 P2 tr id1 Step In1.
    induction Step as [ | ? ? ? ? ? ? ? Step ]; auto.
    inversion Step as [ | ]; subst.
    rewrite Facts.add_in_iff. auto.
    repeat rewrite Facts.add_in_iff. auto.
  Qed.

  (** A spawn step always includes a spawned thread term. *)
  Lemma red_spawn_inversion : forall t1 id t1' t_opt,
    t1 -- (spawn _ id) --> t1' ;; t_opt ->
    exists t2, t_opt = Some t2.
  Proof.
    intros t1 id t1' t_opt Red.
    remember (spawn trm id) as a.
    induction Red; inversion Heqa; eauto.
  Qed.

  (** Inverting the spawn step for program states. *)
  Lemma program_step_spawn_inversion : forall P1 P2 id1 id2,
    program_step P1 (Some (id1, spawn _ id2)) P2 ->
    exists t1, exists t1', exists t2,
      Map.MapsTo id1 t1 P1 /\
      ~ Map.In id2 P1 /\
      t1 -- (spawn _ id2) --> t1' ;; (Some t2) /\
      P2 = Map.add id2 t2 (Map.add id1 t1' P1).
  Proof.
    intros P1 P2 id1 id2 Step.
    inversion Step as [ ? ? ? ? ? ? Red | ]; subst.
    pose proof red_spawn_inversion Red as [ ? false ].
    inversion false.
    repeat esplit; eauto.
  Qed.

  (** The ID generated for a spawned thread is always fresh. *)
  Lemma spawn_fresh : forall P1 P2 P3 tr id1 id2 j id3 a3,
    program_step_many P1 P2 tr ->
    program_step P2 (Some (id1, spawn _ id2)) P3 ->
    index tr j (Some (id3, a3)) ->
    id2 <> id3.
  Proof.
    intros P1 P2 P3 tr id1 id2 j id3 a3 StepMany Step Index Eq.
    subst id3.
    pose proof program_step_spawn_inversion Step
      as [ t1 [ t1' [ t2 [ Maps1 [ NotIn2 [ Red1 ] ] ] ] ] ].
    subst P3.
    apply NotIn2.
    pose proof LFacts.program_step_many_index_inversion StepMany Index
      as [ P5 [ P6 [ tr1 [ tr2 [ StepMany1 [ Step2 [ StepMany2 ? ] ] ] ] ] ] ].
    eapply program_step_many_preserves_tids.
    eapply LFacts.program_step_many_alt.
    eapply program_step_many_trans.
    apply program_step_many_refl.
    apply Step2. apply StepMany2.
    unfold Map.In. inversion Step2; eauto.
  Qed.

  Lemma well_formed_read : forall Sigma P1 t r v P2,
    Typ.typing Sigma P1 ->
    L.program_step P1 (Some (t, read r v)) P2 ->
    Map.In r Sigma.
  Proof.
    intros Sigma P1 t r v P2 Typing Step.
    inversion Step as [ ? id e ? e' PoolRead ThreadStep | ]; subst.
    unfold typing in Typing.
    specializes Typing PoolRead.
    eapply preservation_read; eauto.
  Qed.

  Lemma well_formed_write : forall Sigma P1 t r v P2,
    Typ.typing Sigma P1 ->
    L.program_step P1 (Some (t, write r v)) P2 ->
    Map.In r Sigma.
  Proof.
    intros Sigma P1 t r v P2 Typing Step.
    inversion Step as [ ? id e ? e' PoolRead ThreadStep | ]; subst.
    unfold typing in Typing.
    specializes Typing PoolRead.
    eapply preservation_write; eauto.
  Qed.
End Lambda_Calculus_DRF_Axioms.
