Require Import Arith.
Require Import Decidable.
Require Import List.
Require Import Setoid.

Set Implicit Arguments.

Section Lists.
  Variable A : Type.

  Lemma cons_nil_app : forall (a:A) l,
    a :: l = (a :: nil) ++ l.
  Proof.
    intros. simpl. auto.
  Qed.

  Lemma app_nil_begin : forall (l : list A),
    l = nil ++ l.
  Proof.
    simpl. auto.
  Qed.

  Lemma length_nil : forall (l : list A),
    l = nil <-> length l = 0.
  Proof.
    intros l.
    split.
    intros l_eq. subst. auto.
    intros length_l.
    induction l. auto. simpl in length_l. inversion length_l.
  Qed.

  Lemma length_cons : forall (l : list A) (n : nat),
    length l = S n ->
    exists a, exists l', l = a :: l' /\ length l' = n.
  Proof.
    intros l n length_l.
    induction l; simpl in length_l; inversion length_l; eauto.
  Qed.

  Lemma append_begin_unique : forall (l1 l2 l1' l2' : list A),
    l1 ++ l2 = l1' ++ l2' ->
    length l1 = length l1' ->
    l1 = l1' /\ l2 = l2'.
  Proof.
    intros l1 l2 l1' l2' l_eq length_eq.

    (* Proof by induction on l1. *)
    generalize dependent l1'.
    induction l1 as [ | a1 l1 IH ]; intros l1' l_eq' length_eq.

    (* l1 = nil *)
    simpl in length_eq. symmetry in length_eq.
    rewrite <- length_nil in length_eq. subst l1'.
    simpl in *. auto.

    (* l1 = a1 :: l1 *)
    simpl in *. symmetry in length_eq.
    pose proof length_cons _ length_eq
      as [ a1' [ l1'' [ l1'_eq length_eq' ] ] ].
    subst l1'. inversion l_eq'. subst a1.
    assert (l1 = l1'' /\ l2 = l2') as [ ? ? ]. eapply IH; eauto.
    subst. auto.
  Qed.

  Lemma append_end_unique : forall (l1 l2 l1' l2' : list A),
    l1 ++ l2 = l1' ++ l2' ->
    length l2 = length l2' ->
    l1 = l1' /\ l2 = l2'.
  Proof.
    intros l1 l2 l1' l2' l_eq length_eq.

    assert (length (l1 ++ l2) = length (l1' ++ l2')) as length_eq'.
    rewrite l_eq. auto.
    repeat rewrite app_length in length_eq'.
    rewrite length_eq in length_eq'.
    repeat rewrite (plus_comm _ (length l2')) in length_eq'.
    apply plus_reg_l in length_eq'.
    apply append_begin_unique. auto. auto.
  Qed.

  Lemma length_append_exists : forall (l : list A) m n,
    length l = m + n ->
    exists l1, exists l2,
      length l1 = m /\ length l2 = n /\ l = l1 ++ l2.
  Proof.
    intros l m n length_l.
    generalize dependent l.
    induction m; intros.
    exists (nil:list A). exists l. auto.
    induction l. inversion length_l.
    simpl in length_l. inversion length_l as [ length_l' ].
    pose proof IHm l length_l' as [ l1 [ l2 [ ? [ ] ] ] ].
    exists (a :: l1). exists l2.
    split. simpl. subst m. auto.
    split. auto. simpl. subst l. auto.
  Qed.

  Lemma length_append_exists_1 : forall (l : list A) n,
    length l = n + 1 ->
    exists l1, exists a,
      length l1 = n /\ l = l1 ++ (a :: nil).
  Proof.
    intros l n length_l.
    pose proof length_append_exists _ _ _ length_l
      as [ l1 [ l2 [ ? [ length_l2 ? ] ] ] ].
    destruct l2. inversion length_l2.
    inversion length_l2 as [ length_l2' ].
    rewrite <- length_nil in length_l2'. subst l2 l.
    eauto.
  Qed.

  Variable P : A -> Prop.
  Variable P_dec : forall a, decidable (P a).

  Lemma not_for_all_exists_not_list : forall L,
    ~ (forall a, In a L -> P a) ->
    exists a, In a L /\ ~ (P a).
  Proof.
    intros L not_all_P.
    induction L as [ | a L ].

    contradiction not_all_P.
    intros a in_nil.
    inversion in_nil.

    destruct (P_dec a).
    assert (exists a, In a L /\ ~ P a) as [ a' [ in_L not_a' ] ].
    apply IHL.
    intro all_P. apply not_all_P.
    intros a' in_L. destruct in_L; subst; auto.
    exists a'. simpl. auto.

    exists a. simpl. auto.
  Qed.
End Lists.
