Require Import Metatheory.

Require Import FMapFacts.

Require Import actions.
Require Import var_map.

Set Implicit Arguments.

Module LockMap.
  Definition lock_map := Map.t tid.

  Definition empty : lock_map := Map.empty tid.

  Definition available (Lambda : lock_map) (l : lock) : Prop :=
    Map.find l Lambda = None.

  Definition held_by (Lambda : lock_map) (l : lock) (t : tid) : Prop :=
    Map.find l Lambda = Some t.

  Definition acquire (Lambda : lock_map) (l : lock) (t : tid) : lock_map :=
    Map.add l t Lambda.

  Definition release (Lambda : lock_map) (l : lock) : lock_map :=
    Map.remove l Lambda.

  Definition eq (Lambda Lambda' : lock_map) : Prop :=
    Map.Equal Lambda Lambda'.

  Module MapFacts := FMapFacts.Facts Map.
  Module VarMapFacts := VarMapFacts Map.

  Lemma available_after_acquire : forall Lambda l l' t,
    available Lambda l ->
    l <> l' ->
    available (acquire Lambda l' t) l.
  Proof.
    intros Lambda l l' t avail_before l_neq.
    unfold available, acquire in *.
    rewrite MapFacts.add_neq_o; auto.
  Qed.

  Lemma available_before_acquire : forall Lambda l l' t,
    available (acquire Lambda l' t) l ->
    l <> l' ->
    available Lambda l.
  Proof.
    intros Lambda l l' t avail_after l_neq.
    unfold available, acquire in *.
    erewrite <- MapFacts.add_neq_o; eauto.
  Qed.

  Lemma not_available_after_acquire : forall Lambda l t,
    ~ available (acquire Lambda l t) l.
  Proof.
    intros Lambda l t avail.
    unfold available, acquire in avail.
    assert (Map.MapsTo l t (Map.add l t Lambda)) as maps_to.
    apply Map.add_1. unfold Var_as_OT.eq. auto.
    apply Map.find_1 in maps_to.
    rewrite avail in maps_to. inversion maps_to.
  Qed.

  Lemma held_by_after_acquire : forall Lambda l l' t t',
    held_by Lambda l t ->
    l <> l' ->
    held_by (acquire Lambda l' t') l t.
  Proof.
    intros Lambda l l' t t' held_before l_neq.
    unfold held_by, acquire in *.
    rewrite MapFacts.add_neq_o; auto.
  Qed.

  Lemma held_by_before_acquire : forall Lambda l l' t t',
    held_by (acquire Lambda l' t') l t ->
    l <> l' ->
    held_by Lambda l t.
  Proof.
    intros Lambda l l' t t' held_after l_neq.
    unfold held_by, acquire in *.
    erewrite <- MapFacts.add_neq_o; eauto.
  Qed.

  Lemma not_held_by_after_acquire : forall Lambda l t t',
    t <> t' ->
    ~ held_by (acquire Lambda l t) l t'.
  Proof.
    intros Lambda l t t' t_neq held.
    unfold held_by, acquire in held.
    assert (Map.MapsTo l t (Map.add l t Lambda)) as maps_to.
    apply Map.add_1. unfold Var_as_OT.eq. auto.
    apply Map.find_1 in maps_to.
    rewrite held in maps_to. inversion maps_to.
    apply t_neq. auto.
  Qed.

  Lemma not_held_by_after_release : forall Lambda l t,
    ~ held_by (release Lambda l) l t.
  Proof.
    intros Lambda l t held.
    unfold held_by, release in held.
    apply Map.find_2 in held.
    contradiction ((@Map.remove_1 tid) Lambda l l (Var_as_OT.eq_refl l)).
    unfold Map.In. eauto.
  Qed.

  Lemma available_after_release : forall Lambda l l',
    available Lambda l ->
    l <> l' ->
    available (release Lambda l') l.
  Proof.
    intros Lambda l l' avail_before l_neq.
    unfold available, release in *.
    rewrite MapFacts.remove_neq_o; auto.
  Qed.

  Lemma available_before_release : forall Lambda l l',
    available (release Lambda l') l ->
    l <> l' ->
    available Lambda l.
  Proof.
    intros Lambda l l' avail_after l_neq.
    unfold available, release in *.
    erewrite <- MapFacts.remove_neq_o; eauto.
  Qed.

  Lemma held_by_after_release : forall Lambda l l' t,
    held_by Lambda l t ->
    l <> l' ->
    held_by (release Lambda l') l t.
  Proof.
    intros Lambda l l' t held_before l_neq.
    unfold held_by, release in *.
    rewrite MapFacts.remove_neq_o; auto.
  Qed.

  Lemma held_by_before_release : forall Lambda l l' t,
    held_by (release Lambda l') l t ->
    l <> l' ->
    held_by Lambda l t.
  Proof.
    intros Lambda l l' t held_after l_neq.
    unfold held_by, release in *.
    erewrite <- MapFacts.remove_neq_o; eauto.
  Qed.

  Lemma eq_acquire_acquire : forall Lambda l l' t t',
    l <> l' ->
    acquire (acquire Lambda l t) l' t' = acquire (acquire Lambda l' t') l t.
  Proof.
    intros.
    unfold acquire.
    apply Map.extensionality.
    auto with var_maps.
  Qed.

  Lemma eq_acquire_release : forall Lambda l l' t',
    l <> l' ->
    acquire (release Lambda l) l' t' = release (acquire Lambda l' t') l.
  Proof.
    intros.
    unfold acquire, release.
    apply Map.extensionality.
    auto with var_maps.
  Qed.

  Lemma eq_release_acquire : forall Lambda l l' t,
    l <> l' ->
    release (acquire Lambda l t) l' = acquire (release Lambda l') l t.
  Proof.
    intros.
    unfold acquire, release.
    apply Map.extensionality.
    auto with var_maps.
  Qed.

  Lemma eq_release_release : forall Lambda l l',
    l <> l' ->
    release (release Lambda l) l' = release (release Lambda l') l.
  Proof.
    intros.
    unfold release.
    apply Map.extensionality.
    auto with var_maps.
  Qed.

  Hint Resolve available_before_acquire available_after_acquire
    held_by_before_acquire held_by_after_acquire available_before_release
    available_after_release held_by_before_release held_by_after_release
    eq_acquire_acquire eq_acquire_release eq_release_acquire eq_release_release
    : locks.
End LockMap.
