(** This file implements a semantics called MSIOwner that is an intermediate
   between MSI and SC. *)

Require Import Metatheory.

Require Import maps.
Require Import models.

Module MSIOwner (L : LANGUAGE_MODEL) <: MEMORY_MODEL L.
  Import VarMap.

  Inductive ownership_state :=
  | exclusive : tid -> ownership_state
  | shared : t unit -> ownership_state.

  (* heap consists of an ownership store *)
  Definition heap := t (ownership_state * L.val).

  (* the empty heap *)
  Definition empty_heap : heap := empty _.

  (* heap steps - obey the ownership table. *)
  Inductive step : heap -> effect L.val -> heap -> Prop :=
  | step_alloc : forall S t r v,
    ~ In r S ->
    step S (Some (t, alloc r v)) (add r (shared (empty _), v) S)
  | step_read_shared : forall S t r v tids,
    MapsTo r (shared tids, v) S ->
    In t tids ->
    step S (Some (t, read r v)) S
  | step_read_modified : forall S t r v,
    MapsTo r (exclusive t, v) S ->
    step S (Some (t, read r v)) S
  | step_write_modified : forall S t r v v',
    MapsTo r (exclusive t, v) S ->
    step S (Some (t, write r v')) (add r (exclusive t, v') S)
  | step_invalid_to_shared : forall S t r v tids,
    MapsTo r (shared tids, v) S ->
    ~ In t tids ->
    step S None (add r (shared (add t tt tids), v) S)
  | step_invalid_to_modified : forall S t r v tids,
    MapsTo r (shared tids, v) S ->
    Empty tids ->
    step S None (add r (exclusive t, v) S)
  | step_shared_to_invalid : forall S t r v tids,
    MapsTo r (shared tids, v) S ->
    In t tids ->
    step S None (add r (shared (remove t tids), v) S)
  | step_modified_to_invalid : forall S t r v,
    MapsTo r (exclusive t, v) S ->
    step S None (add r (shared (empty _), v) S)
  | step_spawn : forall S t1 t2,
    step S (Some (t1, spawn _ t2)) S.

  Definition heap_step := step.

  Inductive heap_step_many : heap -> heap -> trace L.val -> Prop :=
  | heap_step_many_refl : forall H,
    heap_step_many H H nil
  | heap_step_many_trans : forall H1 H2 H3 tr e,
    heap_step_many H1 H2 tr ->
    heap_step H2 e H3 ->
    heap_step_many H1 H3 (e :: tr).

  Lemma no_pure_steps : forall H H' t,
    ~ (heap_step H (Some (t, pure _)) H').
  Proof.
    intros H H' t' Step. inversion Step.
  Qed.
End MSIOwner.
