Require Import Metatheory.

Require Import maps.
Require Import model_facts.
Require Import models.
Require Import msi_owner.
Require Import sc.

Module MSIToSCOwner (L : LANGUAGE_MODEL).
  Module SC := Sequential_Consistency L.
  Module MSIO := MSIOwner L.
  Declare Module SysSC : SYSTEM_MODEL L SC.
  Declare Module SysMSIO : SYSTEM_MODEL L MSIO.
  Module SCFacts := ModelFacts L SC SysSC.

  Inductive map_heap :  MSIO.heap -> SC.heap ->Prop :=
  | map_heap_def : forall S1 S2,
    (forall r, ~ VarMap.In r S1 -> ~ VarMap.In r S2) ->
    (forall r o v, VarMap.MapsTo r (o, v) S1 -> VarMap.MapsTo r v S2) ->
    map_heap S1 S2.

  Lemma map_heap_empty :
    map_heap MSIO.empty_heap SC.empty_heap.
  Proof.
    apply map_heap_def.
    intros r NotIn false.
    rewrite* VarMapFacts.empty_in_iff in false.
    intros r o v false.
    rewrite* VarMapFacts.empty_mapsto_iff in false.
  Qed.

  Lemma alloc_step : forall H1 H1' H2 t r v,
    map_heap H1 H2 ->
    MSIO.heap_step H1 (Some (t, alloc r v)) H1' ->
    exists H2',
      SC.heap_step H2 (Some (t, alloc r v)) H2'
      /\ map_heap H1' H2'.
  Proof.
    intros H1 H1' H2 t r v Map Step.
    inversion Step. subst.
    inversion Map. subst.
    esplit. split.

    (* Do alloc. *)
    apply SC.step_alloc. eauto.

    (* Map the new MSIOwner heap to the new SC heap *)
    apply map_heap_def.

      (* Unallocated locations are also unallocated in SC. *)
      intros r' NotIn.
      assert (r <> r').
        intro. subst. apply NotIn.
        esplit. apply VarMap.find_2.
        rewrite* VarMapFacts.add_eq_o.
      rewrite* VarMapFacts.add_neq_in_iff.
      rewrite* VarMapFacts.add_neq_in_iff in NotIn.

      (* Allocated locations are allocated with the same value in SC. *)
      intros r' o v' Maps1.
      destruct (r == r'). subst.

        (* Case: the location being allocated. *)
        assert ((o, v') = (MSIO.shared (VarMap.empty unit), v)) as Eq.
        eapply VarMapFacts.MapsTo_fun.
        apply Maps1. apply VarMap.find_2. rewrite* VarMapFacts.add_eq_o.
        inversion Eq. subst.
        apply VarMap.find_2. rewrite* VarMapFacts.add_eq_o.

        (* Case: the other locations. *)
        rewrite* VarMapFacts.add_neq_mapsto_iff in Maps1.
        apply VarMap.find_2.
        rewrite* VarMapFacts.add_neq_o.
        apply VarMap.find_1. eauto.
  Qed.

  Lemma read_step : forall H1 H1' H2 t r v,
    map_heap H1 H2 ->
    MSIO.heap_step H1 (Some (t, read r v)) H1' ->
    exists H2',
      SC.heap_step H2 (Some (t, read r v)) H2'
      /\ map_heap H1' H2'.
  Proof.
    intros H1 H1' H2 t r v Map Step.
    exists H2.
    inversion Map. subst.
    inversion Step;
      ( subst; split; [ apply SC.step_read; eauto | auto ] ).
  Qed.

  Lemma write_step : forall H1 H1' H2 t r v,
    map_heap H1 H2 ->
    MSIO.heap_step H1 (Some (t, write r v)) H1' ->
    exists H2',
      SC.heap_step H2 (Some (t, write r v)) H2'
      /\ map_heap H1' H2'.
  Proof.
    intros H1 H1' H2 t r v Map Step.
    inversion Step. subst.
    inversion Map. subst.
    esplit. split.

    apply SC.step_write.
    esplit. eauto.

    (* Map the new MSIOwner heap to the new SC heap *)
    apply map_heap_def.

      (* Unallocated locations are also unallocated in SC. *)
      intros r' NotIn In.
      assert (r <> r'). intro. subst. apply NotIn. esplit.
        apply VarMap.find_2.
        rewrite* VarMapFacts.add_eq_o.
      rewrite* VarMapFacts.add_neq_in_iff in In.
      rewrite* VarMapFacts.add_neq_in_iff in NotIn.

      (* Allocated locations are allocated with the same value in SC. *)
      intros r' o v' Maps1.
      destruct (r == r'). subst.

        (* Case: the location being written. *)
        assert ((o, v') = (MSIO.exclusive t, v)) as Eq.
        eapply VarMapFacts.MapsTo_fun. eauto.
        apply VarMap.find_2.
        rewrite* VarMapFacts.add_eq_o.
        inversion Eq. subst.
        apply VarMap.find_2.
        rewrite* VarMapFacts.add_eq_o.

        (* Case: the other locations. *)
        apply VarMap.find_2.
        rewrite* VarMapFacts.add_neq_o.
        apply VarMap.find_1 in Maps1.
        rewrite* VarMapFacts.add_neq_o in Maps1.
        apply VarMap.find_2 in Maps1.
        apply VarMap.find_1. eauto.
  Qed.

  Lemma change_ownership : forall H1 H2 o r v o',
    map_heap H1 H2 ->
    VarMap.MapsTo r (o, v) H1 ->
    map_heap (VarMap.add r (o', v) H1) H2.
  Proof.
    intros H1 H2 o r v o' Map Maps.
    inversion Map as [ ? ? Fresh ]. subst.
    apply map_heap_def.

    (* Unallocated locations are also unallocated in SC. *)
    intros r' NotIn.
    assert (r <> r'). intro. apply NotIn.
      rewrite* VarMapFacts.add_in_iff.
    apply Fresh. intros [ v' Maps' ].
    apply NotIn. exists v'.
    apply VarMap.find_2.
    rewrite* VarMapFacts.add_neq_o.
    apply VarMap.find_1. eauto.

    (* Allocated locations are allocated with the same value in SC. *)
    intros r' o'' v' Maps'.
    destruct (r == r'). subst.

      (* Case: location whose ownership was changed. *)
      assert ((o'', v') = (o', v)) as Eq.
      eapply VarMapFacts.MapsTo_fun. eauto.
      apply VarMap.find_2.
      rewrite* VarMapFacts.add_eq_o.
      inversion Eq. subst. eauto.

      (* Case: other locations. *)
      apply VarMap.find_1 in Maps'.
      rewrite* VarMapFacts.add_neq_o in Maps'.
      apply VarMap.find_2 in Maps'. eauto.
  Qed.

  Lemma empty_step : forall H1 H1' H2,
    map_heap H1 H2 ->
    MSIO.heap_step H1 None H1' ->
    map_heap H1' H2.
  Proof.
    intros H1 H1' H2 Map Step.
    inversion Step; subst;
      eapply change_ownership; eauto.
  Qed.

  Ltac esplitt := esplit; esplit; split; [ | split ].

  Ltac step_many_one :=
    eapply SC.heap_step_many_trans;
      [ apply SC.heap_step_many_refl | ].

  Lemma msi_owner_to_sc_step : forall H1 H1' H2 e,
    map_heap H1 H2 ->
    MSIO.heap_step H1 e H1' ->
    exists H2', exists tr,
      SC.heap_step_many H2 H2' tr /\
      erase_trace tr = erase_trace (e :: nil) /\
      map_heap H1' H2'.
  Proof with eauto.
    intros H1 H1' H2 e MapHeap Step.
    destruct e as [ [ t a ] | ].
    inversion Step; subst.

    forwards [ ? [ ? ? ] ] : alloc_step...
    esplitt.
    step_many_one...
    auto. auto.

    forwards [ ? [ ? ? ] ] : read_step...
    esplitt.
    step_many_one...
    auto. auto.

    forwards [ ? [ ? ? ] ] : read_step...
    esplitt.
    step_many_one...
    auto. auto.

    forwards [ ? [ ? ? ] ] : write_step...
    esplitt.
    step_many_one...
    auto. auto.

    esplitt.
    step_many_one...
    apply SC.step_spawn.
    auto. auto.

    esplitt.
    apply SC.heap_step_many_refl.
    auto.
    inversion Step;
      eapply change_ownership...
  Qed.

  Lemma msi_owner_to_sc_step_many : forall H1 H2 P H1' P' tr,
    map_heap H1 H2 ->
    SysMSIO.system_step_many H1 P H1' P' tr ->
    exists H2', exists tr',
      SysSC.system_step_many H2 P H2' P' tr'
      /\ map_heap H1' H2'.
  Proof with eauto.
    intros H1 H2 P H1' P' tr MapHeap StepMany.
    induction StepMany as [ | ? ? ? ? ? ? ? ? ? ? Step ].
    esplit. esplit. split.
    apply SysSC.system_step_many_refl.
    auto.
    specializes IHStepMany (>>> __)...
    destruct IHStepMany as [ ? [ ? [ StepMany' ? ] ] ].
    inversion Step; subst.

    forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : msi_owner_to_sc_step...
    esplit. esplit. split.
    eapply SCFacts.system_step_many_alt...
    eapply SCFacts.system_step_many_mutual...
    auto.

    forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : msi_owner_to_sc_step...
    esplit. esplit. split.
    eapply SCFacts.system_step_many_alt.
    apply StepMany'.
    eapply SCFacts.system_step_many_empty...
    auto.

    esplit. esplit. split.
    eapply SysSC.system_step_many_trans...
    apply SysSC.system_step_program...
    auto.
  Qed.

  Theorem msi_owner_to_sc : forall P0 H1 P tr,
    SysMSIO.system_step_many MSIO.empty_heap P0 H1 P tr ->
    exists H2, exists tr',
      SysSC.system_step_many SC.empty_heap P H2 P tr'.
  Proof with eauto.
    intros P0 H1 P tr StepMany.
    forwards [ ? [ ? [ ? ? ] ] ] : msi_owner_to_sc_step_many...
    apply map_heap_empty.
  Qed.
End MSIToSCOwner.
