(** We prove our key result: DRF programs have sequentially-consistent semantics
   when run on a write-buffering heap. *)

Require Import coherent.
Require Import init_wb.
Require Import happens_before.
Require Import heap_reduce.
Require Import heap_relation.
Require Import modulo_values.
Require Import preservation.
Require Import sc_facts.
Require Import wb_facts.
Require Import sc_type_preservation.
Require Import wb_type_preservation.

Open Scope list_scope.

Set Implicit Arguments.

Module SC_Sys_Pres := System_Model_Type_Preservation Pres SC_Type_Preservation
  L_SC.
Module WB_Sys_Pres := System_Model_Type_Preservation Pres WB_Type_Preservation
  L_WB.

Hint Constructors L_SC.system_step L_SC.system_step_many L_WB.system_step
  L_WB.system_step_many SC.step SC.heap_step_many WB.step WB.heap_step_many.

(** We define a bisimulation between write-buffering executions and
   sequentially-consistent executions. For each non-empty step taken by the WB
   semantics, the SC semantics takes the same non-empty step. For any empty
   step taken by the WB semantics, the SC semantics takes zero steps. At each
   point in the execution, the WB and SC heaps must be related by the heap-
   reduce relation. *)
Inductive bisimulation : L.program_state -> WB.heap -> SC.heap
  -> L.program_state -> trace L.val -> Prop :=
| bisimulation_refl : forall P,
  bisimulation P WB.empty_heap SC.empty_heap P nil
| bisimulation_empty : forall P0 P H_wb H_wb' H_sc tr,
  bisimulation P0 H_wb H_sc P tr ->
  L_WB.system_step H_wb P None H_wb' P ->
  heap_reduce H_wb' H_sc ->
  bisimulation P0 H_wb' H_sc P tr
| bisimulation_nonempty : forall P0 P P' H_wb H_wb' H_sc H_sc' tr t a,
  bisimulation P0 H_wb H_sc P tr ->
  L_WB.system_step H_wb P (Some (t, a)) H_wb' P' ->
  L_SC.system_step H_sc P (Some (t, a)) H_sc' P' ->
  heap_reduce H_wb' H_sc' ->
  bisimulation P0 H_wb' H_sc' P' (Some (t, a) :: tr).

Lemma bisimulation_implies_heap_reduce : forall P0 H_wb H_sc P tr,
  bisimulation P0 H_wb H_sc P tr ->
  heap_reduce H_wb H_sc.
Proof.
  intros P0 H_wb H_sc P tr bisim.
  inversion bisim; auto using heap_reduce_empty.
Qed.

Lemma bisimulation_implies_WB_step_many : forall P0 H_wb H_sc P tr,
  bisimulation P0 H_wb H_sc P tr ->
  exists tr',
    erase_trace tr = erase_trace tr' /\
    L_WB.system_step_many WB.empty_heap P0 H_wb P tr'.
Proof.
  intros P0 H_wb H_sc P tr bisim.
  induction bisim as [ | | ? ? ? ? ? ? ? ? t a ].
  eauto.
  destruct IHbisim as [ tr' [ ? ? ] ].
  exists (None :: tr'). eauto.
  destruct IHbisim as [ tr' [ ? ? ] ].
  exists (Some (t, a) :: tr'). simpl.
  split. f_equal. auto. eauto.
Qed.

Lemma bisimulation_implies_SC_step_many : forall P0 H_wb H_sc P tr,
  bisimulation P0 H_wb H_sc P tr ->
  L_SC.system_step_many SC.empty_heap P0 H_sc P tr.
Proof.
  intros P0 H_wb H_sc P tr bisim.
  induction bisim; eauto.
Qed.

(* A tactic we use in the bisimulation proof. *)
Ltac split_cut H :=
  match goal with
    | |- ?A /\ _ => cut A; [ intro H; split; [ apply H | idtac ] | idtac ]
  end.

(** If a DRF program takes a non-empty step in the WB semantics, and the
   execution up to this point has been a bisimulation, then the SC semantics
   also takes a step (continuing the bisimulation). *)
Lemma nonempty_step_preserves_bisimulation :
  forall P0 P P' H_sc H_wb H_wb' tr t a,
    Typ.typing Typ.empty_heap_typing P0 ->
    DRF.race_free_program P0 ->
    bisimulation P0 H_wb H_sc P tr ->
    coherent_heap H_wb ->
    L_WB.system_step H_wb P (Some (t, a)) H_wb' P' ->
    coherent_heap H_wb'/\
    exists H_sc',
      L_SC.system_step H_sc P (Some (t, a)) H_sc' P' /\
      bisimulation P0 H_wb' H_sc' P' (Some (t, a) :: tr).
Proof.
  intros P0 P P' H_sc H_wb H_wb' tr_sc t a typ drf bisim coherent steps.
  pose proof (bisimulation_implies_heap_reduce bisim) as reduce.
  pose proof (bisimulation_implies_WB_step_many bisim)
    as [ tr_wb [ erase_eq steps_many_WB ] ].
  pose proof (bisimulation_implies_SC_step_many bisim) as steps_many_SC.
  remember drf as drf2. clear Heqdrf2.
  pose proof WB_Sys_Pres.preservation_init typ steps_many_WB
    as [ Sigma [ typing_H_wb typing_H_sc ] ].

  (* The WB semantics took a nonempty step. What kind of step did it take? *)
  inversion steps as [ ? ? ? ? ? ? program_steps heap_steps | | ]; subst.
  inversion heap_steps as
    [ S_wb B_wb Lambda_wb ? r v fresh_S_wb (* alloc *)
      | S_wb B_wb Lambda_wb ? r v empty_B get_S (* read *)
      | S_wb B_wb Lambda_wb ? r v (* write *)
      | (* commit *)
      | S_wb B_wb Lambda_wb ? l available_l (* acquire *)
      | S_wb B_wb Lambda_wb ? l held_by_t_l is_empty_B_t (* release *)
      | S_wb B_wb Lambda_wb t' is_empty_B_t (* spawn *)
      | S_wb B_wb Lambda_wb ? r v top_B_t_r
    ]; subst.

  (* The step taken was an alloc. Basically we just have to show that the SC
     semantics can do the same alloc (which it can, assuming the domains of the
     stores are identical) and that heap coherency is preserved (which it is,
     because the buffers are the same for all the old locations). *)
  split. eapply alloc_preserves_coherent; eauto.
  destruct H_sc as [ S_sc Lambda_sc ].
  exists (SC.Store.update S_sc r v, Lambda_sc).
  split_cut steps_SC.
  inversion steps_SC. subst.
  eapply bisimulation_nonempty; eauto.
  eapply alloc_preserves_heap_reduce; eauto.
  apply L_SC.system_step_mutual; auto.
  apply SC.step_alloc.
  eapply heap_reduce_implies_fresh; eauto.

  (* The step taken was a read. The logic here is somewhat convoluted:
     basically, we know that since the WB semantics did a read, the SC semantics
     must be able to do a read as well (although we don't know exactly which
     value it sees in the heap). Because P0 is DRF, the new trace (including
     the read) must be well-synchronized, which in turn implies that there are
     no writes buffered for that location in the WB semantics, which in turn
     implies that the SC semantics did indeed read the same value (because the
     two heaps are related by the heap relation). *)
  split. auto.
  exists H_sc. split_cut steps_SC.
  inversion steps_SC; subst.
  eapply bisimulation_nonempty; eauto.
  apply L_SC.system_step_mutual; auto.
  destruct H_sc as [ S_sc Lambda_sc ].
  apply SC.step_read.
  eapply empty_buffer_heap_reduce_implies_get; eauto.
  pose proof heap_reduce_implies_get_exists reduce get_S as [ v' get_S_sc ].
  pose proof L_Axioms.value_independent_reads v' program_steps
    as [ P'' program_steps' ].
  forwards except : well_synchronized_trace_with_read_implies_empty_except.
  apply steps_many_WB.
  apply steps.
  apply equal_modulo_values_well_synchronized
    with (tr1 := (t, read r v') :: erase_trace tr_wb).
  eapply drf_implies_well_synchronized.
  apply typ. apply drf.
  rewrite <- erase_eq. simpl.
  rewrite <- sequential_trace_erase.
  eapply L_SC.system_step_many_trans. eauto.
  eapply L_SC.system_step_mutual. eauto.
  apply* SC.step_read.
  apply* step_many_sequential_trace.
  eapply equal_modulo_values_cons.
  apply equal_modulo_values_refl.
  apply actions_equal_modulo_values_read.
  intros t'.
  destruct (t == t').
  subst. auto.
  unfold empty_except in except.
  apply except. auto.

  (* The WB semantics did a write. There are few logical steps, again.  It's not
     hard to show that the SC semantics takes a step (we simply update the
     store). Given that, it must be the case that only the writing thread had
     any values buffered, so the new buffer is also coherent.  Moreover, the
     heap relation is preserved because both heaps are updated to reflect the
     new value. *)
  destruct H_sc as [ S_sc Lambda_sc ].
  pose proof L_SC.system_step_mutual program_steps
    (SC.step_write S_sc Lambda_sc t r v) as steps_SC.
  pose proof L_SC.system_step_many_trans steps_many_SC steps_SC
    as steps_many_SC'.
  assert (empty_except B_wb t r) as except.
  eapply well_synchronized_trace_with_write_implies_empty_except.
  eauto. eauto. eapply drf_implies_well_synchronized.
  apply typ. apply drf.
  rewrite <- erase_eq. simpl. rewrite <- sequential_trace_erase.
  eauto. eapply step_many_sequential_trace; eauto.
  split. eapply write_preserves_coherent; eauto.
  esplit. split. eauto.
  eapply bisimulation_nonempty; eauto.
  eapply write_preserves_heap_reduce; eauto.
  apply SC.step_write.

  (* The WB semantics did an acquire. This doesn't affect the coherency of the
     heap and is therefore straightforward. *)
  split.
  apply coherent_independent_of_lock_map with (Lambda := Lambda_wb);
    eauto.
  destruct H_sc as [ S_sc Lambda_sc ].
  pose proof heap_reduce_implies_lock_maps_equal reduce; subst.
  exists (S_sc, LockMap.acquire Lambda_sc l t).
  split_cut steps_SC.
  eapply bisimulation_nonempty; eauto.
  eapply heap_reduce_independent_of_lock_map; eauto.
  eapply L_SC.system_step_mutual; eauto.
  apply SC.step_acquire. eauto.

  (* The WB semantics did a release. Again, this is straightforward. *)
  split.
  apply coherent_independent_of_lock_map with (Lambda := Lambda_wb);
    eauto.
  destruct H_sc as [ S_sc Lambda_sc ].
  pose proof heap_reduce_implies_lock_maps_equal reduce; subst.
  exists (S_sc, LockMap.release Lambda_sc l).
  split_cut steps_SC.
  eapply bisimulation_nonempty; eauto.
  eapply heap_reduce_independent_of_lock_map; eauto.
  eapply L_SC.system_step_mutual; eauto.
  apply SC.step_release. eauto.

  (* The WB semantics did a spawn.  Straightforward again. *)
  split. auto.
  exists H_sc. split_cut steps_SC.
  eapply bisimulation_nonempty; eauto.
  eapply L_SC.system_step_mutual; eauto.
  apply SC.step_spawn.

  (* The step taken was a read (from the buffer, not the store). *)
  split. auto.
  exists H_sc. split_cut steps_SC.
  inversion steps_SC; subst.
  eapply bisimulation_nonempty; eauto.
  apply L_SC.system_step_mutual; auto.
  destruct H_sc as [ S_sc Lambda_sc ].
  apply SC.step_read.
  eapply top_heap_reduce_implies_get; eauto.

  (* pure step *)
  split. auto.
  esplit. split. eapply L_SC.system_step_program. auto.
  eapply bisimulation_nonempty; eauto.
Qed.

(** The lemma above, but for many steps. *)
Lemma step_many_preserves_bisimulation :
  forall P0 P P' H_sc H_wb H_wb' tr tr_wb,
    Typ.typing Typ.empty_heap_typing P0 ->
    DRF.race_free_program P0 ->
    bisimulation P0 H_wb H_sc P tr ->
    coherent_heap H_wb ->
    L_WB.system_step_many H_wb P H_wb' P' tr_wb ->
    coherent_heap H_wb'/\
    exists H_sc', exists tr_sc,
      L_SC.system_step_many H_sc P H_sc' P' tr_sc /\
      bisimulation P0 H_wb' H_sc' P' (tr_sc ++ tr).
Proof.
  intros P0 P P' H_sc H_wb H_wb' tr tr_wb typ drf bisim coherent steps_many.

  (* Proof by induction on the number of steps taken by the WB semantics. *)
  induction steps_many
    as [ | H_wb P H_wb'' P'' H_wb' P' tr_wb e steps_many ? steps ].

  (* Base case: zero steps. *)
  split. auto.
  exists H_sc. exists (@nil (effect L.val)).
  simpl. auto.

  (* Inductive case. Was the step taken empty or non-empty? *)
  pose proof IHsteps_many bisim coherent
    as [ coherent'' [ H_sc'' [ tr_sc [ steps_many_SC bisim'' ] ] ] ].
  induction e as [ [ t a ] | ].

  (* The most recent step was a nonempty step. *)
  pose proof nonempty_step_preserves_bisimulation typ drf bisim'' coherent''
    steps as [ coherent' [ H_sc' [ steps_SC' bisim' ] ] ].
  split. auto.
  exists H_sc'. exists (Some (t, a) :: tr_sc).
  simpl. eauto.

  (* The most recent step was an empty step. The program state does not
     change. *)
  inversion steps. subst.
  split.
  eapply empty_step_preserves_coherent. eauto. auto.
  exists H_sc''. exists tr_sc.
  split. auto.
  eapply bisimulation_empty. eauto. auto.
  eapply empty_step_preserves_heap_reduce.
  eapply bisimulation_implies_heap_reduce. eauto. auto.
Qed.

(* The main theorem. Any valid execution in the WB semantics is also valid
   in the SC semantics. *)
Theorem safety : forall H_wb P0 P tr_wb,
  Typ.typing Typ.empty_heap_typing P0 ->
  DRF.race_free_program P0 ->
  L_WB.system_step_many WB.empty_heap P0 H_wb P tr_wb ->
  exists H_sc, exists tr_sc,
    L_SC.system_step_many SC.empty_heap P0 H_sc P tr_sc.
Proof.
  intros H_wb P0 P tr_wb typ drf steps_many.
  pose proof bisimulation_refl P0 as bisim.
  pose proof step_many_preserves_bisimulation typ drf bisim coherent_empty
    steps_many as [ ? [ ? [ ? [ ? ? ] ] ] ].
  eauto.
Qed.
