Require Import Metatheory.

Require Import FMapFacts.

Require Import models.
Require Import var_map.

Set Implicit Arguments.

Module Type STORE (L : LANGUAGE_MODEL).
  Parameter store : Type.
  Parameter empty : store.
  Parameter fresh : store -> loc -> Prop.
  Parameter get : store -> loc -> L.val -> Prop.
  Parameter update : store -> loc -> L.val -> store.

  Parameter get_unique : forall S r v v',
    get S r v ->
    get S r v' ->
    v = v'.

  Parameter get_update_same : forall S r v,
    get (update S r v) r v.

  Parameter get_update_different : forall S r r' v v',
    get S r v ->
    exists v'', get (update S r' v') r v''.

  Parameter get_before_update : forall S r r' v v',
    get (update S r' v') r v ->
    r <> r' ->
    get S r v.

  Parameter get_after_update : forall S r r' v v',
    get S r v ->
    r <> r' ->
    get (update S r' v') r v.

  Parameter fresh_before_update : forall S r r' v,
    fresh (update S r' v) r ->
    fresh S r.

  Parameter fresh_after_update : forall S r r' v,
    fresh S r ->
    r <> r' ->
    fresh (update S r' v) r.

  Parameter not_fresh_after_update : forall S r v,
    ~ fresh (update S r v) r.

  Parameter not_fresh_after_get : forall S r v,
    get S r v ->
    ~ fresh S r.

  Parameter eq_update_update : forall S r r' v v',
    r <> r' ->
    update (update S r v) r' v' = update (update S r' v') r v.

  Parameter fresh_empty : forall r,
    fresh empty r.

  Hint Resolve get_update_same get_update_different get_before_update
    get_after_update fresh_before_update fresh_after_update eq_update_update
    fresh_empty : stores.
End STORE.

Module Store (L : LANGUAGE_MODEL) : STORE L.
  Definition store := Map.t L.val.

  Definition empty : store := Map.empty L.val.

  Definition fresh (S : store) (r : loc) : Prop :=
    ~ Map.In r S.

  Definition get (S : store) (r : loc) (v : L.val) : Prop :=
    Map.MapsTo r v S.

  Definition update (S : store) (r : loc) (v : L.val) : store :=
    Map.add r v S.

  Module MapFacts := FMapFacts.Facts Map.
  Module VarMapFacts := VarMapFacts Map.

  Lemma get_unique : forall S r v v',
    get S r v ->
    get S r v' ->
    v = v'.
  Proof.
    intros S r v v' get1 get2.
    unfold get in *.
    apply Map.find_1 in get1.
    apply Map.find_1 in get2.
    rewrite get1 in get2.
    inversion get2.
    auto.
  Qed.

  Lemma get_update_same : forall S r v,
    get (update S r v) r v.
  Proof.
    intros S r v.
    unfold get, update.
    pose proof MapFacts.add_mapsto_iff S r r v v as [ ? ? ].
    auto.
  Qed.

  Lemma get_update_different : forall S r r' v v',
    get S r v ->
    exists v'', get (update S r' v') r v''.
  Proof.
    intros S r r' v v' get_S.
    unfold get, update in *.
    destruct (eq_var_dec r r').
    subst. pose proof MapFacts.add_mapsto_iff S r' r' v' v' as [ ? ? ]. eauto.
    pose proof MapFacts.add_mapsto_iff S r' r v' v as [ ? ? ]. eauto.
  Qed.

  Lemma get_before_update : forall S r r' v v',
    get (update S r' v') r v ->
    r <> r' ->
    get S r v.
  Proof.
    intros S r r' v v' get_update_S r_neq.
    unfold get, update in *.
    assert (r' <> r) as r'_neq; auto.
    pose proof MapFacts.add_neq_mapsto_iff S v' v r'_neq as get_S.
    apply get_S. auto.
  Qed.

  Lemma get_after_update : forall S r r' v v',
    get S r v ->
    r <> r' ->
    get (update S r' v') r v.
  Proof.
    intros S r r' v v' get_S r_neq.
    unfold get, update in *.
    assert (r' <> r) as r'_neq; auto.
    pose proof MapFacts.add_neq_mapsto_iff S v' v r'_neq as get_update_S.
    apply get_update_S. auto.
  Qed.

  Lemma fresh_before_update : forall S r r' v,
    fresh (update S r' v) r ->
    fresh S r.
  Proof.
    unfold fresh, update.
    auto with var_maps.
  Qed.

  Lemma fresh_after_update : forall S r r' v,
    fresh S r ->
    r <> r' ->
    fresh (update S r' v) r.
  Proof.
    unfold fresh, update.
    eauto with var_maps.
  Qed.

  Lemma not_fresh_after_update : forall S r v,
    ~ fresh (update S r v) r.
  Proof.
    intros S r v false.
    unfold fresh, update in false. apply false.
    unfold Map.In. exists v. apply Map.add_1.
    unfold Var_as_OT.eq. auto.
  Qed.

  Lemma not_fresh_after_get : forall S r v,
    get S r v ->
    ~ fresh S r.
  Proof.
    intros S r v get_S fresh_S.
    unfold get, fresh, Map.In in *. apply fresh_S.
    eauto.
  Qed.

  Lemma eq_update_update : forall S r r' v v',
    r <> r' ->
    update (update S r v) r' v' = update (update S r' v') r v.
  Proof.
    intros. unfold update.
    apply Map.extensionality.
    auto with var_maps.
  Qed.

  Lemma fresh_empty : forall r,
    fresh empty r.
  Proof.
    unfold fresh, empty, Map.In.
    intros r [ e false ].
    contradiction (Map.empty_1 false).
  Qed.
End Store.
