Require Import coherent.
Require Import happens_before.
Require Import init_wb.

Set Implicit Arguments.

Open Scope nat_scope.

Import WB.
Import L_WB.

Lemma happens_before_plus_implies_release_or_spawn :
  forall tr t1 a1 j1 t2 a2 j2,
    index tr j1 (t1, a1) ->
    index tr j2 (t2, a2) ->
    happens_before_plus tr j1 j2 ->
    t1 <> t2 ->
    exists j3, j3 <= j1 /\  happens_before_plus tr j3 j2 /\
      ((exists l, index tr j3 (t1, release _ l))
        \/ exists t3, index tr j3 (t1, spawn _ t3)).
Proof.
  intros tr t1 a1 j1 t2 a2 j2 idx1 idx2 hb12 t1_neq_t2.
  generalize dependent a1.
  induction hb12 as [ j1 j2 hb12 | j1 j3 j2 hb13 hb32 ]; intros a1 idx1.

  (* Base case. Either j1 hb j2 by program order (vacuous) or by sync order (in
     which case j1 is the witness. *)
  inversion hb12 as [ ? ? ? ? ? idx1' idx2' | ? ? ? ? ? idx1' idx2'
    | ? ? ? ? ? idx1' idx2' ]; subst;
  pose proof index_unique idx1 idx1' as eq1; inversion eq1; subst;
  pose proof index_unique idx2 idx2' as eq2; inversion eq2; subst.
  tauto.
  exists j1. split. auto. split. apply Rplus_0. auto. eauto.
  exists j1. split. auto. split. apply Rplus_0. auto. eauto.

  (* Is j1 hb j3 by program order or sync order? *)
  inversion hb13 as [ ? ? ? a3 ? idx1' idx3 | ? ? ? ? ? idx1'
    | ? ? ? ? ? idx1' ]; subst;
    pose proof index_unique idx1 idx1' as eq1; inversion eq1; subst.

  (* Program order. Then we can apply the induction hypothesis to j3 and j2. *)
  pose proof IHhb32 idx2 _ idx3 as [ j4 [ ? ? ] ].
  exists j4. split. omega. auto.

  (* Sync order. j1 is our witness. *)
  esplit. split. apply le_refl. split. eapply Rplus_n; eauto. eauto.

  (* Spawn order. j1 is our witness. *)
  esplit. split. apply le_refl. split. eapply Rplus_n; eauto. eauto.
Qed.

Lemma well_synchronized_implies_release_or_spawn :
  forall tr t1 a1 j1 t2 a2 j2,
    well_synchronized tr ->
    index tr j1 (t1, a1) ->
    index tr j2 (t2, a2) ->
    data_race a1 a2 ->
    t1 <> t2 ->
    j2 < j1 ->
    exists j3, j2 < j3 < j1 /\
      ((exists l, index tr j3 (t1, release _ l))
        \/ exists t3, index tr j3 (t1, spawn _ t3)).
Proof.
  intros tr t1 a1 j1 t2 a2 j2 sync idx1 idx2 race12 t1_neq_t2 j2_lt.
  pose proof sync _ _ _ _ _ _ idx1 idx2 race12 j2_lt as hb12.
  pose proof happens_before_plus_implies_release_or_spawn idx1 idx2 hb12
    t1_neq_t2 as [ j3 [ j3_le [ hb32 idx3 ] ] ].
  exists j3. split; auto.
  split. eapply happens_before_plus_implies_lt. eauto.
  destruct (le_lt_eq_dec _ _ j3_le) as [ | ]; auto.
  subst.
  destruct idx3 as [ [ l idx3 ] | [ t3 idx3 ] ];
    pose proof index_unique idx1 idx3 as eq1; inversion eq1; subst;
      inversion race12.
Qed.

(* This is lemma 5.7 from the paper. If at any point the buffer for a thread
   and location is nonempty, then (1) that thread must have done a write to that
   location and (2) that thread has not performed an unlock since doing said
   write. *)
Lemma not_empty_buffer_implies_write_and_no_unlock :
  forall P0 S B Lambda P tr t r,
    system_step_many empty_heap P0 (S, B, Lambda) P tr ->
    ~ Buffer.is_empty_queue B t r ->
    exists j, exists v,
      index (erase_trace tr) j (t, write r v) /\
      ~ (exists j', j' < j /\
        ((exists l, index (erase_trace tr) j' (t, release _ l))
          \/ (exists t', index (erase_trace tr) j' (t, spawn _ t')))).
Proof.
  intros P0 S2 B2 Lambda2 P2 tr t r steps_many not_empty.

  (* Proof by induction on steps_many. *)
  remember empty_heap as H0.
  remember (S2, B2, Lambda2) as H2.
  generalize dependent Lambda2.
  generalize dependent S2.
  generalize dependent B2.
  induction steps_many as [ H0 | H0 P0 H1 P1 H2 P2 tr e steps_many ? steps ];
    intros B2 not_empty S2 Lambda2 H2_eq.

  (* zero steps - vacuous *)
  subst H0. unfold empty_heap in *. inversion H2_eq; subst.
  contradiction not_empty.
  pose proof Buffer.empty_is_empty as empty. apply empty.

  (* at least one step *)
  subst H0 H2.
  pose proof IHsteps_many (refl_equal _) as IH. clear IHsteps_many.
  pose proof IH B2 not_empty as IH_same_buffer.
  inversion steps as [ ? ? ? ? t1 a1 program_steps heap_steps
    | ? ? ? heap_steps | ]; subst.
  inversion heap_steps as
    [ S1 B1 Lambda1 ? r1 v1 fresh1 (* alloc *)
      | S1 B1 Lambda1 ? r1 v1 empty_t1_r1 get_r1 (* read *)
      | S1 B1 Lambda1 ? r1 v1 (* write *)
      |
      | S1 B1 Lambda1 ? l ? (* acquire *)
      | S1 B1 Lambda1 ? l ? is_empty_B1_t1 (* release *)
      | S1 B1 Lambda1 ? t' is_empty_B1_t1 (* spawn *)
      | S1 B1 Lambda1 ? r1 v1 top_t1_r1 (* read2 *)
    ]; subst.

  (* alloc *)
  pose proof IH_same_buffer S1 Lambda2 (refl_equal _)
    as [ j [ v [ idx no_release ] ] ].
  exists (S j). exists v. split.
  apply index_S. auto.
  intros [ j'  [ j'_lt [ [ l idx' ] | [ t' idx' ] ] ] ];
    inversion idx'; subst.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.

  (* read *)
  pose proof IH_same_buffer S2 Lambda2 (refl_equal _)
    as [ j [ v [ idx no_release ] ] ].
  exists (S j). exists v. split.
  apply index_S. auto.
  intros [ j'  [ j'_lt [ [ l idx' ] | [ t' idx' ] ] ] ];
    inversion idx'; subst.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.

  (* write *)
  destruct ((t, r) ==== (t1, r1)) as [ [ ? ? ] | ].

  (* the queue of interest is the one we just enqueued a value in *)
  simpl in *. subst t1 r1. exists 0. exists v1. split. apply index_0.
  intros [ j' [ false ? ] ]. inversion false.

  (* the queue of interest is *not* the one we just enqueued a value in. *)
  assert (~ Buffer.is_empty_queue B1 t r) as not_empty_B1.
  intros empty_B1. apply not_empty.
  rewrite <- Buffer.is_empty_queue_enqueue_diff; auto.
  auto using VarPair.eq_sym.
  pose proof IH B1 not_empty_B1 S2 Lambda2 (refl_equal _)
    as [ j [ v [ idx no_release ] ] ].
  exists (S j). exists v. split.
  apply index_S. auto.
  intros [ j'  [ j'_lt [ [ l idx' ] | [ t' idx' ] ] ] ];
    inversion idx'; subst.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.

  (* acquire *)
  pose proof IH_same_buffer S2 Lambda1 (refl_equal _)
    as [ j [ v [ idx no_release ] ] ].
  exists (S j). exists v. split.
  apply index_S. auto.
  intros [ j'  [ j'_lt [ [ l' idx' ] | [ t' idx' ] ] ] ];
    inversion idx'; subst.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.

  (* release *)
  pose proof IH_same_buffer S2 Lambda1 (refl_equal _)
    as [ j [ v [ idx no_release ] ] ].
  destruct (eq_var_dec t t1) as [ | ].
  subst t1. unfold Buffer.is_empty_tid in is_empty_B1_t1.
  pose proof is_empty_B1_t1 r. tauto.
  exists (S j). exists v. split.
  apply index_S. auto.
  intros [ j'  [ j'_lt [ [ l' idx' ] | [ t' idx' ] ] ] ];
    inversion idx'; subst.
  tauto.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.

  (* spawn *)
  pose proof IH_same_buffer S2 Lambda2 (refl_equal _)
    as [ j [ v [ idx no_release ] ] ].
  destruct (eq_var_dec t t1) as [ | ].
  subst t1. unfold Buffer.is_empty_tid in is_empty_B1_t1.
  pose proof is_empty_B1_t1 r. tauto.
  exists (S j). exists v. split.
  apply index_S. auto.
  intros [ j'  [ j'_lt [ [ l' idx' ] | [ t'' idx' ] ] ] ];
    inversion idx'; subst.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.
  tauto.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.

  (* read2 *)
  pose proof IH_same_buffer S2 Lambda2 (refl_equal _)
    as [ j [ v [ idx no_release ] ] ].
  exists (S j). exists v. split.
  apply index_S. auto.
  intros [ j'  [ j'_lt [ [ l idx' ] | [ t' idx' ] ] ] ];
    inversion idx'; subst.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.

  (* empty step - commit *)
  inversion heap_steps as [ | | | S1 B1 Lambda1 ? r1 v1 ? B2_eq | | | | ];
    subst.
  assert (~ Buffer.is_empty_queue B1 t r) as not_empty_B1.
  intro is_empty_B1.
  pose proof Buffer.is_empty_queue_dequeue is_empty_B1 B2_eq. tauto.
  pose proof IH B1 not_empty_B1 S1 Lambda2 (refl_equal _)
    as [ j [ v [ idx no_release ] ] ].
  exists j. exists v. split.
  simpl. auto.
  intros [ j'  [ j'_lt [ [ l' idx' ] | [ t'' idx' ] ] ] ];
    simpl in idx'.
  apply no_release. esplit. split. eauto. eauto.
  apply no_release. esplit. split. eauto. eauto.

  (* pure step *)
  pose proof IH_same_buffer S2 Lambda2 (refl_equal _)
    as [ j [ v [ idx no_release ] ] ].
  exists (S j). exists v. split.
  apply index_S. auto.
  intros [ j'  [ j'_lt [ [ l' idx' ] | [ t'' idx' ] ] ] ];
    inversion idx'; subst.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.
  apply no_release. esplit. split. apply lt_S_n. eauto. eauto.
Qed.

(* If a trace is well-synchronized and ends with a read, then only the
   thread doing the read has any values buffered for the location
   being read. *)
Corollary well_synchronized_trace_with_read_implies_empty_except :
  forall P0 S B Lambda P tr t r v S' B' Lambda' P',
    system_step_many WB.empty_heap P0 (S, B, Lambda) P tr ->
    system_step (S, B, Lambda) P (Some (t, read r v)) (S', B', Lambda') P' ->
    well_synchronized ((t, read r v) :: erase_trace tr) ->
    empty_except B t r.
Proof.
  intros P0 S B Lambda P tr t r v S' B' Lambda' P' steps_many steps sync t'
    concurrent.

  inversion steps as [ ? ? ? ? ? ? ? heap_steps | | ]; subst.
  inversion heap_steps as [ | ? ? ? ? ? ? empty_B_t ? | | | | | | ]; subst.

  (* t' is not the reading thread. Proof by contradiction - assume the
     buffer is not empty, and demonstrate that this is impossible. *)
  apply dec_not_not. apply Buffer.is_empty_queue_dec.
  intro not_empty_B_t'.
  pose proof not_empty_buffer_implies_write_and_no_unlock steps_many
    not_empty_B_t' as [ j [ v' [ idx_write no_release ] ] ].

  (* (t, read r v) appears at index 0 *)
  pose proof index_0 (erase_trace tr) (t, read r v) as idx_read.

  (* (t', write r v') appears somewhere in the trace too *)
  pose proof index_S (t, read r v) idx_write as idx_write'.

  (* read r v and write r v' conflict *)
  pose proof data_race_wr r v' v as races.

  (* 0 < S j' *)
  assert (0 < S j) as zero_lt. auto with arith.

  (* use the well-synchronized property *)
  pose proof well_synchronized_implies_release_or_spawn sync idx_write'
    idx_read races (sym_not_equal concurrent) zero_lt
    as [ j3 [ [ ? ? ] [ [ ? idx_release ] | [ ? idx_spawn ] ] ] ].
  inversion idx_release as [ | ? j' ]. subst.
  assert (j' < j) as j'_lt. auto with arith.
  apply no_release. esplit. split. eauto. eauto.
  inversion idx_spawn as [ | ? j' ]. subst.
  assert (j' < j) as j'_lt. auto with arith.
  apply no_release. esplit. split. eauto. eauto.

  (* t' is not the reading thread. Proof by contradiction - assume the
     buffer is not empty, and demonstrate that this is impossible. *)
  apply dec_not_not. apply Buffer.is_empty_queue_dec.
  intro not_empty_B_t'.
  pose proof not_empty_buffer_implies_write_and_no_unlock steps_many
    not_empty_B_t' as [ j [ v' [ idx_write no_release ] ] ].

  (* (t, read r v) appears at index 0 *)
  pose proof index_0 (erase_trace tr) (t, read r v) as idx_read.

  (* (t', write r v') appears somewhere in the trace too *)
  pose proof index_S (t, read r v) idx_write as idx_write'.

  (* read r v and write r v' conflict *)
  pose proof data_race_wr r v' v as races.

  (* 0 < S j' *)
  assert (0 < S j) as zero_lt. auto with arith.

  (* use the well-synchronized property *)
  pose proof well_synchronized_implies_release_or_spawn sync idx_write'
    idx_read races (sym_not_equal concurrent) zero_lt
    as [ j3 [ [ ? ? ] [ [ ? idx_release ] | [ ? idx_spawn ] ] ] ].
  inversion idx_release as [ | ? j' ]. subst.
  assert (j' < j) as j'_lt. auto with arith.
  apply no_release. esplit. split. eauto. eauto.
  inversion idx_spawn as [ | ? j' ]. subst.
  assert (j' < j) as j'_lt. auto with arith.
  apply no_release. esplit. split. eauto. eauto.
Qed.

(* If a trace is well-synchronized and ends with a write, then only the thread
   doing the write may have any values buffered for that location. *)
Corollary well_synchronized_trace_with_write_implies_empty_except :
  forall P0 P P' S S' B B' Lambda Lambda' tr t r v,
    system_step_many empty_heap P0 (S, B, Lambda) P tr ->
    system_step (S, B, Lambda) P (Some (t, write r v)) (S', B', Lambda') P' ->
    well_synchronized ((t, write r v) :: erase_trace tr) ->
    empty_except B t r.
Proof.
  intros P0 P P' S S' B B' Lambda Lambda' tr t r v steps_many steps sync
    t' concurrent.

  inversion steps as [ ? ? ? ? ? ? ? heap_steps | | ]; subst.
  inversion heap_steps as [ | ? ? ? ? ? ? empty_B_t ? | | | | | | ]; subst.

  (* t' is not the writing thread. Proof by contradiction - assume the
     buffer is not empty, and demonstrate that this is impossible. *)
  apply dec_not_not. apply Buffer.is_empty_queue_dec.
  intro not_empty_B_t'.
  pose proof not_empty_buffer_implies_write_and_no_unlock steps_many
    not_empty_B_t' as [ j [ v' [ idx_write1 no_release ] ] ].

  (* (t, read r v) appears at index 0 *)
  pose proof index_0 (erase_trace tr) (t, write r v) as idx_write2.

  (* (t', write r v') appears somewhere in the trace too *)
  pose proof index_S (t, write r v) idx_write1 as idx_write1'.

  (* read r v and write r v' conflict *)
  pose proof data_race_ww r v' v as races.

  (* 0 < S j' *)
  assert (0 < S j) as zero_lt. auto with arith.

  (* use the well-synchronized property *)
  pose proof well_synchronized_implies_release_or_spawn sync idx_write1'
    idx_write2 races (sym_not_equal concurrent) zero_lt
    as [ j3 [ [ ? ? ] [ [ ? idx_release ] | [ ? idx_spawn ] ] ] ].
  inversion idx_release as [ | ? j' ]. subst.
  assert (j' < j) as j'_lt. auto with arith.
  apply no_release. esplit. split. eauto. eauto.
  inversion idx_spawn as [ | ? j' ]. subst.
  assert (j' < j) as j'_lt. auto with arith.
  apply no_release. esplit. split. eauto. eauto.
Qed.
