Require Import Metatheory.

Require Import init_wb.

Set Implicit Arguments.

(** This module guarantees that write-buffering is type-preserving (for type
   systems that satisfy certain reasonable properties. *)
Module WB_Type_Preservation <: MEMORY_MODEL_TYPE_PRESERVATION Typ.
  Module M := WB.

  Import L.
  Import Typ.
  Import WB.

  Module MapFacts := FMapFacts.Facts Map.

  (** The heap is well-typed under Sigma if: (1) dom(S) = dom(Sigma); (2)
     the set of locations for which values are buffered in B is a subset of
     dom(Sigma); (3) all values in S typecheck to the corresponding type in
     Sigma; and (4) all values buffered in B typecheck to the corresponding
     type in Sigma. *)
  Definition well_typed (Sigma : heap_typing) (H : heap) : Prop :=
    let (SB, _) := H in let (S, B) := SB in
      (forall r, ~ Map.In r Sigma ->
        Store.fresh S r /\ Buffer.is_empty_loc B r) /\
      (forall r T, Map.MapsTo r T Sigma ->
        (exists v, Store.get S r v /\ typing_val Sigma v T) /\
        (Buffer.for_all_loc B (fun v => typing_val Sigma v T) r)).

  (** Well-typed implies that the domains of Sigma and S are equal. *)
  Lemma well_typed_fresh_iff : forall Sigma S B Lambda r,
    well_typed Sigma (S, B, Lambda) ->
    (~ Map.In r Sigma <-> Store.fresh S r).
  Proof.
    intros Sigma S B Lambda r Typ.
    destruct Typ as [ Dom Typ ].
    split. intros NotIn. apply* proj1.
    intros Fresh [ T Maps ].
    specializes* Typ (>>> Maps).
    destruct Typ as [ [ ? [ Get ? ] ] ? ].
    apply Store.not_fresh_after_get in Get. tauto.
  Qed.

  (** Well-typed implies that the domain of B is a subset of the domain of S. *)
  Lemma well_typed_fresh_buffer_empty : forall Sigma S B Lambda r,
    well_typed Sigma (S, B, Lambda) ->
    Store.fresh S r ->
    Buffer.is_empty_loc B r.
  Proof.
    intros Sigma S B Lambda r Typ Fresh.
    rewrite* <- well_typed_fresh_iff in Fresh.
    destruct Typ as [ Dom Typ ].
    apply* proj2.
  Qed.

  (** The empty heap is well-typed. *)
  Lemma well_typed_empty : well_typed empty_heap_typing empty_heap.
  Proof.
    split.

    (* Neither S nor B has any values for any locations. *)
    introv not_in. split.
    apply Store.fresh_empty.
    pose proof Buffer.empty_is_empty as is_empty.
    unfold Buffer.is_empty_loc, Buffer.is_empty in *. auto.

    (* All locations are well-typed - vacuous. *)
    introv Maps.
    unfold empty_heap_typing in Maps.
    contradiction (Map.empty_1 Maps).
  Qed.

  Hint Constructors typing_incoming typing_outgoing.

  (** A well-typed heap produces well-typed values for reads and fresh locations
     for allocations. *)
  Lemma typing_step : forall Sigma H t a H',
    well_typed Sigma H ->
    heap_step H (Some (t, a)) H' ->
    typing_incoming Sigma a.
  Proof.
    introv Typ Step.
    destruct H as [ [ S B ] ? ].
    unfold well_typed in Typ.
    destruct Typ as [ Dom Typ ].
    inversion Step as [ | | | | | | | ? ? ? ? ? ? Top ]; subst; eauto.

    (* alloc *)
    apply typing_incoming_alloc.
      intros [ T Maps ].
      specializes* Typ (>>>Hyps __).
      destruct Typ as [ [ v' [ Get Typ ] ] ].
      apply Store.not_fresh_after_get in Get.
      tauto.

    (* read *)
    remember (Map.find r Sigma) as maybe_T.
      symmetry in Heqmaybe_T.
      destruct maybe_T as [ T | ].
      apply Map.find_2 in Heqmaybe_T.
      apply* typing_incoming_read.
      specializes* Typ (>>> Hyps __).
      destruct Typ as [ [ v' [ Get Typ ] ] ].
      assert (v = v'). eapply Store.get_unique; eauto. subst. auto.
      forwards fr : Dom r.
      intros [ T Maps ].
      apply Map.find_1 in Maps.
      rewrite Heqmaybe_T in Maps. inversion Maps.
      apply Store.not_fresh_after_get in H6.
      tauto.

    (* read2 *)
    remember (Map.find r Sigma) as maybe_T.
      symmetry in Heqmaybe_T.
      destruct maybe_T as [ T | ].
      apply Map.find_2 in Heqmaybe_T.
      apply* typing_incoming_read.
      specializes* Typ (>>> Hyps __).
      destruct Typ as [ ? Typ ].
      unfold Buffer.for_all_loc in Typ.
      specializes Typ t.
      eapply Buffer.for_all_top in Typ.
      apply Typ.
      auto.
      rewrite <- MapFacts.not_find_in_iff in Heqmaybe_T...
      specializes* Dom (>>> __ __).
      destruct Dom as [ ? BufferEmpty ].
      unfold Buffer.is_empty_loc in BufferEmpty.
      specializes BufferEmpty t.
      apply Buffer.empty_top in BufferEmpty.
      rewrite BufferEmpty in Top.
      inversion Top.
  Qed.

  Module TypeFacts := Type_Model_Facts Typ.
  Module VarMapFacts := VarMapFacts Map.

  Hint Unfold Var_as_OT.eq.

  (** Type preservation for mutual steps. *)
  Lemma preservation : forall Sigma H t a H' Sigma',
    well_typed Sigma H ->
    heap_step H (Some (t, a)) H' ->
    typing_outgoing Sigma a Sigma' ->
    well_typed Sigma' H'.
  Proof.
    introv Typ Step Out.
    forwards Inc : typing_step; eauto.
    forwards Ext : TypeFacts.typing_extends; eauto.
    destruct H as [ [ S B ] ? ].
    unfold well_typed in Typ.
    destruct Typ as [ Dom Typ ].
    inversions* Step; inversions* Out; inversions* Inc; split; auto.

    (* alloc - doms eq *)
    intros r' not_in.
    forwards fresh_r' : Dom (>>> r').
    intro in_sigma.
    unfold Map.In in *. destruct in_sigma.
    apply not_in. destruct (r == r').
    esplit. apply* Map.add_1.
    esplit. apply* Map.add_2.
    destruct fresh_r'. split*.
    apply* Store.fresh_after_update.
    intro. subst.
    apply not_in. unfold Map.In.
    esplit. apply* Map.add_1.

    (* alloc - well-typed *)
    intros r' T' Maps.
    destruct (r == r'). subst.
      assert (T = T'). eapply VarMapFacts.MapsTo_unique.
        apply Map.add_1. auto. eauto. subst.
      split. esplit. split.
        apply Store.get_update_same.
        apply* typing_val_stability.
      forwards Dom_r' : Dom (>>> r'). auto.
        destruct Dom_r' as [ ? is_empty ].
        unfold Buffer.is_empty_loc in is_empty.
        intros t'. specializes* is_empty t'.
        apply* Buffer.for_all_empty.
    forwards Typ' : Typ (>>> __ __ __).
      eapply Map.add_3; eauto.
      destruct Typ' as [ [ v' [ Get Typ' ] ] Typ'' ].
      split. exists v'. split.
        apply* Store.get_after_update.
        apply* typing_val_stability.
      unfold Buffer.for_all_loc in *.
      intros t'. eapply Buffer.for_all_generalize.
        eauto. intros v''. simpl. intros typing_v''.
        apply* typing_val_stability.

    (* write - doms equal *)
    intros r' NotIn.
    specializes* Dom (>>> __ __).
    destruct Dom as [ Fresh EmptyLoc ].
    split. auto. intros t'.
    rewrite* <- Buffer.is_empty_queue_enqueue_diff.
    intros [ ? ? ]. simpl in *. subst.
    apply NotIn. unfold Map.In. eauto.

    (* write - well-typed *)
    intros r' T' Maps.
    specializes* Typ (>>> __ __ __).
    destruct Typ as [ Typ' All' ].
    split. auto. intros t'.
    destruct ((t, r) ==== (t', r')) as [ [ ? ? ] | Neq ].
    simpl in *. subst.
      assert (T = T'). apply* VarMapFacts.MapsTo_unique. subst.
      apply* Buffer.for_all_enqueue_same.
    apply* Buffer.for_all_enqueue_diff.
      intro. apply Neq. apply* VarPair.eq_sym.
  Qed.

  (** Type preservation for empty steps. *)
  Lemma preservation_empty : forall Sigma H H',
    well_typed Sigma H ->
    heap_step H None H' ->
    well_typed Sigma H'.
  Proof.
    intros Sigma H1 H2 Typ1 Step.
    destruct H1 as [ [ S1 B1 ] Lambda1 ].
    destruct H2 as [ [ S2 B2 ] Lambda2 ].
    unfold well_typed in *.
    destruct Typ1 as [ Dom1 Typ1 ].
    inversion Step as [ | | | ? ? ? t1 r1 v1 ? Deq | | | | ]; subst.
    split.

    (* doms eq *)
    intros r2 NotIn.
    specializes* Dom1 (>>> __ __).
    destruct Dom1 as [ Fresh2 EmptyLoc2 ].
    assert (r1 <> r2). intro. subst.
      apply Buffer.dequeue_not_is_empty_queue in Deq.
      apply Deq. apply* EmptyLoc2.
    split. apply* Store.fresh_after_update.
    intros t2. rewrite Buffer.is_empty_queue_dequeue_diff.
      apply* EmptyLoc2. apply Deq.
      intros [ ? ? ]. simpl in *. tauto.

    (* well-typed *)
    intros r2 T2 Maps.
    specializes* Typ1 (>>> __ __ __).
    destruct Typ1 as [ [ ? [ ? ? ] ] All1 ].
    split.
    destruct (r1 == r2). subst.
      esplit. split. apply Store.get_update_same.
        forwards* Typ2 : Buffer.for_all_dequeue_val.
      esplit. split*. apply* Store.get_after_update.
    unfold Buffer.for_all_loc in *. intro t2.
      specializes All1 (>>> t2).
      forwards* All : Buffer.for_all_dequeue.
  Qed.
End WB_Type_Preservation.
